#include <nitro.h>
#include "dbs_console.h"
#include "dbs_print.h"


typedef struct tLine
{
    u8 color;
    char str[1];
} tLine;


static tDbsConsole* sConsoleList; // Console link list
static tDbsConsole* sDefaultConsole; // The default console is used when this parameter has been set to NULL



// static functions
static void CorrectOutOfRangeDisplay( tDbsConsole* con );
static const char* GetString( const tDbsConsole* con, u32 line );
static int GetColor( const tDbsConsole* con, u32 line );
static BOOL IsInsideBuffer( const tDbsConsole* con, u32 line );
static void Print( tDbsConsole* console, int color, const char* fmt, va_list vlist );
static tLine* GetLine( tDbsConsole* con, u32 line );
static void Draw( tDbsConsole* console );
static tDbsConsole* GetTarget( tDbsConsole* arg );
static void AppendConsole( tDbsConsole* appending );





/*---------------------------------------------------------------------------*

 Create console
  
 *---------------------------------------------------------------------------*/
void dbs_CreateConsole( tDbsConsole* console, void* linebuf, u32 linebuf_size, int x, int y, int width, int height )
{
    SDK_NULL_ASSERT( console );

    console->linenum         = linebuf_size / ( width + 1 ) ; // Divide by width + 1 since 1-byte is used to save color
    console->currentline     = 0;
    console->nowcolumn       = 0;
    console->displayline     = 0;
    MI_CpuClear8( linebuf, linebuf_size );
    console->linebuf         = (u8*)linebuf;
    AppendConsole( console );
    console->x               = (u8) x;
    console->y               = (u8) y;
    console->width           = (u8) width;
    console->height          = (u8) height;
    console->visible         = (u8) TRUE;
    console->linenumber_width = (u8) 0;
    console->stop_bufferfull = (u8) FALSE;
    console->default_color   = DBS_BLACK;

    // automatic default when it's the first one
    if ( sDefaultConsole == NULL )
    {
        sDefaultConsole = console;
    }
    
}




/*---------------------------------------------------------------------------*
  
  Delete console
  
 *---------------------------------------------------------------------------*/
void dbs_DestroyConsole( tDbsConsole* removing )
{
    tDbsConsole* con; // Variable into which the addresses of the next field for the tDbsConsole are stored

    if ( removing == NULL ) return;

    for( con = sConsoleList; ; con = con->next )
    {
        if ( !con ) break;

        if ( con == removing )
        {
            // The first element is removed.
            sConsoleList = removing->next;

            break;
        }
        else if ( con->next == removing )
        {
            // After that.
            con->next = removing->next;

            break;
        }
    }

    if ( removing == sDefaultConsole )
    {
        sDefaultConsole = sConsoleList;
    }
}





/*---------------------------------------------------------------------------*
  
  Destroys all consoles being managed
  
 *---------------------------------------------------------------------------*/
void dbs_CleanupConsole( void )
{
    sConsoleList    = NULL;
    sDefaultConsole = NULL;
}



/*---------------------------------------------------------------------------*
  
  Render on console. This can be done merely by calling the dbs_print function
  without touching the graphics engine at all.
  
 *---------------------------------------------------------------------------*/
void dbs_DrawConsole( void )
{
    tDbsConsole* con;

    if ( ! sConsoleList ) return ; // Does nothing if not even one console could be created

    for ( con = sConsoleList; con != NULL; con = con->next )
    {
        if ( con->visible )
        {
            Draw( con );
        }
    }

}




/*---------------------------------------------------------------------------*
  
  Displays text on the default console
  
 *---------------------------------------------------------------------------*/
void dbs_Puts( const char* fmt, ... )
{
    va_list vlist;
    
    if ( ! sDefaultConsole ) return;

    va_start( vlist, fmt );
    Print( sDefaultConsole, sDefaultConsole->default_color, fmt, vlist );
    va_end( vlist );
    
}




/*---------------------------------------------------------------------------*
  
  Displays color text on the default console
  
 *---------------------------------------------------------------------------*/
void dbs_PutsColor( int color, const char* fmt, ... )
{
    va_list vlist;
    
    if ( ! sDefaultConsole ) return;

    va_start( vlist, fmt );
    Print( sDefaultConsole, color, fmt, vlist );
    va_end( vlist );

}




/*---------------------------------------------------------------------------*
  
  Displays text on the specified console. Display is made on the default console if nULL is set for the console
  
 *---------------------------------------------------------------------------*/
void dbs_CPrintf( tDbsConsole* console, const char* fmt, ... )
{
    va_list vlist;

    va_start( vlist, fmt );
    dbs_CVPrintf( console, fmt, vlist );
    va_end( vlist );
}

void dbs_CVPrintf( tDbsConsole* console, const char* fmt, va_list vlist )
{
    tDbsConsole* target;
    
    if ( ! sConsoleList ) return;

    target = GetTarget( console );

    Print( target, target->default_color, fmt, vlist );
}




/*---------------------------------------------------------------------------*
  
  Displays text using the specified console and color. Display is made on the default console if nULL is set for the console
  
 *---------------------------------------------------------------------------*/
void dbs_PrintColorConsole( tDbsConsole* console, int color, const char* fmt, ... )
{
    tDbsConsole* target;
    va_list vlist;
    
    if ( ! sConsoleList ) return;

    target = GetTarget( console );


    va_start( vlist, fmt );
    Print( target, color, fmt, vlist );
    va_end( vlist );
    
}





/*---------------------------------------------------------------------------*
  
  Delete all console contents including buffers.
  
 *---------------------------------------------------------------------------*/
void dbs_ClearConsole( tDbsConsole* console )
{
    tDbsConsole* con = GetTarget( console );
    
    con->currentline     = 0;
    con->displayline     = 0;
    MI_CpuClear8( con->linebuf, sizeof(u8) * ( con->width + 1 ) * ( con->linenum ) );
    
}





/*---------------------------------------------------------------------------*

  Move the visible area. Forward if the value of 'lines' is positive, reverse if the value of 'lines' is negative.
  
 *---------------------------------------------------------------------------*/
void dbs_ScrollConsole( tDbsConsole* console, int lines )
{
    tDbsConsole* con = GetTarget( console );

    // If 'displayline' is about to go zero, it's stopped at zero.
    if ( lines < 0 && con->displayline < -lines )
    {
        con->displayline = 0;
        return ;
    }

    con->displayline += lines;

    CorrectOutOfRangeDisplay( con );

}




/*---------------------------------------------------------------------------*

 Made so that the olded line is visible as long as it's in the buffer
  
 *---------------------------------------------------------------------------*/
void dbs_ScrollConsoleToFirstLine( tDbsConsole* console )
{
    tDbsConsole* con = GetTarget( console );

    con->displayline = con->currentline - con->linenum;

}



/*---------------------------------------------------------------------------*
  
  Scroll to the position where only the last line is visible.
  
 *---------------------------------------------------------------------------*/
void dbs_ScrollConsoleToLastLine( tDbsConsole* console )
{
    tDbsConsole* con = GetTarget( console );

    con->displayline = con->currentline - 1;

}




/*---------------------------------------------------------------------------*
  
  Return to the previous page
  
 *---------------------------------------------------------------------------*/
void dbs_ScrollConsoleToPrevPage( tDbsConsole* console )
{
    tDbsConsole* con = GetTarget( console );
    
    dbs_ScrollConsole( con, - con->height );
}




/*---------------------------------------------------------------------------*
  
  Go to the next page
  
 *---------------------------------------------------------------------------*/
void dbs_ScrollConsoleToNextPage( tDbsConsole* console )
{
    tDbsConsole* con = GetTarget( console );

    dbs_ScrollConsole( con, con->height );
}




/*---------------------------------------------------------------------------*
  
  Change the default console
  
 *---------------------------------------------------------------------------*/
void dbs_SetDefaultConsole( tDbsConsole* console )
{
    SDK_ASSERT( console );

    sDefaultConsole = console;

}



/*---------------------------------------------------------------------------*
  
  Get the default console
  
 *---------------------------------------------------------------------------*/
tDbsConsole* dbs_GetDefaultConsole( void )
{
    return sDefaultConsole;
}






/*---------------------------------------------------------------------------*
  
  Switch the console display
  
 *---------------------------------------------------------------------------*/
void dbs_SetConsoleVisibility( tDbsConsole* console, BOOL show )
{
    GetTarget( console )->visible = (u8)show;
}



/*---------------------------------------------------------------------------*
  
  Get the viewable status of the console
  
 *---------------------------------------------------------------------------*/
BOOL dbs_GetConsoleVisibility( tDbsConsole* console )
{
    return GetTarget( console )->visible;
}



/*---------------------------------------------------------------------------*
  
  Sets how many lines to use when displaying line numbers on the console. During printing, a line number
  is inserted at the start of each character string.
  This number cannot be added and removed dynamically.
  
 *---------------------------------------------------------------------------*/
void dbs_SetConsoleLineNumberWidth( tDbsConsole* console, int width )
{
    SDK_ASSERT( width >= 0 );

    GetTarget( console )->linenumber_width = (u8)width;
    
}




/*---------------------------------------------------------------------------*
  
  Specify what to do if the buffer becomes full
  
 *---------------------------------------------------------------------------*/
void dbs_SetConsoleStopOnBufferFull( tDbsConsole* console,  BOOL stop )
{
    GetTarget( console )->stop_bufferfull = (u8)stop;
}




/*---------------------------------------------------------------------------*
  
  Gets what to do if the buffer becomes full
  
 *---------------------------------------------------------------------------*/
BOOL dbs_GetConsoleStopOnBufferFull( const tDbsConsole* console )
{
    return GetTarget( (tDbsConsole*)console )->stop_bufferfull;
}



/*---------------------------------------------------------------------------*
  
  Add to the list
  
 *---------------------------------------------------------------------------*/
static void AppendConsole( tDbsConsole* appending )
{
    appending->next = sConsoleList;
    sConsoleList = appending;
}





static void Print( tDbsConsole* console, int color, const char* fmt, va_list vlist )
{
    int i;
    tLine* linebuf = GetLine( console, console->currentline );
    char str[64];
    
    (void) OS_VSNPrintf( str, sizeof(str), fmt, vlist );

    // Colors are handled line-by-line.
    linebuf->color = (u8)color;

    i = 0;
    while ( str[i] )
    {
        if ( console->nowcolumn == 0 )
        {
            linebuf = GetLine( console, console->currentline );
            linebuf->color = (u8)color;

            // Clear the line buffer to be written to ahead of time
            MI_CpuFill8( linebuf->str, ' ', console->width );
        }

        // Draw characters
        if ( str[i] == 0x0d || str[i] == 0x0a )
        {
            // Newline
            console->nowcolumn = 0;
            console->currentline ++;

            // DOS newline support
            if ( str[i] == 0x0d && str[i+1] == 0x0a )
            {
                i+=2;

                continue;
            }

            // UNIX/MAC newlines
            i++;
            continue;
        }
        else if ( console->nowcolumn < console->width )
        {
            linebuf->str[console->nowcolumn] = str[i];

            console->nowcolumn++;

            if ( console->nowcolumn == console->width )
            {
                // General processing
                console->nowcolumn = 0;
                console->currentline ++;
            }
        }

        i++;
    }

    // Automatically made to scroll when looking at the lowest part of the console
    if( console->currentline > console->displayline + console->height )
    {
        // Work-around because several lines may be skipped.
        console->displayline = console->currentline - console->height + 1;
    }
}


static void Draw( tDbsConsole* console )
{
    int i;

    for ( i = 0; i < console->height; i++ )
    {
        char tmpstr[64]; // Since there is no null character part in the string buffer for the console, it's made here.
        
        MI_CpuCopy8( GetString( console, console->displayline + i ), tmpstr, console->width );
        tmpstr[ console->width ] = '\0';
        
        dbs_Print( console->x, console->y + i, tmpstr );
    }
    
}


static tDbsConsole* GetTarget( tDbsConsole* arg )
{
    if ( arg )
    {
        return arg;
    }
    else
    {
        return sDefaultConsole;
    }
    
}


static tLine* GetLine( tDbsConsole* con, u32 line )
{
    int offset; // Which byte does the start of the line match?

	offset = (int)(( line % con->linenum ) * ( con->width + 1 ));

    return (tLine*)( & con->linebuf[ offset ] ) ;
    
}



/*---------------------------------------------------------------------------*
  
  Get character string from console buffer If a range that extends beyond the buffer was specified,
  a pointer to 32 blank characters is returned.
  
 *---------------------------------------------------------------------------*/
static const char* GetString( const tDbsConsole* con, u32 line )
{
                                  // 12345678901234567890123456789012
    static const char* whitespace = "                                ";

    if ( IsInsideBuffer( con, line ) )
    {
        return GetLine( (tDbsConsole*)con, line )->str;
    }
    else
    {
        return whitespace;
    }
}


static int GetColor( const tDbsConsole* con, u32 line )
{
    if ( IsInsideBuffer( con, line ) )
    {
        return GetLine( (tDbsConsole*)con, line )->color;
    }
    else
    {
        return DBS_BLACK;
    }
}


static BOOL IsInsideBuffer( const tDbsConsole* con, u32 line )
{
    // Two u32 values are compared with attention paid to possible overflow.&#x81;u Used in place of v
    // Returns invalid results if the difference exceed 0xFFFFFFFF/2 (=2147483647).
#define GREATER( x, y )  ( 0 < (u32)((x) - (y)) && (u32)((x) - (y)) <= 0xFFFFFFFF/2 )

    u32 bufstart;

    bufstart = con->currentline - con->linenum - 1;
    
    if ( GREATER( line, bufstart ) &&
         GREATER( con->currentline, line ) )
    {
        return TRUE;
    }
    else
    {
        return FALSE;
    }
}




static void CorrectOutOfRangeDisplay( tDbsConsole* con )
{

    // The oldest line is left visible if returned to often.
    if ( con->displayline < con->currentline - con->linenum )
    {
        dbs_ScrollConsoleToFirstLine( con );
    }

}
