/*---------------------------------------------------------------------------*

  Display for on-screen debugging

  - Displays text on the BG screen
  - Use with 16-color, 256x256 screens only
  
 *---------------------------------------------------------------------------*/


#include <nitro.h>
#include "dbs_print.h"
#include "dbs_print_data.h"




static BOOL          s_initialized;                                 // Flag representing whether or not initialization is finished
static tDbsInitParam s_initparam;                                   // For saving initialization parameters
static GXScrFmtText  s_screen_data[24][32];                         // Screen data for text string
static void*         s_screen_loadaddr;                             // Address in which screen is to be loaded
static int           s_charname_top;                                // Lead character frame for font data
static int           s_pltt_base;                                   // Start of palette no.
static int           s_pltt_num;                                    // Number of palettes to be used for this text string
static u8            s_converted_font[ sizeof( s_font_chardata ) ]; // Font after pixel color is changed
static char          s_format_buffer[ 256 ];                        // Buffer for text string format.
static GXRgb         s_fgcolor[] = { GX_RGB(  0,  0,  0 ),          // black
                                    GX_RGB( 31,  0,  0 ),           // red
                                    GX_RGB(  0, 31,  0 ),           // green
                                    GX_RGB( 31, 25,  0 ),           // yellow
                                    GX_RGB(  0,  0, 31 ),           // blue
                                    GX_RGB( 31,  0, 31 ),           // magenta
                                    GX_RGB(  0, 31, 31 ),           // cyan
                                    GX_RGB( 31, 31, 31 )            // white
};
static GXRgb         s_bgcolor = GX_RGB( 31, 31, 31 );



// static functions
static int  GetEscapeCode( const char* p, int * advance );
static void DummyVBlankIntr(void);
static void PutString             ( int x, int y, const char* str, u32 strbufsize );
static void ConvertFontPixelcolor ( const u8* font_in, u8* font_out, int fontsize, int fg_color, int bg_color );





/*---------------------------------------------------------------------------*
  
  Initializes text string drawing. May be called as many times as needed. The graphics engin is set
  each time this function is called.

  param: Where is the BG screen to be used for the text screen? enum value defined by dbs_print.h
  fontchar_location : Location for storing character data for fonts. Number of bytes from the character base on the BG screen
  fg_color Pixel color assigned as the text color for the string
  bg_color          : Pixel color assigned as the background color for the string
  pltt_base         : Beginning of palette to be used for this text string.
  pltt_num          : Which colors should be made available for use as the text color? Secures a palette with only those colors which have been used.

  Initial Settings Example 1

  fg_color  = 8
  bg_color  = 15
  pltt_base = 12
  When pltt_num = 4

  O O O O O O O O O O O O O O O O 
  O O O O O O O O O O O O O O O O 
  O O O O O O O O O O O O O O O O 
  O O O O O O O O O O O O O O O O 
  O O O O O O O O O O O O O O O O 
  O O O O O O O O O O O O O O O O 
  O O O O O O O O O O O O O O O O 
  O O O O O O O O O O O O O O O O 
  O O O O O O O O O O O O O O O O 
  O O O O O O O O O O O O O O O O 
  O O O O O O O O O O O O O O O O 
  O O O O O O O O O O O O O O O O 
  O O O O O O O O * O O O O O O * 
  O O O O O O O O * O O O O O O * 
  O O O O O O O O * O O O O O O * 
  O O O O O O O O * O O O O O O * 

  The * parts above are used for text strings

  
  Initial Settings Example 2
  
  fg_color  = 15
  bg_color  = 0
  pltt_base = 0
  When pltt_num = 2

  OOOOOOOOOOOOOOOOOOOOOOOOOOOOOOO*
  OOOOOOOOOOOOOOOOOOOOOOOOOOOOOOO*
  OOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOO
  OOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOO
  OOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOO
  OOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOO
  OOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOO
  OOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOO
  OOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOO
  OOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOO
  OOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOO
  OOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOO
  OOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOO
  OOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOO
  OOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOO
  OOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOO

  The * parts above are used for text strings In this case the background is transparent.
  
 *---------------------------------------------------------------------------*/
void dbs_Init       ( tDbsInitParam param, int fontchar_location, int fg_color, int bg_color, int pltt_base, int pltt_num )
{
    GXCharBGText16* bgchar;


    // Saves initial values
    s_initparam    = param;
	s_charname_top = fontchar_location / (int)sizeof( GXCharFmt16 );
    s_pltt_base    = pltt_base;
    s_pltt_num     = pltt_num;

    switch( param )
    {
    case eDBS_INIT_MAIN_BG0:
        bgchar = (GXCharBGText16*)G2_GetBG0CharPtr();
        s_screen_loadaddr = G2_GetBG0ScrPtr();
        break;
    case eDBS_INIT_MAIN_BG1:
        bgchar = (GXCharBGText16*)G2_GetBG1CharPtr();
        s_screen_loadaddr = G2_GetBG1ScrPtr();
        break;
    case eDBS_INIT_MAIN_BG2:
        bgchar = (GXCharBGText16*)G2_GetBG2CharPtr();
        s_screen_loadaddr = G2_GetBG2ScrPtr();
        break;
    case eDBS_INIT_MAIN_BG3:
        bgchar = (GXCharBGText16*)G2_GetBG3CharPtr();
        s_screen_loadaddr = G2_GetBG3ScrPtr();
        break;
    case eDBS_INIT_SUB_BG0:
        bgchar = (GXCharBGText16*)G2S_GetBG0CharPtr();
        s_screen_loadaddr = G2S_GetBG0ScrPtr();
        break;
    case eDBS_INIT_SUB_BG1:
        bgchar = (GXCharBGText16*)G2S_GetBG1CharPtr();
        s_screen_loadaddr = G2S_GetBG1ScrPtr();
        break;
    case eDBS_INIT_SUB_BG2:
        bgchar = (GXCharBGText16*)G2S_GetBG2CharPtr();
        s_screen_loadaddr = G2S_GetBG2ScrPtr();
        break;
    case eDBS_INIT_SUB_BG3:
        bgchar = (GXCharBGText16*)G2S_GetBG3CharPtr();
        s_screen_loadaddr = G2S_GetBG3ScrPtr();
        break;
    default:
        OS_Panic( "dbs_init : invalid parameter\n" );
        break;
    }


    // Process fonts
    if ( ! s_initialized ) // Processing is performed the first time only
    {
        ConvertFontPixelcolor( s_font_chardata, s_converted_font, sizeof(s_font_chardata), fg_color, bg_color );
    }

    // Load fonts
    MI_DmaCopy32( GX_GetDefaultDMA(), s_converted_font, & bgchar->ch[ s_charname_top ], sizeof( s_font_chardata ) );

    // Load palettes
    if( eDBS_INIT_MAIN_BG0 <= param && param <= eDBS_INIT_MAIN_BG3 ) // Upper screen
    {
        int i;

        for( i = 0 ; i < s_pltt_num; i++ )
        {
            GX_LoadBGPltt ( & s_fgcolor[ i ], sizeof(GXBGPltt16) * ( s_pltt_base + i ) + sizeof(GXRgb) * fg_color,  sizeof(GXRgb) );

            if( bg_color != 0 ) // The background is only loaded when it is other than transparent
            {
                GX_LoadBGPltt ( & s_bgcolor,  sizeof(GXBGPltt16) * ( s_pltt_base + i ) + sizeof(GXRgb) * bg_color,  sizeof(GXRgb) );
            }
        }
        
    }
    else // Lower screen
    {
        int i;

        for( i = 0 ; i < s_pltt_num; i++ )
        {
            GXS_LoadBGPltt ( & s_fgcolor[ i ], sizeof(GXBGPltt16) * ( s_pltt_base + i ) + sizeof(GXRgb) * fg_color,  sizeof(GXRgb) );

            if( bg_color != 0 ) // The background is only loaded when it is other than transparent
            {
                GXS_LoadBGPltt ( & s_bgcolor,      sizeof(GXBGPltt16) * ( s_pltt_base + i ) + sizeof(GXRgb) * bg_color,  sizeof(GXRgb) );
            }
        }
        
    }

    // Clear the screen buffer
    MI_CpuClear8( s_screen_data, sizeof( s_screen_data ) );

    s_initialized = TRUE;
    
}




/*---------------------------------------------------------------------------*
  
  Load screen data into the graphics engine. Just as with the normal load function,
  it is necessary to call this function during the VBLANK period.
  
 *---------------------------------------------------------------------------*/
void dbs_Load       ( void )
{
    if ( ! s_initialized ) return; // Do nothing if not initialized.

    DC_StoreRange( s_screen_data, sizeof( s_screen_data ) );
    MI_DmaCopy32( GX_GetDefaultDMA(), s_screen_data, s_screen_loadaddr, sizeof( s_screen_data ) );

    // Clear after loading
    MI_CpuClear8( s_screen_data, sizeof( s_screen_data ) );

}




/*---------------------------------------------------------------------------*
  
  Draw text string using screen coordinates (x, y) as the base point. Since a display of only one frame will result,
  this function should usually be called for each frame.
  
 *---------------------------------------------------------------------------*/
void dbs_Print      ( int x, int y, const char* fmt, ... )
{
    va_list va;
    va_start(va, fmt);
    (void)vsnprintf( s_format_buffer, sizeof(s_format_buffer), fmt, va );
    va_end(va);

    PutString( x, y, s_format_buffer, sizeof(s_format_buffer) );
}


void dbs_VPrint           ( int x, int y, const char *fmt, va_list vlist )
{
    (void)vsnprintf( s_format_buffer, sizeof(s_format_buffer), fmt, vlist );

    PutString( x, y, s_format_buffer, sizeof(s_format_buffer) );
    
}



/*---------------------------------------------------------------------------*
  
  Displays context information. This function will presumably be used as the user callback when an exception occurs.
   

 *---------------------------------------------------------------------------*/
void dbs_DisplayException ( u32 context_addr, void* /* arg*/ )
{
    static int bg_offset = 8 * 8; // Exception information is displayed offset and drawn off the screen.
    OSContext* context = (OSContext*) context_addr;
    int i;

    if ( ! s_initialized ) return; // Do nothing if not initialized.

    // Prints exception data.
    for( i=0; i<13; i++ )
    {
        dbs_Print( ( i % 2) * 16, i / 2, "R%02d=%08X ", i, context->r[i] );
    }

    dbs_Print( ( i % 2 ) * 16, i / 2, "SP =%08X ", context->sp );
    i++;
    
    dbs_Print( ( i % 2 ) * 16, i / 2, "LR =%08X ",  context->lr );
    i++;
    
    dbs_Print( ( i % 2 ) * 16, i / 2, "PC =%08X " , context->pc_plus4 );
    i++;


    DC_StoreRange( s_screen_data, sizeof( s_screen_data ) );

    // A VBLANK function which does nothing is set in order to enable interrupts.
    OS_SetIrqFunction( OS_IE_V_BLANK, DummyVBlankIntr );
    (void)OS_EnableIrqMask( OS_IE_V_BLANK );
    (void)OS_EnableIrq();
    (void)GX_VBlankIntr(TRUE);

    // Load during the VBLANK period
    OS_WaitIrq( 1, OS_IE_V_BLANK );
    MI_DmaCopy32( GX_GetDefaultDMA(), s_screen_data,
                  ((GXScrText32x32*)s_screen_loadaddr)->scr[24], // Load to an area off the screen
                  sizeof(GXScrFmtText) * 32 * 8 );

    // Main loop for moving the exception data display position using the controller
    while ( 1 )
    {
        u16 pad;

        pad = PAD_Read();

        if ( pad & PAD_KEY_UP )
        {
            bg_offset ++;
        }
        else if ( pad & PAD_KEY_DOWN )
        {
            bg_offset --;
        }

        if( bg_offset > 8 * 8 )
        {
            bg_offset = 8 * 8;
        }
        else if ( bg_offset < 0 )
        {
            bg_offset = 0;
        }
        
        switch( s_initparam )
        {
        case eDBS_INIT_MAIN_BG0:
            G2_SetBG0Offset( 0, bg_offset );
            break;
        case eDBS_INIT_MAIN_BG1:
            G2_SetBG1Offset( 0, bg_offset );
            break;
        case eDBS_INIT_MAIN_BG2:
            G2_SetBG2Offset( 0, bg_offset );
            break;
        case eDBS_INIT_MAIN_BG3:
            G2_SetBG3Offset( 0, bg_offset );
            break;
        case eDBS_INIT_SUB_BG0:
            G2S_SetBG0Offset( 0, bg_offset );
            break;
        case eDBS_INIT_SUB_BG1:
            G2S_SetBG1Offset( 0, bg_offset );
            break;
        case eDBS_INIT_SUB_BG2:
            G2S_SetBG2Offset( 0, bg_offset );
            break;
        case eDBS_INIT_SUB_BG3:
            G2S_SetBG3Offset( 0, bg_offset );
            break;
        }
        
        OS_WaitIrq( 1, OS_IE_V_BLANK );
    }

}





/*---------------------------------------------------------------------------*
  
  Re-writes the character data of the font. 4 bits per pixel for character data to be used
  Configured such that Transparent = 0, Color text = 1, and Background color = 2.
  
 *---------------------------------------------------------------------------*/
static void ConvertFontPixelcolor( const u8* font_in, u8* font_out, int fontsize, int fg_color, int bg_color )
{
    int i;
    int table[3]; //

    SDK_NULL_ASSERT( font_in );
    SDK_NULL_ASSERT( font_out );
    SDK_ASSERT( fontsize > 0 );
    SDK_ASSERT( fg_color < 16 );
    SDK_ASSERT( bg_color < 16 );

    table[0] = 0;
    table[1] = fg_color;
    table[2] = bg_color;
    
    for( i = 0; i < fontsize; i++ )
    {
        int high, low;

        high = table[ ( font_in[ i ] & 0xF0 ) >> 4 ];
        low  = table[ ( font_in[ i ] &  0xF ) >> 0 ];
        
        font_out[ i ] = (u8) ( ( high << 4 ) | low );

    }

}





/*---------------------------------------------------------------------------*
  
  Sets the text string for the screen
  
 *---------------------------------------------------------------------------*/
static void PutString( int x, int y, const char* str, u32 strbufsize )
{
    int i;
    int start_x = x;
    int color_index = 0; // s_fgcolor index

    for( i = 0 ; str[i] != '\0' && i < strbufsize; i++ )
    {
        if ( x < 0 )             continue; // Does nothing if out of range
        if ( y < 0 || y >= 24 )  continue; // Does nothing if out of range

        if ( str[i] == '\e' ) // Change color if the escape character is encountered
        {
            int code, advance;
            
            code = GetEscapeCode( & str[i], & advance );

            if ( code >= 0 )
            {
                color_index = code - DBS_BLACK;
            }

            i += advance - 1; // 1 is subtracted since it will be incremented using i++ in a for statement
            
            continue;
            
        }

        // Makes a newline if it is a newline character
        if ( str[i] == '\n' )
        {
            x = start_x;
            y ++;
            continue;
        }
        

        s_screen_data[ y ][ x ] = GX_SCRFMT_TEXT( s_pltt_base + color_index % s_pltt_num, 0, 0, s_charname_top + str[i] - ' ' );

        x ++;
        // Automatically causes a new line when the right edge is reached.
        if ( x >= 32 )
        {
            x = start_x;
            y ++;
        }

    }

}




/*---------------------------------------------------------------------------*
  
  A VBLANK function which which does nothing is set during exception display. Without this, there would be problems when reading dbs_Load using the VBLANK function.
   
  
 *---------------------------------------------------------------------------*/
static void DummyVBlankIntr(void)
{
    OS_SetIrqCheckFlag( OS_IE_V_BLANK );
}




/*---------------------------------------------------------------------------*

  Gets the text color change code

  p : Start of escape sequence
  advance : Variable for returning how many bytes were used

  returns:
  -1 Failure (an unknown escape sequence was encountered)
  It was 30 "\e[30m", in otherwords it was set to black
  &#x81;F
  &#x81;F
  It was 37 "\e[37m", in otherwords it was set to white
  
 *---------------------------------------------------------------------------*/
static int  GetEscapeCode( const char* p, int * advance )
{
    int code = -1; // Error value stored here
    
    SDK_ASSERT( *p == '\e' ); // Don't expect this function to be called except in the case of an escape sequence

    // Error check
    if ( p[1] != '[' )
    {
        *advance = 1;
        return code;
    }

    // Error Check Part 2
    if ( p[2] != '3' )
    {
        *advance = 2;
        return code;
    }

    // Error Check Part 3
    if ( p[3] < '0' || p[3] > '7' )
    {
        *advance = 3;
        return code;
    }

    // Error Check Part 4
    if ( p[4] != 'm' )
    {
        *advance = 4;
        return code;
    }

    code = DBS_BLACK + p[3] - '0';
    *advance = 5;

    return code;
        
}
