#include <nitro.h>
#include <ninet/ip.h>
#include <dwc.h>

#include "userdata.h"
#include "init.h"
#include "dbs.h"

/** --------------------------------------------------------------------
  typedef
  ----------------------------------------------------------------------*/
typedef struct stUDBackupInfo
{
    CARDBackupType type;
    const char *comment;
} DTUDBackupInfo;

typedef struct stUDPlayerInfo
{
    DWCUserData   userData;
    DWCFriendData friendList[FRIEND_LIST_LEN];
} DTUDPlayerInfo;

/** --------------------------------------------------------------------
  statics
  ----------------------------------------------------------------------*/
static DTUDPlayerInfo s_PlayerInfo;
static u16 s_CardLockId;

/** --------------------------------------------------------------------
  Backup is decided in compile time.
  ----------------------------------------------------------------------*/
#if !defined(DTUD_BACKUP_TYPE)
//#define DTUD_BACKUP_TYPE CARD_BACKUP_TYPE_EEPROM_4KBITS
#define DTUD_BACKUP_TYPE CARD_BACKUP_TYPE_EEPROM_64KBITS
//#define DTUD_BACKUP_TYPE CARD_BACKUP_TYPE_EEPROM_512KBITS
//#define DTUD_BACKUP_TYPE CARD_BACKUP_TYPE_FLASH_2MBITS
//#define DTUD_BACKUP_TYPE CARD_BACKUP_TYPE_FRAM_256KBITS
#endif

/*---------------------------------------------------------------------------*
  Name:         DTUDs_CheckBackup

  Description:  Checks the backup status of the card.

  Arguments:    None.

  Returns:      BOOL - If the backup device on the card is in a readable state: TRUE
                       if it's in an unreadable state: returns FALSE.
 *---------------------------------------------------------------------------*/
static BOOL
DTUDs_CheckBackup( void )
{
    BOOL end = FALSE;
    s32 ret;

    ret = OS_GetLockID();
    if (ret == OS_LOCK_ID_ERROR)
    {
        return FALSE;
    }
    s_CardLockId = (u16)ret;

    CARD_LockBackup(s_CardLockId);
    end = CARD_IdentifyBackup( DTUD_BACKUP_TYPE );
    CARD_UnlockBackup(s_CardLockId);

    return end;
}

/*---------------------------------------------------------------------------*
  Name:         DTUD_LoadBackup

  Description:  Loads from the card's backup

  Arguments:    src  - distribution source address
                buf - Transfer destination memory address
                size - Byte length of data being read

  Returns:      BOOL - If data is successfully read from the card: returns TRUE,
                       if failed: returns FALSE.
 *---------------------------------------------------------------------------*/
BOOL
DTUD_LoadBackup( u32 src, void* buf, u32 size )
{
    int last_result;

    CARD_LockBackup(s_CardLockId);
    if (CARD_IsBackupEeprom())
    {
        CARD_ReadEeprom(src, buf, size );
    }
    else if (CARD_IsBackupFlash())
    {
        CARD_ReadFlash(src, buf, size );
    }
    else if (CARD_IsBackupFram())
    {
        CARD_ReadFram(src, buf, size );
    }
    CARD_UnlockBackup(s_CardLockId);

    last_result = CARD_GetResultCode();
    if (last_result != CARD_RESULT_SUCCESS)
    {
        return FALSE;
    }

    return TRUE;
}

/*---------------------------------------------------------------------------*
  Name:         DTUD_SaveBackup

  Description:  Saves to the card's backup

  Arguments:    dst  - distribution addresses
                buf - Transfer source memory address
                size - Byte length of data being read

  Returns:      BOOL - If data write to the card is successful: returns TRUE,
                       if failed: returns FALSE.
 *---------------------------------------------------------------------------*/
BOOL
DTUD_SaveBackup( u32 dst, const void* buf, u32 size )
{
    int last_result;
    
    CARD_LockBackup(s_CardLockId);
    if (CARD_IsBackupEeprom())
    {
        CARD_WriteAndVerifyEeprom(dst, buf, size );
    }
    else if (CARD_IsBackupFlash())
    {
        CARD_WriteAndVerifyFlash(dst, buf, size );
    }
    else if (CARD_IsBackupFram())
    {
        CARD_WriteAndVerifyFram(dst, buf, size );
    }
    CARD_UnlockBackup(s_CardLockId);

    last_result = CARD_GetResultCode();
    if (last_result != CARD_RESULT_SUCCESS)
    {
        return FALSE;
    }

    return TRUE;
}

/*---------------------------------------------------------------------------*
  Name:         DTUD_Init

  Description:  If there is user data on the card, it is used,
                   otherwise, new user data is created.

  Arguments:    NONE.

  Returns:      BOOL - If there is user data on the card: returns TRUE,
                       if not: returns FALSE.
 *---------------------------------------------------------------------------*/
BOOL
DTUD_Init( void )
{

    // If there is backup data and there is user data there,
    // load everything and return TRUE.
    if ( DTUDs_CheckBackup() )
    {
        (void)DTUD_LoadBackup( 0, &s_PlayerInfo, sizeof(DTUDPlayerInfo) );

        OS_TPrintf("Load From Backup\n");

        if ( DWC_CheckUserData( &s_PlayerInfo.userData ) )
        {
            DWC_ReportUserData( &s_PlayerInfo.userData );
            return TRUE;
        }
    }

    // If valid user data had not been saved
    OS_TPrintf("no Backup UserData\n");

    // Create user data.
    DWC_CreateUserData( &s_PlayerInfo.userData, DTUD_INITIAL_CODE );

    OS_TPrintf("Create UserData.\n");
    DWC_ReportUserData( &s_PlayerInfo.userData );

    // Initialize the friend list
    MI_CpuClear32( &s_PlayerInfo.friendList, sizeof( s_PlayerInfo.friendList ) );
    
    return FALSE;
}

/*---------------------------------------------------------------------------*
  Name:         DTUD_CheckUpdateUserData

  Description:  Checks for the Dirty flag in user data.

  Arguments:    NONE.

  Returns:      BOOL - If user data is changed: returns TRUE,
                       if not: returns FALSE.
 *---------------------------------------------------------------------------*/
BOOL
DTUD_CheckUpdateUserData( void )
{
    return DWC_CheckDirtyFlag( &s_PlayerInfo.userData );
}

/*---------------------------------------------------------------------------*
  Name:         DTUD_SaveUserData

  Description:  Saves user data to the card

  Arguments:    NONE.

  Returns:      NONE.
 *---------------------------------------------------------------------------*/
void
DTUD_SaveUserData( void )
{
    if ( DTUDs_CheckBackup() )
    {
        DWC_ClearDirtyFlag( &s_PlayerInfo.userData );
        DTUD_SaveBackup( 0, &s_PlayerInfo.userData, sizeof(DWCUserData) );
    }
}

/*---------------------------------------------------------------------------*
  Name:         DTUD_DeleteUserData

  Description:  Deletes user data from local and the card.

  Arguments:    NONE.

  Returns:      NONE.
 *---------------------------------------------------------------------------*/
void
DTUD_DeleteUserData( void )
{
    if ( DTUDs_CheckBackup() )
    {
        MI_CpuClear32( &s_PlayerInfo.userData, sizeof( DWCUserData ) );
        DTUD_SaveBackup( 0, &s_PlayerInfo.userData, sizeof(DWCUserData) );
    }
}

/*---------------------------------------------------------------------------*
  Name:         DTUD_GetUserData

  Description:  Gets user data.

  Arguments:    NONE.

  Returns:      Returns the user data pointer.
 *---------------------------------------------------------------------------*/
DWCUserData* DTUD_GetUserData( void )
{
    return &s_PlayerInfo.userData;
}

/*---------------------------------------------------------------------------*
  Name:         DTUD_SaveFriendList

  Description:  Saves the friend list to the card

  Arguments:    NONE.

  Returns:      NONE.
 *---------------------------------------------------------------------------*/
void
DTUD_SaveFriendList( void )
{
    if ( DTUDs_CheckBackup() )
    {
        DTUD_SaveBackup( (u32)s_PlayerInfo.friendList - (u32)&s_PlayerInfo,
                         &s_PlayerInfo.friendList, sizeof( s_PlayerInfo.friendList ) );
    }
}

/*---------------------------------------------------------------------------*
  Name:         DTUD_DeleteFriendList

  Description:  Deletes the friend list from local and the card.

  Arguments:    NONE.

  Returns:      NONE.
 *---------------------------------------------------------------------------*/
void
DTUD_DeleteFriendList( void )
{
    if ( DTUDs_CheckBackup() )
    {
        MI_CpuClear32( &s_PlayerInfo.friendList, sizeof( s_PlayerInfo.friendList ) );
        DTUD_SaveBackup( (u32)s_PlayerInfo.friendList - (u32)&s_PlayerInfo,
                         &s_PlayerInfo.friendList, sizeof( s_PlayerInfo.friendList ) );
    }
}

/*---------------------------------------------------------------------------*
  Name:         DTUD_GetFriendList

  Description:  Gets the friend list

  Arguments:    NONE.

  Returns:      Returns the friend list pointer.
 *---------------------------------------------------------------------------*/
DWCFriendData* DTUD_GetFriendList( void )
{
    return s_PlayerInfo.friendList;
}

