#include <nitro.h>
#include <ninet/ip.h>
#include <dwc.h>

#include "main.h"
#include "init.h"
#include "utility.h"
#include "userdata.h"
#include "screen.h"

#include "dbs.h"

/** --------------------------------------------------------------------
  defines
  ----------------------------------------------------------------------*/
#define GAME_NAME        "dwctest" // Game name used in this sample
#define GAME_SECRET_KEY  "d4q9GZ"  // Secret key used in this sample
#define GAME_PRODUCTID   10824      // Product ID used in this sample

/** --------------------------------------------------------------------
  globals
  ----------------------------------------------------------------------*/
KeyControl		g_KeyCtrl;

/** --------------------------------------------------------------------
  statics
  ----------------------------------------------------------------------*/
static u8 s_Work[ DWC_INIT_WORK_SIZE ] ATTRIBUTE_ALIGN( 32 );
static DWCFriendsMatchControl s_FMCtrl;
static DWCInetControl		  s_ConnCtrl;
static char stFriendKey[sizeof(DWCFriendData)]; // Friend key
static int  s_state  = state_init;
static BOOL s_logined;
static BOOL s_canceled;

static int  s_profileID = 0;
static u64  s_friendkey = 0;

/** --------------------------------------------------------------------
  static functions
  ----------------------------------------------------------------------*/
static int update_friend( void );
static int update_init( void );
static int update_connect( void );
static int update_login( void );
static int update_online( void );
static int update_initFriendKey( void );
static int update_changeFriendKey( void );
static int update_initFriendData( void );
static int update_inputFriendIndex( void );
static int update_deleteFriendData( void );
static int update_disconnect( void );
static int update_save( void );
static int update_error( void );

static void cb_login( DWCError error, int profileID, void* param );
static void cb_updateServers(DWCError error, BOOL isChanged, void* param);
static void cb_close( DWCError error, BOOL isLocal, BOOL isServer, u8  aid, int index, void* param );
static int  InputNumber( u8 maxFigure );
static void DispMenuMsg( const u8 figure, const u8 maxFigure );
static void DispMessage( const char* main_msg, const char* msg_a, const char* msg_b, const char* msg_x, const char* msg_y,  const char* msg_l );

/*---------------------------------------------------------------------------*
  V-blank interrupt function
 *---------------------------------------------------------------------------*/
static void VBlankIntr(void)
{
    //---- upload pseudo screen to VRAM
    DC_FlushRange(gScreen, sizeof(gScreen));
    GX_LoadBG0Scr(gScreen, 0, sizeof(gScreen));

    OS_SetIrqCheckFlag(OS_IE_V_BLANK);
}


/** --------------------------------------------------------------------
  Initialization processing
  ----------------------------------------------------------------------*/
static void
initFunctions( void )
{
	OS_Init();
    OS_InitTick();
    OS_InitAlarm();
    RTC_Init();
    CARD_Init();
    OS_SetThreadStackWarningOffset(OS_GetCurrentThread(), 0x100);
    PXI_Init();
    GX_Init();
    FS_Init( MI_DMA_MAX_NUM ); 

    GX_DispOff();
    GXS_DispOff();

    // Enable V-blank interrupt
    OS_SetIrqFunction(OS_IE_V_BLANK, VBlankIntr);
    (void)OS_EnableIrqMask(OS_IE_V_BLANK);
    (void)OS_EnableIrq();
    (void)OS_EnableInterrupts();
    (void)GX_VBlankIntr(TRUE);
	GX_SetDispSelect(GX_DISP_SELECT_SUB_MAIN);

    Heap_Init();
    Heap_SetDebug( TRUE );

    // Initialize output
    InitScreen();
    dbs_DemoInit();
    DTDB_SetPrintOverride( TRUE );

    // Initialize key input
    InitKeyData( &g_KeyCtrl );

    // Starts display
    OS_WaitVBlankIntr();              // Waiting the end of V-Blank interrupt
    GX_DispOn();
    GXS_DispOn();
}

/** --------------------------------------------------------------------
  
  ----------------------------------------------------------------------*/

/** --------------------------------------------------------------------
  function called once per frame
  ----------------------------------------------------------------------*/
static void
update( void )
{
    int errcode;
    DWCError err;
    DWCErrorType errtype;

    switch ( s_state )
    {
    case state_init:			 s_state = update_init();			 break;
    case state_connect:			 s_state = update_connect();		 break;
    case state_login:			 s_state = update_login();			 break;
    case state_online:			 s_state = update_online();			 break;
    case state_initFriendKey:    s_state = update_initFriendKey();   break;
    case state_changeFriendKey:  s_state = update_changeFriendKey(); break;
    case state_initFriendData:   s_state = update_initFriendData();  break;
    case state_inputFriendIndex: s_state = update_inputFriendIndex();break;
    case state_deleteFriendData: s_state = update_deleteFriendData();break;
    case state_disconnect:		 s_state = update_disconnect();		 break;
    case state_save:             s_state = update_save();            break;
    case state_error:			 s_state = update_error();			 break;
    }

    err = DWC_GetLastErrorEx(&errcode, &errtype);

    if ( err != DWC_ERROR_NONE && s_state != state_error )
    {
        // Here, just change the status to "waiting to resolve error," if an error occurs.
        OS_TPrintf( "\n*******************************\n" );
        OS_TPrintf( "ERROR : %d : %d\n", err, errcode );
        OS_TPrintf( "Push START BUTTON to continue\n" );
        OS_TPrintf( "*******************************\n\n" );
        s_state = state_error;
    }
}

/** --------------------------------------------------------------------
  Initial state
  ----------------------------------------------------------------------*/
static int
update_init( void )
{
    // Does the user data on the DS device coincide with that on the DS card?
    if ( DWC_CheckValidConsole( DTUD_GetUserData() ) )
    {
        DispMessage( "STATE : CONNECTING", NULL, NULL, NULL, NULL, NULL );

        DWC_InitInet( &s_ConnCtrl );
        
        DWC_SetAuthServer( DWC_CONNECTINET_AUTH_RELEASE );
        
        DWC_ConnectInetAsync();

        // Gets the friend registration key.
        s_friendkey = DWC_CreateFriendKey( DTUD_GetUserData() );
        
        return state_connect;
    }
    else
    {
        // IF it does not coincide, a connection to the Internet cannot be made.
        OS_TPrintf( "UserData is not Valid Console.\n" );
        return state_init;
    }

    return state_init;
}

/** --------------------------------------------------------------------
  connecting
  ----------------------------------------------------------------------*/
static int
update_connect( void )
{
    DWC_ProcessInet();

    if ( DWC_CheckInet() )
    {
        int status;

        status = DWC_GetLastError(NULL);

        if ( status == DWC_ERROR_NONE )
        {
            OS_TPrintf("Connected!..start FriendsMatch.\n");

            s_logined = FALSE;

            DWC_InitFriendsMatch(&s_FMCtrl, DTUD_GetUserData(),
                                 GAME_PRODUCTID, GAME_NAME, GAME_SECRET_KEY,
                                 0, 0,
                                 DTUD_GetFriendList(), FRIEND_LIST_LEN);

            // Login using the authentication function
            DWC_LoginAsync(L"Name", NULL, cb_login, NULL);

            return state_login;
        }
        else
        {
            OS_TPrintf("Error\n");
            return state_disconnect;
        }
    }

    return state_connect;
}

/** --------------------------------------------------------------------
  During login
  ----------------------------------------------------------------------*/
int
update_login( void )
{
    DWC_ProcessFriendsMatch();  // Update DWC communication processing

    if ( s_logined )
    {
        // Login successful
        DispMessage( "STATE : ONLINE", "save data", "shutdown", "change friend key", "delete friend data", "friend state" );

        // Gets the result of the ingamesn check.
	    if( DWC_GetIngamesnCheckResult() == DWC_INGAMESN_INVALID )
        {
            // If an inappropriate name is determined, a special operation is necessary.
		    OS_TPrintf("BAD ingamesn is detected by NAS.\n");
        }

        // Gets the friend registration key.
        s_friendkey = DWC_CreateFriendKey( DTUD_GetUserData() );

        s_canceled = FALSE;

        return state_online;
    }

    return state_login;
}

/** --------------------------------------------------------------------
  Online
  ----------------------------------------------------------------------*/
int
update_online( void )
{
    DWC_ProcessFriendsMatch();  // Update DWC communication processing

    if ( g_KeyCtrl.trg & PAD_BUTTON_A )
    {
        DispMessage( "SAVE DATA?", "save", "cancel", NULL, NULL, NULL );
        return state_save;
    }
    else if ( g_KeyCtrl.trg & PAD_BUTTON_B )
    {
        DWC_ShutdownFriendsMatch();
        return state_disconnect;
    }
    else if ( g_KeyCtrl.trg & PAD_BUTTON_X )
    {
        DispMessage( "CHANGE FRIEND KEY?", "yes", "no", NULL, NULL, NULL );
        return state_initFriendKey;
    }
    else if ( g_KeyCtrl.trg & PAD_BUTTON_Y )
    { 
        DispMessage( "DELETE FRIEND DATA?", "yes", "no", NULL, NULL, NULL );
        return state_initFriendData;
    }
    else if ( g_KeyCtrl.trg & PAD_BUTTON_L )
    {
        DTUD_DispFriendList();
        return state_online;
    }

    return state_online;
}

/*---------------------------------------------------------------------------*
  Name:         update_initFriendKey

  Description:  performs friends registration key initialization

  Arguments:    NONE.

  Returns:      Returns state_online if the friend list is full and registration is not possible or registration is cancelled.
                Returns state_change if exchange registration starts.
 *---------------------------------------------------------------------------*/
static int
update_initFriendKey( void )
{
    int  friendIdx = -1;

    DWC_ProcessFriendsMatch();  // Update DWC communication processing

    if ( g_KeyCtrl.trg & PAD_BUTTON_A )
    {
        // Clears the friend registration key to '0'
        MI_CpuFill8( stFriendKey, '0', sizeof(stFriendKey));

        //Check empty space on the friend list
        friendIdx = DTUD_GetAvailableFriendListIndex();
        if ( friendIdx == -1 )
        {
            OS_TPrintf( "Failed to register new friend. Friend list is full.\n" );
            DispMessage( "STATE : ONLINE", "save data", "shutdown", "change friend key", "delete friend data", "friend state" );
            return state_online;
        }
        else
        {
            return state_changeFriendKey;
        }
    }
    else if( g_KeyCtrl.trg & PAD_BUTTON_B )
    {
        // If canceled, it returns to online status
        DispMessage( "STATE : ONLINE", "save data", "shutdown", "change friend key", "delete friend data", "friend state" );
        return state_online;
    }
    return state_initFriendKey;
}

/*---------------------------------------------------------------------------*
  Name:         update_changeFriendKey

  Description:  Runs friend registration key input and friend registration.

  Arguments:    NONE.

  Returns:      Returns state_init if registration is complete.
                Returns state_changeFriendKey if it has not finished.
 *---------------------------------------------------------------------------*/
static int
update_changeFriendKey( void )
{
    DWCUserData*   s_UserData;
    DWCFriendData* s_FriendList;
    int  friendIdx = -1;
    u8   maxFigure = 12;// Maximum 12 digits for friend registration key

    DWC_ProcessFriendsMatch();  // Update DWC communication processing
    
    if ( g_KeyCtrl.trg & PAD_BUTTON_B )
    {
        // Pressing the B button cancels input
        DispMessage( "STATE : ONLINE", "save data", "shutdown", "change friend key", "delete friend data", "friend state" );
        return state_online;        
    }
    // friend registration key input
    else if( InputNumber( maxFigure ) )
    {
        // Validity check of friend registration key
        s_UserData = DTUD_GetUserData();
        if( DWC_CheckFriendKey( s_UserData, DWC_Acc_NumericKeyToFriendKey( stFriendKey )))
        {
            // Register friend using friend registration key
            friendIdx = DTUD_GetAvailableFriendListIndex();
            s_FriendList = DTUD_GetFriendList();
            DWC_CreateFriendKeyToken( &s_FriendList[friendIdx], DWC_Acc_NumericKeyToFriendKey( stFriendKey ) );
        }
        else
        {
            OS_TPrintf( "Invalid Friend Code!\n" );
        }
        DTUD_DispFriendList();
        DispMessage( "STATE : ONLINE", "save data", "shutdown", "change friend key", "delete friend data", "friend state" );
        return state_online;
    }
    
    return state_changeFriendKey;
}

/*---------------------------------------------------------------------------*
  Name:         update_initFriendData

  Description:  performs initialization to delete friends

  Arguments:    NONE.

  Returns:      Returns state_online if deletion completes or is cancelled.
                Returns state_deleteFriendData if performing exchange registration.
 *---------------------------------------------------------------------------*/
static int
update_initFriendData( void )
{
    DWC_ProcessFriendsMatch();  // Update DWC communication processing

    if ( g_KeyCtrl.trg & PAD_BUTTON_A )
    {
        DTUD_DispFriendList();
        
        // Clears the friend registration key to '0'
        MI_CpuFill8( stFriendKey, '0', sizeof(stFriendKey));

        return state_inputFriendIndex;
    }
    else if( g_KeyCtrl.trg & PAD_BUTTON_B )
    {
        DispMessage( "STATE : ONLINE", "save data", "shutdown", "change friend key", "delete friend data", "friend state" );
        return state_online;
    }

    return state_initFriendData;
    
}

/*---------------------------------------------------------------------------*
  Name:         update_inputFriendIndex

  Description:  Runs numerical value input for the friend list index.

  Arguments:    NONE.

  Returns:      BOOL - If deletion is completed or canceled: returns TRUE
                       Returns FALSE if it has not finished.
 *---------------------------------------------------------------------------*/
static int
update_inputFriendIndex( void )
{
    u8 maxFigure = 2;
    u8 i;

    DWC_ProcessFriendsMatch();  // Update DWC communication processing

    if ( g_KeyCtrl.trg & PAD_BUTTON_B )
    {
        // Pressing the B button cancels input
        DispMessage( "STATE : ONLINE", "save data", "shutdown", "change friend key", "delete friend data", "friend state" );
        return state_online;        
    }
    else if( InputNumber( maxFigure ) )
    {
        // Checks whether the friends list can be edited
        if( DWC_CanChangeFriendList() )
        {
            DispMessage( "Is Data deleted, Really?", "yes", "no", NULL, NULL, NULL );
            PrintString( 1, 15, 0xf, "Set friend index:");
            for(i = 0; i < maxFigure; i++){
                PrintString( (s16)(18+i), 15, 0xf, "%c", stFriendKey[i] );
            }
            return state_deleteFriendData;
        }
        else
        {
            OS_TPrintf( "Invalid Friend Index!\n" );
            DTUD_DispFriendList();
            DispMessage( "STATE : ONLINE", "save data", "shutdown", "change friend key", "delete friend data", "friend state" );
            return state_online;
        }
    }
    return state_inputFriendIndex;
}

/*---------------------------------------------------------------------------*
  Name:         update_deleteFriendData

  Description:  Runs friend delete for the friend list index.

  Arguments:    NONE.

  Returns:      Returns state_online when the deletion is complete or cancelled.
                Returns state_changeFriendData if it has not finished.
 *---------------------------------------------------------------------------*/
static int
update_deleteFriendData( void )
{
    DWCFriendData* s_FriendList;
    int friendIdx;
    u8 maxFigure = 2;

    DWC_ProcessFriendsMatch();  // Update DWC communication processing
        
    if ( g_KeyCtrl.trg & PAD_BUTTON_A )
    {
        // Converts input data to int format.
        friendIdx = (int)((stFriendKey[0]-48)*10+(stFriendKey[1]-48));

        // Delete friend from friend roster
        s_FriendList = DTUD_GetFriendList();
        DWC_DeleteBuddyFriendData(&s_FriendList[friendIdx]);

        DTUD_DispFriendList();
        DispMessage( "STATE : ONLINE", "save data", "shutdown", "change friend key", "delete friend data", "friend state" );
        return state_online;
    }
    else if( g_KeyCtrl.trg & PAD_BUTTON_B )
    {
        DispMessage( "STATE : ONLINE", "save data", "shutdown", "change friend key", "delete friend data", "friend state" );
        return state_online;
    }
    return state_deleteFriendData;
}

/** --------------------------------------------------------------------
  Disconnects
  ----------------------------------------------------------------------*/
static int
update_disconnect( void )
{
    if ( DWC_CleanupInetAsync() )
    {
        return state_init;
    }

    return state_disconnect;
}

/** --------------------------------------------------------------------
  Saves data
  ----------------------------------------------------------------------*/
static int
update_save( void )
{
    DWC_ProcessFriendsMatch();  // Update DWC communication processing
    
    if ( g_KeyCtrl.trg & PAD_BUTTON_A )
    {
        if( DWC_CheckDirtyFlag( DTUD_GetUserData() ) ){
            DTUD_SaveUserData();
            OS_TPrintf( "Saved User Data.\n" );
        }
        
        DTUD_SaveFriendList();
        OS_TPrintf( "Saved Friend Data.\n" );

        DispMessage( "STATE : ONLINE", "save data", "shutdown", "change friend key", "delete friend data", "friend state" );
        return state_online;
    }
    else if ( g_KeyCtrl.trg & PAD_BUTTON_B )
    {
        DispMessage( "STATE : ONLINE", "save data", "shutdown", "change friend key", "delete friend data", "friend state" );
        return state_online;
    }
    return state_save;
}

/** --------------------------------------------------------------------
  Waiting for error release
  ----------------------------------------------------------------------*/
static int update_error( void )
{
    int errcode;
    DWCError err;
    DWCErrorType errtype;
    
    if ( g_KeyCtrl.trg & PAD_BUTTON_START )
    {
        // Releases errors with the start button
        err = DWC_GetLastErrorEx(&errcode, &errtype);

        DWC_ClearError();
        
        if ( errtype == DWC_ETYPE_DISCONNECT )
        {
            // When disconnecting, run the "FriendsMatch" end operation and Internet connection Cleanup.
            // Runs Clearup
            DWC_ShutdownFriendsMatch();
            return state_disconnect;
        }
        else if ( errtype == DWC_ETYPE_FATAL )
        {
            // This is a Fatal Error, so end
            OS_Panic("++FATAL ERROR\n");
            return state_init;  // not reached;
        }
        else if ( errtype == DWC_ETYPE_LIGHT )
        {
            // For small errors, don't do anything and go back to online status.
            return state_online;
        }
        else {
            // Once only the FriendsMatch operation is completed, run initialization of FriendsMatch again.
            DWC_ShutdownFriendsMatch();
            return state_connect;
        }
    }

    return state_error;
}

/*---------------------------------------------------------------------------*
  Name:         InputNumber

  Description:  performs friends registration key numerical value input

  Arguments:    maxFigure - Maximum number of input digits

  Returns:      If input is complete: returns TRUE,
                if input is incomplete: returns FALSE.
 *---------------------------------------------------------------------------*/
static BOOL InputNumber( u8 maxFigure )
{
    static u8 figure = 0;

    if(figure > maxFigure){
        figure = 0;
    }
    
    DispMenuMsg(figure, maxFigure);

    if (g_KeyCtrl.trg & PAD_BUTTON_A)
    {
        figure = 0;
        return TRUE;
    }
    else if (g_KeyCtrl.trg & PAD_KEY_UP)
    {
        // Increment numeric value during selection using the up/down buttons on the crosskey
        stFriendKey[figure]++;
        if(stFriendKey[figure] > '9')
        {
            stFriendKey[figure] = '0';
        }
        // Displays numerical values during input
        DispMenuMsg(figure ,maxFigure);
    }
    else if (g_KeyCtrl.trg & PAD_KEY_DOWN)
    {
        // Increment numeric value during selection using the up/down buttons on the crosskey
        stFriendKey[figure]--;
        if(stFriendKey[figure] < '0')
        {
            stFriendKey[figure] = '9';
        }
        // Displays numerical values during input
        DispMenuMsg(figure, maxFigure);
    }
    else if (g_KeyCtrl.trg & PAD_KEY_LEFT)
    {
        // Move the number of selected digits using the left/right buttons on the crosskey
        if (figure > 0 && figure <= maxFigure-1 ) figure--;
        // Displays numerical values during input
        DispMenuMsg(figure, maxFigure);
    }
    else if (g_KeyCtrl.trg & PAD_KEY_RIGHT)
    {
        // Move the number of selected digits using the left/right buttons on the crosskey
        if (figure >= 0 && figure < maxFigure-1 ) figure++;
        // Displays numerical values during input
        DispMenuMsg(figure, maxFigure);
    }
    return FALSE;
}

/*---------------------------------------------------------------------------*
  Name:         DispMenuMsg

  Description:  Displays the numerical value during friend registration

  Arguments:    figure    - number of digits during input
                maxFigure - maximum number of digits

  Returns:      NONE.
 *---------------------------------------------------------------------------*/
static void DispMenuMsg( const u8 figure, const u8 maxFigure )
{
    int i;

    int x = 16;

    DispMessage( "NUMBER INPUT", "set data", "cancel", NULL, NULL, NULL );
    
    // Currently entering a numeric value
    // Register friend friend registration key being input
    if( maxFigure == 12 )
    {
        PrintString( 1, 15, 0xf, "Set friend key:");
        for(i = 0; i < maxFigure; i++){
            PrintString( (s16)(x+i), 15, 0xf, "%c", stFriendKey[i] );
            if(i%4 == 3 && i != maxFigure - 1)
            {
                x++;
                PrintString( (s16)(x+i), 15, 0xf, "-" );
            }
        }
        // Display an underline to indicate the numeric value currently being selected
        PrintString( (s16)(16+figure+figure/4), 16, 0xf, "-" );
    }
    else if( maxFigure == 2 )
    {
        x += 2;
        PrintString( 1, 15, 0xf, "Set friend index:");
        for(i = 0; i < maxFigure; i++){
            PrintString( (s16)(x+i), 15, 0xf, "%c", stFriendKey[i] );
        }
        // Display an underline to indicate the numeric value currently being selected
        PrintString( (s16)(18+figure), 16, 0xf, "-" );
    }
}

/** --------------------------------------------------------------------
  ----------------------------------------------------------------------*/
/** --------------------------------------------------------------------
  Callback function
  ----------------------------------------------------------------------*/
static void
cb_login(DWCError error, int profileID, void* param)
{
    (void)param;
    if (error == DWC_ERROR_NONE)
    {
        // Get authentication success status and profile ID
        OS_TPrintf("Login succeeded. profileID = %u\n\n", profileID);

        s_profileID = profileID;  // Records the profile ID.

        // Check whether or not stUserData has been updated.
        if ( DTUD_CheckUpdateUserData() )
        {
            DTUD_SaveUserData();
        }

        // Synchronous processing of the friend register and the buddy list on the GameSpy server
        DWC_UpdateServersAsync(NULL,
                               cb_updateServers, NULL,
                               NULL, NULL,
                               NULL, NULL);
        
        // Set the callback for closing connections
        DWC_SetConnectionClosedCallback(cb_close, NULL);
    }
    else
    {
        OS_TPrintf("Login Error\n");
    }
}

static void
cb_updateServers(DWCError error, BOOL isChanged, void* param)
{
    (void)param;
    (void)isChanged;

    if ( error == DWC_ERROR_NONE )
    {
        // Successful synchronous processing of friend roster.
        OS_TPrintf( "Update Servers succeeded.\n" );
        
        s_logined = TRUE;
    }
    else
    {
        OS_TPrintf("Error\n");
    }
}

static void
cb_close(DWCError error, BOOL isLocal, BOOL isServer, u8  aid, int index, void* param)
{
    (void)isLocal;
    (void)isServer;
    (void)param;

    OS_TPrintf( "Closed Callback err:%d, aid:%d, idx:%d\n",
               error, aid, index );
}

/** -------------------------------------------------------------------------
  ---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*
  Name:         DispMessage

  Description:  Displays general messages

  Arguments:    main_msg - current mode
                msg_a    - operation when the A button is pressed
                msg_b    - operation when the B button is pressed
                msg_x    - operation when the X button is pressed
                msg_y    - operation when the Y button is pressed
                msg_l    - operation when the L button is pressed

  Returns:      NONE.
 *---------------------------------------------------------------------------*/
static void
DispMessage( const char* main_msg, const char* msg_a, const char* msg_b, const char* msg_x, const char* msg_y,  const char* msg_l )
{
    s16 x = 1;
    s16 y = 5;

    ClearScreen();
    
    PrintString( 0, 0, 0xf, "================================" );
    PrintString( 2, 1, 0xf, "*** FRIEND DATA SAMPLE ***" );
    PrintString( 0, 2, 0xf, "================================" );
    
    if ( main_msg == NULL ) return;

    PrintString( 0, y, 0xf, "--------------------------------" );
    y++;
    // Displays a message
    while ( *main_msg != '\0' )
    {
        if ( *main_msg == '\n' )
        {
            x = 1;  y++;
        }
        else
        {
            PrintString( x, y, 0xf, "%c", *main_msg );
            x++;
        }
        main_msg++;
    }
    y++;
    PrintString( 0, y, 0xf, "--------------------------------" );

    y += 2;

    // Displays an explanation of the A button and explanations of the B, X, and Y buttons
    if ( msg_a != NULL )
    {
        PrintString( 1, y, 0xf, "A BUTTON:%s", msg_a );
    }
    y++;
    if ( msg_b != NULL )
    {
        PrintString( 1, y, 0xf, "B BUTTON:%s", msg_b );
    }
    y += 2;
    if ( msg_x != NULL )
    {
        PrintString( 1, y, 0xf, "X BUTTON:%s", msg_x );
    }
    y++;
    if ( msg_y != NULL )
    {
        PrintString( 1, y, 0xf, "Y BUTTON:%s", msg_y );
    }
    y += 2;
    if ( msg_l != NULL )
    {
        PrintString( 1, y, 0xf, "L BUTTON:%s", msg_l );
    }
}

/** --------------------------------------------------------------------
  Status display function called every frame.
  ----------------------------------------------------------------------*/
void
DrawConsole( void )
{
    static int i = 0;
    char loop[] = "/|\\-";

    dbs_Print( 0, 0, "s:%d", DWC_GetMatchingState() );
    dbs_Print( 13, 0, "n:%d", DWC_GetNumConnectionHost() );
    dbs_Print( 0, 1, "w:%d", DWC_GetLinkLevel() );
    dbs_Print( 13, 1, "p:%d", s_profileID );
    dbs_Print( 30, 0, "%c", loop[((i)>>2)&3] );
    dbs_Print( 0, 2, "friendkey:%012lld", s_friendkey );

    i++;
}

/** --------------------------------------------------------------------
  Entrypoint
  ----------------------------------------------------------------------*/
void NitroMain ()
{
    initFunctions();

    // Specify debug display level
    DWC_SetReportLevel((u32)( DWC_REPORTFLAG_ALL &
                              ~DWC_REPORTFLAG_QR2_REQ ));
    
    // Initialize DWC library
    DWC_Init( s_Work );

    // Reads user data
    DTUD_Init();
    // Set function for allocating memory
    DWC_SetMemFunc( DTDB_GetAllocFunc(), DTDB_GetFreeFunc() );

    while (1)
    {
        DrawConsole();

        ReadKeyData();

        update();

        dbs_DemoUpdate();
        OS_WaitIrq(TRUE, OS_IE_V_BLANK);
        Heap_Debug();
        dbs_DemoLoad();
    }

    OS_Terminate();
}
