#include <nitro.h>
#include <ninet/ip.h>
#include <dwc.h>

#include "main.h"
#include "init.h"
#include "utility.h"

#include "dbs.h"


//----------------------------------------------------------------------------
// define
//----------------------------------------------------------------------------
#define GAME_NAME        "dwctest" // Game name used in this sample
#define GAME_SECRET_KEY  "d4q9GZ"  // Secret key used in this sample
#define GAME_PRODUCTID   10824      // Product ID used in this sample

#define GAME_FRAME       1         // Game frames to be anticipated (1 is taken as 1/60)

#define GAME_NUM_MATCH_KEYS 3      // Number of additional keys for matchmaking

#define NETCONFIG_USE_HEAP 1

// DWC_MATCH_OPTION_SC_CONNECT_BLOCK}b`CNIvVXCb`
//#define GAME_MATCH_OP_SC_CONNECT_BLOCK

// switch for testing saving and loading to and from the storage server
//#define GAME_USE_STORAGE_SERVER

#ifdef GAME_MATCH_OP_SC_CONNECT_BLOCK
// The time in seconds after completing a single case of matchmaking until the next matchmaking
#define GAME_MATCH_OP_SC_BLOCK_TIME 10
#endif

// When changing the authentication server from the default development-oriented one to a product-oriented one, be sure to enable the following definition:
//  
//#define USE_AUTHSERVER_RELEASE 1

//----------------------------------------------------------------------------
// typedef
//----------------------------------------------------------------------------
typedef struct tagRegFriendParam
{
    int step;    // 0: Currently selecting a command, 1: Currently entering friend information to be added or deleted
    u32 figure;  // Digits of the profile ID being entered
    u32 value;   // Value being entered
} RegFriendParam;


//----------------------------------------------------------------------------
// variable
//----------------------------------------------------------------------------
static GameControl stGameCnt;  // Game control data structure
static KeyControl stKeyCnt;    // Key input control structure
static DWCFriendsMatchControl stDwcCnt;    // FriendsMatch control structure

static int stNumEntry  = 2;    // Specify the number of players making up the network
static int stServerIdx = 0;    // Friend list index for the server DS to be connected to

static BOOL stCloseFlag = FALSE;    // TRUE if closed by the local host

// Additional key data structure for peer matchmaking without a friend specification
static GameMatchExtKeys stMatchKeys[GAME_NUM_MATCH_KEYS] = { 0, };

#ifdef GAME_MATCH_OP_SC_CONNECT_BLOCK
static OSTick stMatchFinishTime = 0;  // the time matchmaking ends during server-client matchmaking
#endif

#ifdef USE_PRIVATE_USERDATA
// If make PR_USERDATA=TRUE, use fixed user data for debugging
// Values are set inside @private_settings.h
// friend data for sample MP exchange
static DWCFriendData stSampleFriendData;
// sample friend registration key
static char stSampleFriendKey[sizeof(DWCFriendData)];
#endif

// Structure storing user data.
static DWCUserData stUserData;
static DWCInetControl stConnCtrl;

// Overwrite debug output
static BOOL sPrintOverride;

static u8 stNumBackup;
static u16 card_lock_id;

static u32 loopcnt = 0;
static char loopstr[] = "/|\\-";

//----------------------------------------------------------------------------
// prototype
//----------------------------------------------------------------------------
static void InitGameData(void);
static void SetGameMatchKeys(void);

static GameMode GameMain(void);
static GameMode GameRegisterFriendMain(void);
static GameMode GameLogonMain(void);
static GameMode GameMatchMain(void);
static GameMode GameConnectedMain(void);
static GameMode NetConfigMain(void);
static GameMode StartIPMain(void);
static GameMode StopIPMain(void);
static GameMode BackupUserdataMain(void);

static int GetAvailableFriendListIndex(void);

static void LoginCallback(DWCError error, int profileID, void* param);

static void UpdateServersCallback(DWCError error, BOOL isChanged, void* param);
static void FriendStatusCallback(int index, u8 status, const char* statusString, void* param);
static void DeleteFriendListCallback(int deletedIndex, int srcIndex, void* param);
static void BuddyFriendCallback(int index, void* param);

static void LoginToStorageCallback(DWCError error, void* param);
static void SaveToServerCallback(BOOL success, BOOL isPublic, void* param);
static void LoadFromServerCallback(BOOL success, int index, char* data, int len, void* param);

static void ConnectToAnybodyCallback(DWCError error, BOOL cancel, void* param);
static void ConnectToFriendsCallback(DWCError error, BOOL cancel, void* param);
static void SetupGameServerCallback(DWCError error,
                                    BOOL cancel,
                                    BOOL self,
                                    BOOL isServer,
                                    int index,
                                    void* param);
static void ConnectToGameServerCallback(DWCError error,
                                        BOOL cancel,
                                        BOOL self,
                                        BOOL isServer,
                                        int index,
                                        void* param);
static void ConnectionClosedCallback(DWCError error,
                                     BOOL isLocal,
                                     BOOL isServer,
                                     u8 aid,
                                     int index,
                                     void* param);
static void NewClientCallback(int index, void* param);
static int  EvaluateAnybodyCallback(int index, void* param);
static int  EvaluateFriendCallback(int index, void* param);
#ifdef DWC_STOP_SC_SERVER
static void StopSCMatchingCallback(void* param);
#endif

static DWCError HandleDWCError(GameMode* gameMode);

static void DispErrorMessage(DWCError error, int errorCode, DWCErrorType errorType);
static void DispMenuMsgWithCursor(const GameSequence* gameSeq, int index, void* param);
static void DispFriendList(BOOL overLoad);
static void DispFriendListData(void);

static void RegFriendModeDispCallback(int curIdx, void* param);
static void LogonModeDispCallback(int curIdx, void* param);
static void MatchModeDispCallback(int curIdx, void* param);
static void BackupModeDispCallback(int curIdx, void* param);

static void GameWaitVBlankIntr(void);
static void VBlankIntr(void);


//----------------------------------------------------------------------------
// initialized variable
//----------------------------------------------------------------------------
// Game sequence list
static GameSequence gameSeqList[GAME_MODE_NUM] =
{
    {
        "GAME MAIN MODE",
        GameMain,
        NULL,
        6,
        {
            "Login", GAME_MODE_LOGIN, TRUE,
            "Register friend", GAME_MODE_REG_FRIEND, FALSE,
            "NetConfig", GAME_MODE_NETCONFIG, FALSE,
            "StartIP", GAME_MODE_STARTIP, FALSE,
            "StopIP", GAME_MODE_STOPIP, FALSE,
            "UserData Backup", GAME_MODE_BACKUP, FALSE,
        },
    },
    {
        "REGISTER FRIEND MODE",
        GameRegisterFriendMain,
        RegFriendModeDispCallback,
        3,
        {
            "Set", GAME_MODE_NUM, FALSE,
            "Delete", GAME_MODE_NUM, FALSE,
            "Return", GAME_MODE_MAIN, FALSE,
        },
    },
    {
        "GAME LOGON MODE",
        GameLogonMain,
        LogonModeDispCallback,
        5,
        {
            "Connect to anybody", GAME_MODE_CONNECTED, TRUE,
            "Connect to friends", GAME_MODE_CONNECTED, TRUE,
            "Setup game server", GAME_MODE_CONNECTED, TRUE,
            "Connect to game server", GAME_MODE_CONNECTED, TRUE,
            "Logout", GAME_MODE_MAIN, FALSE
        },
    },
    {
        "GAME CONNECTED MODE",
        GameConnectedMain,
        NULL,
        3,
        {
            "DWC_SendReliable", GAME_MODE_NUM, FALSE,
            "DWC_SendUnreliable", GAME_MODE_NUM, FALSE,            
            "Close connections", GAME_MODE_LOGIN, FALSE
        },
    },
    {
        "NET CONFIG GUI",
        NetConfigMain,        NULL,        0,
    },
    {
        "START TO CONNECT INTERNET",
        StartIPMain,        NULL,        0,
    },
    {
        "STOP TO CONNECT INTERNET",
        StopIPMain,        NULL,        0,
    },
    {
        "BACKUP USERDATA",
        BackupUserdataMain,
        BackupModeDispCallback,
        4,
        {
            "Load UserData", GAME_MODE_NUM, FALSE,
            "Save UserData", GAME_MODE_NUM, FALSE,
            "Report UserData", GAME_MODE_NUM, FALSE,
            "return", GAME_MODE_MAIN, FALSE,
        }
    },
};


// Since it's impossible to save in this sample, this is used in place of saved data
static GameSaveData saveData =
{
    {
        "Player1", "Player2", "Player3", "Player4"
    },
    0, FALSE,
    {
        0,
    }
};

// Sample data set for use as GameSpy status data
static const char gsStatusSample[16] = { 10, 9, 8, 0, 0, 1, 2, 3, };

static const struct
{
    CARDBackupType type;
    const char *comment;
} stBackupTypeTable[] = {
    { CARD_BACKUP_TYPE_EEPROM_4KBITS,   "EEPROM    4 kb" },
    { CARD_BACKUP_TYPE_EEPROM_64KBITS,  "EEPROM   64 kb" },
    { CARD_BACKUP_TYPE_EEPROM_512KBITS, "EEPROM  512 kb" },
    { CARD_BACKUP_TYPE_FLASH_2MBITS,    "FLASH     2 Mb" },
    { CARD_BACKUP_TYPE_FRAM_256KBITS,   "FRAM    256 kb" },
} ;
enum { BACKUP_TABLE_MAX = sizeof(stBackupTypeTable) / sizeof(*stBackupTypeTable) } ;


/*---------------------------------------------------------------------------*
  Function for allocating memory
 *---------------------------------------------------------------------------*/
static void*
AllocFunc( DWCAllocType name, u32   size, int align )
{
    void * ptr;
    OSIntrMode old;
    (void)name;
    (void)align;

    old = OS_DisableInterrupts();

    ptr = OS_AllocFromMain( size );

    OS_RestoreInterrupts( old );

#if 0
    OS_TPrintf("##ALLOC:%d:%p\n",size, ptr);

    if ( (u32) ptr == 0x0210f160 )
    {
        OS_TPrintf("#########################leak!\n");
    }
#endif

    return ptr;
}

/*---------------------------------------------------------------------------*
  Function for freeing memory
 *---------------------------------------------------------------------------*/
static void
FreeFunc( DWCAllocType name, void* ptr,  u32 size  )
{
    OSIntrMode old;
    (void)name;
    (void)size;

    if ( !ptr ) return;

    old = OS_DisableInterrupts();

    OS_FreeToMain( ptr );

    OS_RestoreInterrupts( old );
}

/** --------------------------------------------------------------------
  For console output using OS_TPrintf()
  ----------------------------------------------------------------------*/
#ifdef SDK_FINALROM
#undef OS_TVPrintf
#undef OS_TPrintf
#endif
extern void OS_TVPrintf(const char *fmt, va_list vlist );
static char common_buffer[256];        // thread unsafe, but less use of stack

void OS_TVPrintf(const char *fmt, va_list vlist )
{
    (void)OS_VSNPrintf(common_buffer, sizeof(common_buffer), fmt, vlist);
#ifndef SDK_FINALROM
    OS_PutString(common_buffer);
#endif

    if ( sPrintOverride )
    {
        dbs_CVPrintf( NULL, fmt, vlist );
    }
}
#ifdef SDK_FINALROM
extern void OS_TPrintf(const char *fmt, ...);
void OS_TPrintf(const char *fmt, ...)
{
    va_list vlist;

    va_start(vlist, fmt);
    OS_TVPrintf(fmt, vlist);
    va_end(vlist);
}
#endif

#if defined( MULTI_LOGINID_DEFINE )
#define FIX_PLAYERID 10
static u64 pseudo_userid;
extern void clearUserData( void );
void clearUserData( void )
{
    DWC_CreateUserData( &stUserData, 'ABCD'  );
    DWCi_Acc_SetUserId( &stUserData.pseudo, pseudo_userid );
    DWCi_Acc_SetPlayerId( &stUserData.pseudo, FIX_PLAYERID );
}
#endif

/*---------------------------------------------------------------------------*
  Main routine
 *---------------------------------------------------------------------------*/
static u8 s_Work[ DWC_INIT_WORK_SIZE ] ATTRIBUTE_ALIGN( 32 );
void NitroMain ()
{
    int friendIdx = 0;
    int ret;

	OS_Init();
    OS_InitTick();
    OS_InitAlarm();
    RTC_Init();
    CARD_Init();
    ret = OS_GetLockID();
    if (ret == OS_LOCK_ID_ERROR)
    {
        OS_Panic("demo fatal error! OS_GetLockID() failed");
    }
    card_lock_id = (u16)ret;

    // Initial settings for checking for stack overflow
    OS_SetThreadStackWarningOffset(OS_GetCurrentThread(), 0x100);
    
    PXI_Init();
    
    Heap_Init();
	Alarm_Init();
//    Heap_SetDebug( TRUE );

    GX_Init();
    FS_Init( MI_DMA_MAX_NUM ); 

    GX_DispOff();
    GXS_DispOff();

    // Enable V-blank interrupt
    OS_SetIrqFunction(OS_IE_V_BLANK, VBlankIntr);
    (void)OS_EnableIrqMask(OS_IE_V_BLANK);
    (void)OS_EnableIrq();
    (void)OS_EnableInterrupts();
    (void)GX_VBlankIntr(TRUE);
	GX_SetDispSelect(GX_DISP_SELECT_SUB_MAIN);

    // Initialize output
    dbs_DemoInit();
    sPrintOverride = TRUE; // OS_TPrintf() output is linked to the console

    // Starts display
    OS_WaitVBlankIntr();              // Waiting the end of V-Blank interrupt
    GX_DispOn();
    GXS_DispOn();

    // Initialize key input
    InitKeyData(&stKeyCnt);

    InitGameData();

    // Delete console data for debugging.
    //DWC_Debug_DWCInitError( s_Work, DWC_INIT_RESULT_DESTROY_OTHER_SETTING );

    // Specify debug display level
    DWC_SetReportLevel((unsigned long)(DWC_REPORTFLAG_ALL & ~DWC_REPORTFLAG_QR2_REQ));
	gsSetDebugLevel(GSIDebugCat_All, GSIDebugType_All, GSIDebugLevel_Verbose);

    // Initialize DWC library
    ret = DWC_Init( s_Work );
    OS_TPrintf( "DWC_Init() result = %d\n", ret );

    if ( ret == DWC_INIT_RESULT_DESTROY_OTHER_SETTING )
    {
        OS_TPrintf( "Wi-Fi setting might be broken.\n" );
    }

    // Create user data
    DWC_CreateUserData( &stUserData, 'ABCD' );
#if defined( MULTI_LOGINID_DEFINE )
    pseudo_userid = DWCi_Acc_GetUserId( &stUserData.pseudo );
    DWCi_Acc_SetPlayerId( &stUserData.pseudo, FIX_PLAYERID );
    OS_Printf("***** p:uid: %lld\n", pseudo_userid);
#endif

#if 0
    OS_TPrintf("+++VALID CONSOLE %d +++", DWC_CheckValidConsole( &stUserData ) );
    {
        char test[16];
        sprintf(test,"%u",0xa1546344);

        OS_PutString(test);
    }
#endif

#ifdef USE_PRIVATE_USERDATA
// If make PR_USERDATA=TRUE, use fixed user data for debugging
#include "@private_settings.h"
#endif

    // Display user data
    DWC_ReportUserData( &stUserData );

    // Heap usage indicator display ON
    //Heap_SetDebug( TRUE );

    // Set function for allocating memory
    DWC_SetMemFunc( AllocFunc, FreeFunc );
        
    // Main loop
    while (1){
        // Display menu
        DispMenuMsgWithCursor(&gameSeqList[stGameCnt.mode], 0, NULL);

        // Call loop function for sequence
        stGameCnt.mode = gameSeqList[stGameCnt.mode].seqFunc();
    }

	OS_Terminate();
}


/*---------------------------------------------------------------------------*
  Function for initializing game data and host data
 *---------------------------------------------------------------------------*/
static void InitGameData(void)
{

    // Initialize game control structure
    stGameCnt.mode     = GAME_MODE_MAIN;
    stGameCnt.blocking = FALSE;
    MI_CpuClear8(stGameCnt.friendStatus, sizeof(stGameCnt.friendStatus));

    // It's assumed that player data has been loaded from saved data.
    // For the time being, any console can use data having index 0.
    stGameCnt.userData.playerIdx = 0;
    MI_CpuCopy32(saveData.playerData[stGameCnt.userData.playerIdx].playerName,
                 stGameCnt.userData.playerName, sizeof(stGameCnt.userData.playerName));
    stGameCnt.userData.profileID = saveData.profileID;
    stGameCnt.userData.isValidProfile = saveData.isValidProfile;
    MI_CpuCopy32(&saveData.friendList, &stGameCnt.friendList, sizeof(GameFriendList));
}


/*---------------------------------------------------------------------------*
  Additional key set function for matchmaking using game definitions
 *---------------------------------------------------------------------------*/
static void SetGameMatchKeys(void)
{
    int ranking = (int)((OS_GetTick()*1592653589UL+453816691UL) & 0xff);

    // Ranking set as integer key
    stMatchKeys[0].isStr  = 0;
    strcpy(stMatchKeys[0].keyStr, "game_rank");
    stMatchKeys[0].ivalue = ranking;
    stMatchKeys[0].keyID  =
        DWC_AddMatchKeyInt(stMatchKeys[0].keyID,
                           stMatchKeys[0].keyStr,
                           &stMatchKeys[0].ivalue);
    if (stMatchKeys[0].keyID == 0) OS_TPrintf("AddMatchKey failed 0.\n");

    // Evaluation type used for matchmaking set as text string key
    stMatchKeys[1].isStr  = 1;
    strcpy(stMatchKeys[1].keyStr, "game_mtype");
    strcpy(stMatchKeys[1].svalue, "eval_rank");
    stMatchKeys[1].keyID  =
        DWC_AddMatchKeyString(stMatchKeys[1].keyID,
                              stMatchKeys[1].keyStr,
                              stMatchKeys[1].svalue);
    if (stMatchKeys[1].keyID == 0) OS_TPrintf("AddMatchKey failed 1.\n");

    // Set one more integer key for the time being
    stMatchKeys[2].isStr  = 0;
    strcpy(stMatchKeys[2].keyStr, "game_manner");
    stMatchKeys[2].ivalue = ranking%11;
    stMatchKeys[2].keyID  =
        DWC_AddMatchKeyInt(stMatchKeys[2].keyID,
                           stMatchKeys[2].keyStr,
                           &stMatchKeys[2].ivalue);
    if (stMatchKeys[2].keyID == 0) OS_TPrintf("AddMatchKey failed 2.\n");
}


/*---------------------------------------------------------------------------*
  Main function when offline
 *---------------------------------------------------------------------------*/
static GameMode GameMain(void)
{
    GameSequence *gameSeq = &gameSeqList[GAME_MODE_MAIN];
    GameMode returnSeq = GAME_MODE_NUM;
    int curIdx = 0;

#if 0
// sample MP exchange friend data display test
{
    DWCFriendData friendData;
    int i;
    
    DWC_CreateExchangeToken(&stUserData, &friendData);
    for (i = 0; i < 12; i++){
        OS_TPrintf("0x%02x, ", ((u8 *)&friendData)[i]);
    }
    OS_TPrintf("\n");
}
#endif


    while (1){
        dbs_Print( 0, 0, "s:%d", DWC_GetMatchingState() );
        dbs_Print( 7, 0, "n:%d", DWC_GetNumConnectionHost() );
        dbs_Print( 0, 1, "w:%d", DWC_GetLinkLevel() );
        dbs_Print( 10,1, "p:%d", stGameCnt.userData.profileID );
	    dbs_Print( 30, 0, "%c",   loopstr[((loopcnt)>>2)&3] );
		loopcnt++;
		ReadKeyData();  // Get key data

        DWC_ProcessFriendsMatch();  // FriendsMatch communication update

        HandleDWCError(&returnSeq);  // Error processing

        if (stGameCnt.blocking){
            // Prohibit key operations during DWC processing
            // Wait for V-blank
            GameWaitVBlankIntr();
            continue;
        }

        // Authentication complete if the next mode to go to has been set
        if (returnSeq == GAME_MODE_LOGIN){
            // If login has been selected
            if (stGameCnt.userData.isValidProfile){
                // Escape from loop when the profile ID can be retrieved when authentication succeeded
                break;
            }
            else {
                // Cannot proceed to next mode if authentication fails
                returnSeq = GAME_MODE_NUM;
            }
        }
        else if (returnSeq != GAME_MODE_NUM){
            // Always escape from loop in other cases
            break;
        }
            
        ////////// Process the following key operations
        if (stKeyCnt.trg & PAD_BUTTON_A){
            // Select the menu with the A button
            returnSeq = gameSeq->menuList[curIdx].mode;
            stGameCnt.blocking = gameSeq->menuList[curIdx].blocking;

            switch (returnSeq){
            case GAME_MODE_LOGIN:  // Login

                // FriendsMatch Library initialization
                // The send/receive packet size of packets used by the library is set to default (8KB)
                //  
                DWC_InitFriendsMatch(&stDwcCnt, &stUserData, GAME_PRODUCTID, GAME_NAME,
                                     GAME_SECRET_KEY, 0, 0,
                                     stGameCnt.friendList.keyList, GAME_MAX_FRIEND_LIST);

                // Login using the authentication function
                if (!DWC_LoginAsync(L"_INGAMESN_", NULL, LoginCallback, &returnSeq)){
                    // If login cannot start, do not enter blocking state.
                    // If error or double login has not occurred, should not be FALSE.
                    OS_TPrintf("Can't call DWC_LoginAsync().\n");
                    returnSeq          = GAME_MODE_NUM;
                    stGameCnt.blocking = 0;
                }
                break;
            default:
                break;
            }
        }
//#ifdef USE_PRIVATE_USERDATA
#if defined( MULTI_LOGINID_DEFINE )
        // Only when using fixed user data for debugging when PR_USERDATA=TRUE
        // Register friend data for testing.
        // Since processing for checking validity is being skipped, do not use code as-is
        else if (stKeyCnt.trg & PAD_BUTTON_SELECT){
            // When the Select button is pressed, register friend data used in MP exchange
            u64 userid[] = {
                0x000006989862f8c6,
                0x000002726494d2c7,
                0x00000558406532c7,
                0x00000498666552c6,
                    };
            int i;
            for ( i = 0 ; i < sizeof(userid)/sizeof(u64) ; ++i )
            {
                int index = GetAvailableFriendListIndex();
                if ( userid[i] == pseudo_userid ) continue;
                if (index != -1)
                {
                    DWCUserData usr;
                    DWCFriendData frnd;

                    DWC_CreateUserData( &usr, 'ABCD' );
                    DWCi_Acc_SetUserId( &usr.pseudo, userid[i] );
                    DWCi_Acc_SetPlayerId( &usr.pseudo, FIX_PLAYERID );
                    DWC_CreateExchangeToken( &usr, &frnd );
                    MI_CpuCopy8(&frnd,
                                &stGameCnt.friendList.keyList[index],
                                sizeof(DWCFriendData));

                    OS_TPrintf("Set MP friend data to friend list. %d\n", index);
                }
            }
        }
#endif
#if 0
        else if (stKeyCnt.trg & PAD_BUTTON_START){
            // When the Start button is pressed, register the friend registration key
            int index = GetAvailableFriendListIndex();
            if (index != -1){
                DWC_CreateFriendKeyToken(&stGameCnt.friendList.keyList[index],
                                         DWC_Acc_NumericKeyToFriendKey(stSampleFriendKey));

                OS_TPrintf("Set numeric friend key to friend list.\n");
            }
        }
#endif
#if defined( MULTI_LOGINID_DEFINE )
        else if (stKeyCnt.trg & PAD_BUTTON_START){
            // When the Start button is pressed, clear the user data.

            clearUserData();
            OS_TPrintf("ClearUserData\n");
            DWC_ReportUserData( &stUserData );
        }
#endif
        else if (stKeyCnt.trg & PAD_KEY_UP){
            // Move the cursor using up on the crosskey
            curIdx--;
            if (curIdx < 0) curIdx = gameSeq->numMenu-1;
            // Redisplay menu list
            DispMenuMsgWithCursor(gameSeq, curIdx, NULL);
        }
        else if (stKeyCnt.trg & PAD_KEY_DOWN){
            // Move the cursor using down on the crosskey
            curIdx++;
            if (curIdx >= gameSeq->numMenu) curIdx = 0;
            // Redisplay menu list
            DispMenuMsgWithCursor(gameSeq, curIdx, NULL);
        }
        else if ( stKeyCnt.trg & PAD_BUTTON_Y){
            Heap_Dump();
        }
        ////////// Key operation processing ends here

        // Wait for V-blank
        GameWaitVBlankIntr();
    }

    return returnSeq;
}


/*---------------------------------------------------------------------------*
  Main function for registering friends
 *---------------------------------------------------------------------------*/
static GameMode GameRegisterFriendMain(void)
{
    GameSequence *gameSeq = &gameSeqList[GAME_MODE_REG_FRIEND];
    GameMode returnSeq = GAME_MODE_NUM;
    RegFriendParam cntParam;
    int curIdx = 0;
    int friendIdx;
    int maxFigure;
    int validFlg;

    cntParam.step   = 0;
    cntParam.figure = 0;

    while (1){
        ReadKeyData();  // Get key data

        // Escape if the next mode to go to has been set
        if (returnSeq != GAME_MODE_NUM){
            break;
        }
            
        ////////// Process the following key operations
        if (cntParam.step == 0){
            // Currently selecting command
            if (stKeyCnt.trg & PAD_BUTTON_A){
                // Select the menu with the A button
                returnSeq = gameSeq->menuList[curIdx].mode;
                stGameCnt.blocking = gameSeq->menuList[curIdx].blocking;

                if ((curIdx == 0) || (curIdx == 1)){
                    // Register the profile ID of the friend (currently the raw profile ID)
                    // or delete friend data by specifying an index
                    cntParam.step   = 1;
                    cntParam.figure = 0;
                    cntParam.value  = 0;
                    
                    // Redisplay menu list
                    DispMenuMsgWithCursor(gameSeq, curIdx, &cntParam);

                    if (curIdx == 0){
                        maxFigure = 10;  // The profile ID is a maximum of 10 digits in length
                    }
                    else {
                        maxFigure = 2;   // The friend roster index is a maximum of 2 digits
                    }
                }
            }
            else if (stKeyCnt.trg & PAD_KEY_UP){
                // Move the cursor using up on the crosskey
                curIdx--;
                if (curIdx < 0) curIdx = gameSeq->numMenu-1;
                // Redisplay menu list
                DispMenuMsgWithCursor(gameSeq, curIdx, &cntParam);
            }
            else if (stKeyCnt.trg & PAD_KEY_DOWN){
                // Move the cursor using down on the crosskey
                curIdx++;
                if (curIdx >= gameSeq->numMenu) curIdx = 0;
                // Redisplay menu list
                DispMenuMsgWithCursor(gameSeq, curIdx, &cntParam);
            }
        }
        else {
            // Currently entering a numeric value
            if (stKeyCnt.trg & PAD_BUTTON_A){
                // Add/delete friend using the A button
                if (curIdx == 0){
                    // Register friend
                    if (cntParam.value && DWC_CanChangeFriendList()){
                        friendIdx = GetAvailableFriendListIndex();
                        if (friendIdx == -1){
                            OS_TPrintf("Failed to register new friend. Friend list is full.\n");
                        }
                        else {
                            DWC_SetGsProfileId(&stGameCnt.friendList.keyList[friendIdx],
                                               (int)cntParam.value);
                        }
                        validFlg = 1;
                    }
                    else {
                        validFlg = 0;  // Profile ID = 0 is invalid
                    }
                }
                else {
                    // Delete friend
                    if ((cntParam.value < GAME_MAX_FRIEND_LIST) &&
                        DWC_CanChangeFriendList()){
                        // Delete friend from friend roster
                        DWC_DeleteBuddyFriendData(&stGameCnt.friendList.keyList[cntParam.value]);
                        validFlg = 1;
                    }
                    else {
                        validFlg = 0;  // Not on friend roster
                    }
                }

                if (validFlg){
                    // If input is valid, return to command selection
                    cntParam.step = 0;
                    // Redisplay menu list
                    DispMenuMsgWithCursor(gameSeq, curIdx, &cntParam);
                }
            }
            else if (stKeyCnt.trg & PAD_BUTTON_B){
                // Return to command selection using the B button
                cntParam.step = 0;
                // Redisplay menu list
                DispMenuMsgWithCursor(gameSeq, curIdx, &cntParam);
            }
            else if (stKeyCnt.trg & PAD_KEY_UP){
                // Increment/decrement the numeric value being selected using up/down buttons on the crosskey (no check if 2^32 is exceeded)
                if ((cntParam.value/UtilPow(10, cntParam.figure))%10 == 9){
                    cntParam.value -= UtilPow(10,  cntParam.figure)*9;
                }
                else {
                    cntParam.value += UtilPow(10, cntParam.figure);
                }
                // Redisplay menu list
                DispMenuMsgWithCursor(gameSeq, curIdx, &cntParam);
            }
            else if (stKeyCnt.trg & PAD_KEY_DOWN){
                // Increment/decrement the numeric value being selected using up/down buttons on the crosskey (no check if 2^32 is exceeded)
                if ((cntParam.value/UtilPow(10, cntParam.figure))%10 == 0){
                    cntParam.value += UtilPow(10, cntParam.figure)*9;
                }
                else {
                    cntParam.value -= UtilPow(10, cntParam.figure);
                }
                // Redisplay menu list
                DispMenuMsgWithCursor(gameSeq, curIdx, &cntParam);
            }
            else if (stKeyCnt.trg & PAD_KEY_LEFT){
                // Move the number of selected digits using the left/right buttons on the crosskey
                if (cntParam.figure < maxFigure-1) cntParam.figure++;
                // Redisplay menu list
                DispMenuMsgWithCursor(gameSeq, curIdx, &cntParam);
            }
            else if (stKeyCnt.trg & PAD_KEY_RIGHT){
                // Move the number of selected digits using the left/right buttons on the crosskey
                if (cntParam.figure > 0) cntParam.figure--;
                // Redisplay menu list
                DispMenuMsgWithCursor(gameSeq, curIdx, &cntParam);
            }
        }
        ////////// Key operation processing ends here

        // Wait for V-blank
        GameWaitVBlankIntr();
    }

    return returnSeq;
}


/*---------------------------------------------------------------------------*
  Main function after login
 *---------------------------------------------------------------------------*/
static GameMode GameLogonMain(void)
{
    GameSequence *gameSeq = &gameSeqList[GAME_MODE_LOGIN];
    GameMode returnSeq = GAME_MODE_NUM;
    int curIdx = 0;
    BOOL boolRet;

    stNumEntry  = 2;
    stServerIdx = 0;

#if 0
// Sample friend registration key display test
{
    char tmpstr[16];
    
    DWC_Acc_FriendKeyToNumericKey(DWC_CreateFriendKey(&stUserData), tmpstr);
    OS_TPrintf("friendKeyToken '%s'\n", tmpstr);
}
#endif

#ifdef GAME_USE_STORAGE_SERVER
// Login process for test that saves to and loads from GameSpy storage server.
{
    if (!DWC_LoginToStorageServerAsync(LoginToStorageCallback, NULL)){
        OS_TPrintf("DWC_LoginToStorageServerAsync() failed.\n");
    }
}
#endif

#if 0
// Set up test for matchmaking complete options for the number of specified players or less
{
    DWCMatchOptMinComplete optval;

    optval.valid    = 1;
    optval.minEntry = 2;
    optval.timeout  = 30000;

    if (DWC_SetMatchingOption(DWC_MATCH_OPTION_MIN_COMPLETE, &optval, sizeof(optval))){
        OS_TPrintf("Failed to set matching option\n");
    }
}
#endif

#ifdef GAME_MATCH_OP_SC_CONNECT_BLOCK
// After connection is complete, test the option setting for blocking new connections
{
    BOOL boolval = TRUE;

    DWC_SetMatchingOption(DWC_MATCH_OPTION_SC_CONNECT_BLOCK, &boolval, sizeof(boolval));

    stMatchFinishTime = 0;  // Initialize the matchmaking finish time
}
#endif

    while (1){
        dbs_Print( 0, 0, "s:%d", DWC_GetMatchingState() );
        dbs_Print( 7, 0, "n:%d", DWC_GetNumConnectionHost() );
        dbs_Print( 0, 1, "w:%d", DWC_GetLinkLevel() );
        dbs_Print( 10,1, "p:%d", stGameCnt.userData.profileID );
	    dbs_Print( 30, 0, "%c",   loopstr[((loopcnt)>>2)&3] );
		loopcnt++;

        ReadKeyData();  // Get key data

        DWC_ProcessFriendsMatch();  // Update DWC communication processing

        HandleDWCError(&returnSeq);  // Error processing

        if (stGameCnt.blocking){
            // Prohibit key operations during DWC processing (only for cancellations)
            if (stKeyCnt.trg & PAD_BUTTON_B){
                // Cancel matchmaking using the B button
                boolRet = DWC_CancelMatching();
                if (!boolRet){
                    OS_TPrintf("Now unable to cancel matching.\n");
                }
            }

            // Wait for V-blank
            GameWaitVBlankIntr();
            continue;
        }

        // Escapes the loop if the next mode to enter has been set
        if (returnSeq != GAME_MODE_NUM) break;

        ////////// Process the following key operations
        if (stKeyCnt.trg & PAD_BUTTON_A){
            // Select the menu with the A button
            returnSeq = gameSeq->menuList[curIdx].mode;
            stGameCnt.blocking = gameSeq->menuList[curIdx].blocking;
            boolRet   = TRUE;

            switch (curIdx){
            case 0:  // Call peer matchmaking without a friend specification
                if (!(stKeyCnt.cont & PAD_BUTTON_X)){
                    // Supplemental key data settings for matchmaking if the X button hasn't been pressed
                    SetGameMatchKeys();
                }
                
                boolRet =
                    DWC_ConnectToAnybodyAsync((u8)stNumEntry,
                                              "game_mtype = 'eval_rank'",
                                              ConnectToAnybodyCallback,
                                              &returnSeq,
                                              EvaluateAnybodyCallback,
                                              NULL);
                break;
            case 1:  // Call peer matchmaking with friend specification
                if (!(stKeyCnt.cont & PAD_BUTTON_X)){
                    // Supplemental key data settings for matchmaking if the X button hasn't been pressed
                    SetGameMatchKeys();
                }

                // If NULL is specified for the friendIdxList argument, a search is made for all members on the friend roster
                // Search all friends on friend list.
                boolRet = 
                    DWC_ConnectToFriendsAsync(NULL,
                                              0,
                                              (u8)stNumEntry,
                                              TRUE,//FALSE,
                                              ConnectToFriendsCallback,
                                              &returnSeq,
                                              EvaluateFriendCallback,
                                              NULL);
                break;
            case 2:  // begin server-client matchmaking as the server DS
                boolRet =
                    DWC_SetupGameServer((u8)stNumEntry,
                                        SetupGameServerCallback,
                                        &returnSeq,
                                        NewClientCallback,
                                        NULL);
                break;
            case 3:  // begin server-client matchmaking as the client DS
                boolRet =
                    DWC_ConnectToGameServerAsync(stServerIdx,
                                                 ConnectToGameServerCallback,
                                                 &returnSeq,
                                                 NewClientCallback,
                                                 NULL);
                break;
            case 4:  // Logout
                DWC_ShutdownFriendsMatch();   // Shut down DWC FriendsMatch library
                break;
            default:
                break;
            }

            if (!boolRet){
                // If matchmaking could not start, do not enter blocking state.
                // Cannot be FALSE, except when an error has occurred, the user has not logged in, a connection was already made, or the function was called twice.
                // or the function was called twice.
                OS_TPrintf("Can't call matching function.\n");
                returnSeq          = GAME_MODE_NUM;
                stGameCnt.blocking = 0;
            }
        }
        else if (stKeyCnt.trg & PAD_BUTTON_Y){
            if (!(stKeyCnt.cont & PAD_BUTTON_X)){
                // As long as the X button has not been pressed,
                // the current status of friends can be gotten and displayed (statusString) using the Y button
                DispFriendList(TRUE);
                OS_TPrintf("\n");
            }
            else {
                // As long as the X button is being pressed,
                // the current status of friends can be gotten and displayed (statusData) using the Y button
                DispFriendListData();
                OS_TPrintf("\n");
            }
        }
        else if (stKeyCnt.trg & PAD_BUTTON_SELECT){
#ifndef GAME_USE_STORAGE_SERVER  // When testing GameSpy Status settings
            if (!(stKeyCnt.cont & PAD_BUTTON_X)){
                // As long as the X button has not been pressed,
                // the GameSpy Status string can be set using the select button
                if (DWC_SetGsStatusString("SampleString-00")){
                    OS_TPrintf("Set GsStatusString.\n");
                }
                else {
                    OS_TPrintf("Failed to set GsStatusString.\n");
                }
            }
            else {
                // As long as the X button is being pressed,
                // the GameSpy Status data can be set using the select button
                if (DWC_SetGsStatusData(gsStatusSample, sizeof(gsStatusSample))){
                    OS_TPrintf("Set GsStatusData.\n");
                }
                else {
                    OS_TPrintf("Failed to set GsStatusData.\n");
                }
            }

#else  // When testing saving to the storage server
            // Save the test key to the storage server using the select button
            char keyvalues[12];
            BOOL boolResult;

            if (stKeyCnt.cont & PAD_BUTTON_X){
                // save Private data if the X button is pressed
                sprintf(keyvalues, "\\test_key_prv\\%lld", OS_GetTick());
                boolResult = DWC_SavePrivateDataAsync(keyvalues, NULL);
            }
            else {
                // save Public data if the X button is not pressed
                sprintf(keyvalues, "\\test_key_pub\\%lld", OS_GetTick());
                boolResult = DWC_SavePublicDataAsync(keyvalues, NULL);
            }

            if (boolResult){
                OS_TPrintf("Saved '%s'.\n", keyvalues);
            }
            else {
                OS_TPrintf("Can't save to storage server.\n");
            }
#endif
        }
        else if (stKeyCnt.trg & PAD_BUTTON_START){
#ifndef GAME_USE_STORAGE_SERVER  // When testing GameSpy Status settings
            char statusData[DWC_GS_STATUS_STRING_LEN];
            
            if (!(stKeyCnt.cont & PAD_BUTTON_X)){
                // As long as the X button has not been pressed,
                // the GameSpy Status string can be gotten using the Start button
                if (DWC_GetGsStatusString(statusData)){
                    OS_TPrintf("Got my GsStatusString : %s\n", statusData);
                }
                else {
                    OS_TPrintf("Failed to get GsStatusString.\n");
                }
            }
            else {
                // As long as the X button is being pressed,
                // the GameSpy Status data can be gotten using the Start button
                int size, i;

                size = DWC_GetGsStatusData(statusData);
                if (size != -1){
                    OS_TPrintf("Got my GsStatusData\n");
                    for (i = 0; i < size; i++){
                        OS_TPrintf("%02d, ", statusData[i]);
                    }
                    OS_TPrintf("\n");
                }
                else {
                    OS_TPrintf("Failed to get GsStatusData.\n");
                }
            }

#else  // When testing saving to the storage server
            // Load test data from the storage server using the Start button
            BOOL boolResult;

#if 1
            if (stKeyCnt.cont & PAD_BUTTON_X){
                // load self's Private data if the X button is pressed
                boolResult = DWC_LoadOwnPrivateDataAsync("\\test_key_prv\\test_key_pub", NULL);
            }
            else {
                // load self's Public data if the X button is not pressed
                boolResult = DWC_LoadOwnPublicDataAsync("\\test_key_prv\\test_key_pub", NULL);
            }
#else
            // Load host data on the friend roster
            boolResult = DWC_LoadOthersDataAsync("\\test_key", stServerIdx, NULL);
#endif
            if (!boolResult) OS_TPrintf("Can't load from storage server.\n");
#endif
        }
        else if (stKeyCnt.trg & PAD_KEY_UP){
            // Move the cursor using up on the crosskey
            curIdx--;
            if (curIdx < 0) curIdx = gameSeq->numMenu-1;
            // Redisplay menu list
            DispMenuMsgWithCursor(gameSeq, curIdx, NULL);
        }
        else if (stKeyCnt.trg & PAD_KEY_DOWN){
            // Move the cursor using down on the crosskey
            curIdx++;
            if (curIdx >= gameSeq->numMenu) curIdx = 0;
            // Redisplay menu list
            DispMenuMsgWithCursor(gameSeq, curIdx, NULL);
        }
        else if (stKeyCnt.trg & PAD_KEY_LEFT){
            // Decrement the number of entries using the left button on the crosskey
            if (stNumEntry > 2){
                stNumEntry--;
                // Redisplay menu list
                DispMenuMsgWithCursor(gameSeq, curIdx, NULL);
            }
        }
        else if (stKeyCnt.trg & PAD_KEY_RIGHT){
            // Increment the number of entries using the right button on the crosskey
            if (stNumEntry < DWC_MAX_CONNECTIONS){
                stNumEntry++;
                // Redisplay menu list
                DispMenuMsgWithCursor(gameSeq, curIdx, NULL);
            }
        }
        else if (stKeyCnt.trg & PAD_BUTTON_L){
            // select the friend list index with the LR button for the
            // server-client matchmaking server DS
            if (curIdx == 3){
                if (stServerIdx > 0){
                    stServerIdx--;
                    // Redisplay menu list
                    DispMenuMsgWithCursor(gameSeq, curIdx, NULL);
                }
            }
        }
        else if (stKeyCnt.trg & PAD_BUTTON_R){
            // select the friend list index with the LR button for the
            // server-client matchmaking server DS
            if (curIdx == 3){
                if (stServerIdx < GAME_MAX_FRIEND_LIST-1){
                    stServerIdx++;
                    // Redisplay menu list
                    DispMenuMsgWithCursor(gameSeq, curIdx, NULL);
                }
            }
        }
        ////////// Key operation processing ends here

        // Wait for V-blank
        GameWaitVBlankIntr();
    }

    return returnSeq;
}


/** -------------------------------------------------------------------------
  DWC Transport
  ---------------------------------------------------------------------------*/
const int		SIZE_SEND_BUFFER = 512;
const int		SIZE_RECV_BUFFER = 4 * 1024;

static u8		sSendBuffer[ SIZE_SEND_BUFFER ];
static u8		sRecvBuffer[ DWC_MAX_CONNECTIONS-1 ][ SIZE_RECV_BUFFER ];

/** -------------------------------------------------------------------------
  A transmission completion callback.
  ---------------------------------------------------------------------------*/
static void
SendDoneCallback( int size, u8 aid )
{
	OS_TPrintf( "to aid = %d  size = %d\n", aid, size );
}

/** -------------------------------------------------------------------------
  Receive complete callback
  ---------------------------------------------------------------------------*/
static void
UserRecvCallback( u8 aid, u8* buffer, int size )
{
	OS_TPrintf( "from aid = %d  size = %d  buffer[0] = %X\n", aid, size, buffer[0] );
}

/** -------------------------------------------------------------------------
  Set receive buffer for DWC transport
  ---------------------------------------------------------------------------*/
static void
SetRecvBuffer( void )
{
    u8* pAidList;
	int i;
    int j = 0;
	int connectingNum = DWC_GetAIDList(&pAidList);

    OS_TPrintf( "connection host = %d\n", connectingNum );
    
    for ( i = 0; i < connectingNum; ++i )
    {
		if ( pAidList[i] == DWC_GetMyAID() )
        {
            j++;
			continue;
        }

        DWC_SetRecvBuffer( pAidList[i], &sRecvBuffer[i-j], SIZE_RECV_BUFFER );
    }
}

/** -------------------------------------------------------------------------
  Reliable send
  ---------------------------------------------------------------------------*/
static void
BroadCastReliable( void )
{
#if 1
    // If sending to all members using the AID list
    u8* pAidList;
    int numHost;
    int i;

    numHost = DWC_GetAIDList(&pAidList);

    for (i = 0; i < numHost; i++){
        if (pAidList[i] == DWC_GetMyAID()) continue;

        DWC_SendReliable( pAidList[i], sSendBuffer, SIZE_SEND_BUFFER );
    }

#else
    // If sending to all members using an AID bitmap
    u32 bitmap = DWC_SendReliableBitmap( DWC_GetAIDBitmap(), sSendBuffer, SIZE_SEND_BUFFER );
    OS_TPrintf("Sent to %x.\n", bitmap);
#endif
}

/** -------------------------------------------------------------------------
  Unreliable transfer
  ---------------------------------------------------------------------------*/
static void
BroadCastUnreliable( void )
{
#if 1
    // If sending to all members using the AID list
    u8* pAidList;
    int numHost;
    int i;

    numHost = DWC_GetAIDList(&pAidList);

    for (i = 0; i < numHost; i++){
        if (pAidList[i] == DWC_GetMyAID()) continue;

        DWC_SendUnreliable( pAidList[i], sSendBuffer, SIZE_SEND_BUFFER );
    }

#else
    // If sending to all members using an AID bitmap
    u32 bitmap = DWC_SendUnreliableBitmap( DWC_GetAIDBitmap(), sSendBuffer, SIZE_SEND_BUFFER );
    OS_TPrintf("Sent to %x.\n", bitmap);
#endif
}

/*---------------------------------------------------------------------------*
  Main function after mesh connection has been completed
 *---------------------------------------------------------------------------*/
static GameMode GameConnectedMain(void)
{
    GameSequence *gameSeq = &gameSeqList[GAME_MODE_CONNECTED];
    GameMode returnSeq = GAME_MODE_NUM;
    int closeResult;
    int curIdx = 0;
    
    // Sets the send callback	
    DWC_SetUserSendCallback( SendDoneCallback ); 

    // Set the receive callback	
    DWC_SetUserRecvCallback( UserRecvCallback ); 
    
    // Set the receive buffer
    // when server-client matchmaking is set here
    // the receive buffers of clients that join after this won't be set!
    //SetRecvBuffer();

    // Set the callback for closing connections
    DWC_SetConnectionClosedCallback(ConnectionClosedCallback, &returnSeq);

#if 0
	// Set the packet loss rate
	if ( DWC_GetMyAID() != 0 )
    {
		DWC_SetSendDrop( 30, 0 );
		DWC_SetRecvDrop( 30, 0 );        
    }
#endif
#if 0
    // Set the delay (unit: msec)  
	if ( DWC_GetMyAID() != 0 )
    {
		DWC_SetSendDelay( 500, 0 );
		DWC_SetRecvDelay( 500, 0 );        
    }
#endif    

    OS_TPrintf("AID bitmap = %x\n", DWC_GetAIDBitmap());
    OS_TPrintf( "my aid = %d\n", DWC_GetMyAID() );

    while (1){        
        dbs_Print( 0, 0, "s:%d", DWC_GetMatchingState() );
        dbs_Print( 7, 0, "n:%d", DWC_GetNumConnectionHost() );
        dbs_Print( 0, 1, "w:%d", DWC_GetLinkLevel() );
        dbs_Print( 10,1, "p:%d", stGameCnt.userData.profileID );
	    dbs_Print( 30, 0, "%c",   loopstr[((loopcnt)>>2)&3] );
		loopcnt++;

        ReadKeyData();  // Get key data

        DWC_ProcessFriendsMatch();  // Update DWC communication processing

        HandleDWCError(&returnSeq);  // Error processing

        if (stGameCnt.blocking){
            // Prohibit key operations during DWC processing
            // Wait for V-blank
            GameWaitVBlankIntr();
            continue;
	    }

#ifdef GAME_MATCH_OP_SC_CONNECT_BLOCK
        if ((stMatchFinishTime != 0) &&
            (OS_TicksToSeconds(OS_GetTick()-stMatchFinishTime) > GAME_MATCH_OP_SC_BLOCK_TIME)){
            // After a single case of matchmaking is complete, release the block on new connections once a set period of time has elapsed
            DWC_ClearMOSCConnectBlock();

            stMatchFinishTime = 0;
            OS_TPrintf("Clear matching block!.\n");
        }
#endif

        // Escapes the loop if the next mode to enter has been set
        if (returnSeq != GAME_MODE_NUM) break;

        ////////// Process the following key operations
        if (stKeyCnt.trg & PAD_BUTTON_A){
            // Select the menu with the A button
            returnSeq = gameSeq->menuList[curIdx].mode;
            stGameCnt.blocking = gameSeq->menuList[curIdx].blocking;

            switch (curIdx){
            case 0:             // Reliable transfer
                BroadCastReliable();
				break;
            case 1:             // Unreliable transfer
				BroadCastUnreliable();
                break;
            case 2:             // Close connection
                stCloseFlag = TRUE;
                closeResult = DWC_CloseAllConnectionsHard();
                break;
            default:
                break;
            }
        }
#ifdef DWC_STOP_SC_SERVER
        else if (stKeyCnt.trg & PAD_BUTTON_START){
            // cancel the server-client matchmaking with the START button
            BOOL boolResult = DWC_StopSCMatchingAsync(StopSCMatchingCallback, NULL);
            OS_TPrintf("DWC_StopSCMatchingAsync : result %d\n", boolResult);
        }
#endif
        else if (stKeyCnt.trg & PAD_BUTTON_SELECT){
            // Forcibly disconnect the host with aid = 1 using the select button.
            // Test when disconnecting a host whose power has been cut during communications
            u32 bitmap = 0x02;
#if 1
            // AID specification version
            closeResult = DWC_CloseConnectionHard(1);
#else
            // AID bitmap specification version
            closeResult = DWC_CloseConnectionHardBitmap(&bitmap);
#endif

            OS_TPrintf("Closed connection to aid = 1, result = %d, bitmap = %x\n",
                       closeResult, bitmap);

            // Do not perform blocking since processing will continue regardless of the result.
            // Also, call the close callback from within the CloseConnectonHard function.
            stGameCnt.blocking = FALSE;
        }
        else if (stKeyCnt.trg & PAD_KEY_UP){
            // Move the cursor using up on the crosskey
            curIdx--;
            if (curIdx < 0) curIdx = gameSeq->numMenu-1;
            // Redisplay menu list
            DispMenuMsgWithCursor(gameSeq, curIdx, NULL);
        }
        else if (stKeyCnt.trg & PAD_KEY_DOWN){
            // Move the cursor using down on the crosskey
            curIdx++;
            if (curIdx >= gameSeq->numMenu) curIdx = 0;
            // Redisplay menu list
            DispMenuMsgWithCursor(gameSeq, curIdx, NULL);
        }
        ////////// Key operation processing ends here

        // Wait for V-blank
        GameWaitVBlankIntr();
    }

    return returnSeq;
}


/*---------------------------------------------------------------------------*
  Function for getting an empty index on the friend roster
 *---------------------------------------------------------------------------*/
static int GetAvailableFriendListIndex(void)
{
    int i;

    for (i = 0; i < GAME_MAX_FRIEND_LIST; i++){
        if (!DWC_IsValidFriendData(&stGameCnt.friendList.keyList[i]))
            return i;
    }

    return -1;
}


/*---------------------------------------------------------------------------*
  Login callback function
 *---------------------------------------------------------------------------*/
static void LoginCallback(DWCError error, int profileID, void *param)
{
    BOOL result;
	RTCTime time;
	RTCDate date;

	if (error == DWC_ERROR_NONE){
        // Get authentication success status and profile ID
        OS_TPrintf("Login succeeded. profileID = %u\n\n", profileID);

        // Gets the result of the ingamesn check.
		if(DWC_GetIngamesnCheckResult() == DWC_INGAMESN_INVALID)
			OS_TPrintf("BAD ingamesn is detected by NAS.\n");
		else if(DWC_GetIngamesnCheckResult() == DWC_INGAMESN_VALID)
			OS_TPrintf("GOOD ingamesn is detected by NAS.\n");
		else
			OS_TPrintf("ingamesn is not checked yet.\n");
		
		// display the time on the authentication server
		if(DWC_GetDateTime(&date, &time) == TRUE)
			OS_TPrintf("NasTime = %02d/%02d/%02d %02d:%02d:%02d\n", date.year+2000, date.month, date.day, time.hour, time.minute, time.second);
		else
			DWC_Printf(DWC_REPORTFLAG_AUTH, "DWC_GetNasTime failed\n");


        // Check whether or not stUserData has been updated.
        if ( DWC_CheckDirtyFlag( &stUserData ) )
        {
            DWC_ClearDirtyFlag( &stUserData );

            OS_TPrintf("*******************************\n");
            OS_TPrintf("You must save the DWCUserData!!\n");
            OS_TPrintf("*******************************\n");

            // Always check using this timing, and if a dirty flag is active,
            // be sure to save DWCUserData in backup on the DS card.
            // saveUserDataToDSCard( &stUserData );
        }

        stGameCnt.userData.profileID = profileID;
        stGameCnt.userData.isValidProfile = TRUE;

        // Test setting the GameSpy Status string
        //DWC_SetGsStatusData(gsStatusSample, sizeof(gsStatusSample));
        

        // begin friend list synchronization
        result = DWC_UpdateServersAsync(NULL,//stGameCnt.userData.playerName,
                                        UpdateServersCallback, param,
                                        FriendStatusCallback, param,
                                        DeleteFriendListCallback, param);
        if (result == FALSE){
            // If update was not possible, release the blocking state.
            // Cannot be FALSE, except when an error has occurred, the user has not logged in, or the function was called twice.
            OS_TPrintf("Can't call DWC_UpdateServersAsync().\n");
             *(GameMode *)param = GAME_MODE_NUM;
            stGameCnt.blocking  = 0;
        }
        else {
            // Register the callback for establishing friends on the GameSpy server
            DWC_SetBuddyFriendCallback(BuddyFriendCallback, NULL);

            // Register a callback for making notification that storage save/load is complete
            DWC_SetStorageServerCallback(SaveToServerCallback,
                                         LoadFromServerCallback);
        }
    }
    else {
        // Authentication failed
        OS_TPrintf("Login failed. %d\n\n", error);

        stGameCnt.blocking = FALSE;  // Cancel blocking
    }
}


/*---------------------------------------------------------------------------*
  friend list synchronization complete callback function
 *---------------------------------------------------------------------------*/
static void UpdateServersCallback(DWCError error, BOOL isChanged, void* param)
{
#pragma unused(param)

    if (error == DWC_ERROR_NONE){
        // friend list synchronization completed
        OS_TPrintf("Updating servers succeeded.\n");

        if (isChanged){
            // Save the friend roster if it has been modified
            // (Should be saved)
            MI_CpuCopy32(&stGameCnt.friendList, &saveData.friendList, sizeof(GameFriendList));
        }

        // Display the friend roster
        DispFriendList(TRUE);
    }
    else {
        // Do nothing particular even if processing failed
        OS_TPrintf("Failed to update GameSpy servers. %d\n\n", error);
    }

    stGameCnt.blocking = FALSE;  // Cancel blocking
}


/*---------------------------------------------------------------------------*
  Callback function for notification of changes in friend status
 *---------------------------------------------------------------------------*/
static void FriendStatusCallback(int index, u8 status, const char* statusString, void* param)
{
#pragma unused(param)

    OS_TPrintf("Friend[%d] changed status -> %d (statusString : %s).\n",
               index, status, statusString);
}


/*---------------------------------------------------------------------------*
  Friend roster delete callback
 *---------------------------------------------------------------------------*/
static void DeleteFriendListCallback(int deletedIndex, int srcIndex, void* param)
{
#pragma unused(param)

    OS_TPrintf("friend[%d] was deleted (equal friend[%d]).\n",
               deletedIndex, srcIndex);
}


/*---------------------------------------------------------------------------*
  Callback function for establishing GameSpy friends
 *---------------------------------------------------------------------------*/
static void BuddyFriendCallback(int index, void* param)
{
#pragma unused(param)

    OS_TPrintf("I was authorized by friend[%d].\n", index);
}


/*---------------------------------------------------------------------------*
  Callback function for completion of login to storage server
 *---------------------------------------------------------------------------*/
static void LoginToStorageCallback(DWCError error, void* param)
{
#pragma unused(param)

    OS_TPrintf("Login to storage server: result %d\n", error);
}


/*---------------------------------------------------------------------------*
  Callback function for completing data save to storage server
 *---------------------------------------------------------------------------*/
static void SaveToServerCallback(BOOL success, BOOL isPublic, void* param)
{

    OS_TPrintf("Saved data to server. %08x\n", (u32)param);
    OS_TPrintf("result %d, isPublic %d.\n", success, isPublic);
}


/*---------------------------------------------------------------------------*
  Callback function for completing data load from storage server
 *---------------------------------------------------------------------------*/
static void LoadFromServerCallback(BOOL success, int index, char* data, int len, void* param)
{

    OS_TPrintf("Loaded data from server. %08x\n", (u32)(param));
    OS_TPrintf("result %d, index %d, data '%s', len %d\n",
               success, index, data, len);
}


/*---------------------------------------------------------------------------*
  Callback function for connecting without a friend specification
 *---------------------------------------------------------------------------*/
static void ConnectToAnybodyCallback(DWCError error, BOOL cancel, void* param)
{

    if (error == DWC_ERROR_NONE){
        if (!cancel){
            // Establish connection with unknown players
            OS_TPrintf("Succeeded to connect to anybody\n\n");

            // Set the receive buffer
            SetRecvBuffer();
        }
        else {
            OS_TPrintf("Canceled matching.\n");
            // Returns to the status after login
            *(GameMode *)param = GAME_MODE_LOGIN;
        }
    }
    else {
        OS_TPrintf("Failed to connect to anybody. %d\n\n", error);
    }

    stGameCnt.blocking = FALSE;  // Cancel blocking
}


/*---------------------------------------------------------------------------*
  Callback function for connecting with a friend specification
 *---------------------------------------------------------------------------*/
static void ConnectToFriendsCallback(DWCError error, BOOL cancel, void* param)
{

    if (error == DWC_ERROR_NONE){
        if (!cancel){
            // Establish connection with friend
            OS_TPrintf("Succeeded to connect to friends\n\n");

            // Set the receive buffer
            SetRecvBuffer();
        }
        else {
            OS_TPrintf("Canceled matching.\n");
            // Returns to the status after login
            *(GameMode *)param = GAME_MODE_LOGIN;
        }
    }
    else {
        OS_TPrintf("Failed to connect to friends. %d\n\n", error);
    }

    stGameCnt.blocking = FALSE;  // Cancel blocking
}


/*---------------------------------------------------------------------------*
  server-client matchmaking server launch callback function
 *---------------------------------------------------------------------------*/
static void SetupGameServerCallback(DWCError error,
                                    BOOL cancel,
                                    BOOL self,
                                    BOOL isServer,
                                    int  index,
                                    void* param)
{
#pragma unused(isServer)

    if (error == DWC_ERROR_NONE){
        if (!cancel){
            // A new client has been added to the network
            OS_TPrintf("A friend of mine joined the game.\n");
            OS_TPrintf("friendListIndex = %d.\n\n", index);

            // Set the receive buffer
            SetRecvBuffer();

#ifdef GAME_MATCH_OP_SC_CONNECT_BLOCK
            // Record the time when a single case of matchmaking has completed
            stMatchFinishTime = OS_GetTick();
#endif
        }
        else {
            if (self){
                // The self has cancelled matching
                OS_TPrintf("Canceled matching.\n\n");
                // Returns to the status after login
                // The reference point for param is auto variable returnSeq, but when matchmaking
                // can be cancelled by the user, it's within the scope.
                *(GameMode *)param = GAME_MODE_LOGIN;
            }
            else {
                // the client DS canceled matchmaking
                OS_TPrintf("Client (friendListIndex = %d) canceled matching.\n\n", index);

                if (DWC_GetNumConnectionHost() == 1){
                    // If nobody is connected yet, blocking is not released.
                    // After at least one matchmaking has been completed and the number of connections is 1
                    // it is already in a blocking-released state, so there is no problem.
                    return;
                }
            }
        }
    }
    else {
        // Error occurred
        OS_TPrintf("Game server error occured. %d\n\n", error);
    }

    stGameCnt.blocking = FALSE;  // Cancel blocking
}


/*---------------------------------------------------------------------------*
  server-client matchmaking server connection callback function
 *---------------------------------------------------------------------------*/
static void ConnectToGameServerCallback(DWCError error,
                                        BOOL cancel,
                                        BOOL self,
                                        BOOL isServer,
                                        int  index,
                                        void* param)
{

    if (error == DWC_ERROR_NONE){
        if (!cancel){
            if (self){
                // If the local host has succeeded in connecting to the network established by the server DS
                //  
                OS_TPrintf("Succeeded to join the game.\n\n");
            }
            else {
                // If a new client has been added to the network
                OS_TPrintf("New client joined the game.\n");
                if (index != -1){
                    OS_TPrintf("he is my friend [%d].\n\n", index);
                }
                else {
                    OS_TPrintf("he is not my friend.\n\n");
                }
            }

            // Set the receive buffer
            SetRecvBuffer();
        }
        else {
            if (self){
                // The self has cancelled matching
                OS_TPrintf("Canceled matching.\n\n");
                // Returns to the status after login
                *(GameMode *)param = GAME_MODE_LOGIN;
            }
            else {
                if (isServer){
                    // the server DS canceled matchmaking.
                    OS_TPrintf("GameServer canceled matching.\n\n");
                    // Matchmaking is ended and the status returns to that after login
                    *(GameMode *)param = GAME_MODE_LOGIN;
                }
                else {
                    // Another client DS has cancelled matchmaking
                    OS_TPrintf("Host (friendListIndex = %d) canceled matching.\n\n", index);
                }
            }
        }
    }
    else {
        // Error occurred
        OS_TPrintf("Failed to join the game. %d\n\n", error);
    }

    stGameCnt.blocking = FALSE;  // Cancel blocking
}


/*---------------------------------------------------------------------------*
  Callback function for closing connections
 *---------------------------------------------------------------------------*/
static void ConnectionClosedCallback(DWCError error,
                                     BOOL isLocal,
                                     BOOL isServer,
                                     u8  aid,
                                     int index,
                                     void* param)
{

    if (error == DWC_ERROR_NONE){
        if (!isLocal && isServer){
            // When using server-client matchmaking and the local host is a client,
            // the server DS closed the connection.
            // Since the server DS has closed, the local host also closes.
            stCloseFlag = TRUE;
            DWC_CloseAllConnectionsHard();
        }
        
        if (DWC_GetNumConnectionHost() == 1){
            // When all connections are closed,
            if ((DWC_GetLastMatchingType() != DWC_MATCH_TYPE_SC_SV) || stCloseFlag){
                // when the server-client matchmaking server DS calls the close function
                // and closes, for other matchmaking types,
                // the following mode is set here.
                *(GameMode *)param = GAME_MODE_LOGIN;
                stCloseFlag = FALSE;

                // In preparation for more matchmaking, get local host data and register it with the GameSpy server
                // Nothing happens if you call this function when using server-client type matchmaking
                DWC_RegisterMatchingStatus();
            }
            // When the self  is on the server-client matchmaking server DS,
            // do nothing if a client closes the connections of all members
        }

        if (isLocal){
            OS_TPrintf("Closed connection to aid %d (friendListIndex = %d) Rest %d.\n",
                       aid, index, DWC_GetNumConnectionHost());
        }
        else {
            OS_TPrintf("Connection to aid %d (friendListIndex = %d) was closed. Rest %d.\n",
                       aid, index, DWC_GetNumConnectionHost());
        }
    }
    else {
        OS_TPrintf("Failed to close connections. %d\n\n", error);
    }
}


/*---------------------------------------------------------------------------*
  new connection client notification callback function for server-client matchmaking
 *---------------------------------------------------------------------------*/
static void NewClientCallback(int index, void* param)
{
#pragma unused(param)

    // Controller operations are disabled until connection processing for newly connected clients ends
    //  
    // Do not set blocking during actual game processing
    //stGameCnt.blocking = TRUE;
    
    OS_TPrintf("New client started to connect.\n");
    if (index != -1){
        OS_TPrintf("He is my friend [%d].\n", index);
    }
    else {
        OS_TPrintf("He is not my friend.\n");
    }
}


/*---------------------------------------------------------------------------*
  callback function for evaluating players when using peer matching without a friend specification
 *---------------------------------------------------------------------------*/
static int  EvaluateAnybodyCallback(int index, void* param)
{
#pragma unused(param)
    const char* matchType = DWC_GetMatchStringValue(index, stMatchKeys[1].keyStr, "NONE");
    int rank;
    int diff;
#if 0  // Set this to 1 if you want to perform an evaluation test
    int full = 200;
#else  // Put here since it's usually a hassle when matchmaking isn't always performed
    int full = 300;
#endif
    
    if (!strncmp(matchType, "eval_rank", strlen("eval_rank"))){
        // When the other party also sets "eval_rank" as the matchmaking type
        rank = DWC_GetMatchIntValue(index, stMatchKeys[0].keyStr, -1);
        if (rank == -1){
            return 0;  // matchmaking is not performed without a "game_rank" key
        }

        diff = (stMatchKeys[0].ivalue > rank) ? stMatchKeys[0].ivalue-rank : rank-stMatchKeys[0].ivalue;

        OS_TPrintf("Player[%d]'s rank is %d (mine %d). %d point.\n",
                   index, rank, stMatchKeys[0].ivalue, full-diff);

        // Assign point score such that the smaller the difference in rank as compared to the local host, the higher the points assigned, with 200 points being the maximum.
        // Do not perform matchmaking for scores of 200 or higher (Return values of 0 and lower)
        return full-diff;
    }
    else if (!strncmp(matchType, "NONE", strlen("NONE"))){
        return 1;  // If there is no "game_mtype" key, perform matchmaking evenly
    }

    return 1;
}


/*---------------------------------------------------------------------------*
  callback function for evaluating players when using peer matching with a friend specification
 *---------------------------------------------------------------------------*/
static int  EvaluateFriendCallback(int index, void* param)
{
#pragma unused(param)
    const char* matchType = DWC_GetMatchStringValue(index, stMatchKeys[1].keyStr, "NONE");
    
    if (!strncmp(matchType, "eval_rank", strlen("eval_rank"))){
        // When the other party also sets "eval_rank" as the matchmaking type
        return 1;  // matchmaking is OK
    }
    else {
        return 0;  // matchmaking is not performed without a "game_mtype" key
    }
}


#ifdef DWC_STOP_SC_SERVER
/*---------------------------------------------------------------------------*
  server-client matchmaking cancellation complete callback function
 *---------------------------------------------------------------------------*/
static void StopSCMatchingCallback(void* param)
{
#pragma unused(param)

    OS_TPrintf("StopSCMatchingCallback was called.\n");
}
#endif


/*---------------------------------------------------------------------------*
  Function for processing DWC library errors
 *---------------------------------------------------------------------------*/
static DWCError HandleDWCError(GameMode* gameMode)
{
    int errorCode;
    DWCError error;
    DWCErrorType errorType;

    // Get DWC error
    error = DWC_GetLastErrorEx(&errorCode, &errorType);

    // Return immediately if there is no error
    if (error == DWC_ERROR_NONE) return DWC_ERROR_NONE;

    // Display message that corresponds to the error
    DispErrorMessage(error, errorCode, errorType);

    DWC_ClearError();        // Clear error
    stGameCnt.blocking = 0;  // Cancel blocking

    switch (errorType){
    case DWC_ETYPE_LIGHT:
        // Always return to status after login since at this point
        // Must return to post-login state
        *gameMode = GAME_MODE_LOGIN;
        break;

    case DWC_ETYPE_SHUTDOWN_FM:
        // Terminate FriendsMatch processing and release the GameSpy heap
        DWC_ShutdownFriendsMatch();

        // Return to status before login
        *gameMode = GAME_MODE_MAIN;
        break;

    case DWC_ETYPE_DISCONNECT:
        // Terminate FriendsMatch processing and release the GameSpy heap
        DWC_ShutdownFriendsMatch();

        // Completely initialize Internet connections
        DWC_CleanupInet();

        // Return to status before login
        *gameMode = GAME_MODE_MAIN;
        break;

    case DWC_ETYPE_FATAL:
        // Stop due to Fatal Error
        OS_Panic("========== Stop process ==========\n");
        break;
    }

    return error;
}

/*---------------------------------------------------------------------------*
  Function for displaying communication errors
 *---------------------------------------------------------------------------*/
static void DispErrorMessage(DWCError error, int errorCode, DWCErrorType errorType)
{

    if (error == DWC_ERROR_NONE) return;

    OS_TPrintf("ERROR: %d, code:%d, type:%d\n", error, -1*errorCode, errorType);
    
    switch (error){
    case DWC_ERROR_DS_MEMORY_ANY:
        OS_TPrintf("ERROR: DS_MEMORY_ANY.\n");
        break;
    case DWC_ERROR_AUTH_ANY:
        OS_TPrintf("ERROR: AUTH_ANY.\n");
        break;
    case DWC_ERROR_AUTH_OUT_OF_SERVICE:
        OS_TPrintf("ERROR: This game's network service has ended.\n");
        break;
    case DWC_ERROR_AUTH_STOP_SERVICE:
        OS_TPrintf("ERROR: Now server is temporarily unavailable.\nPlease re-login later.\n");
        break;
    case DWC_ERROR_AC_ANY:
        OS_TPrintf("ERROR: AC ANY.\n");
        break;
    case DWC_ERROR_NETWORK:
        OS_TPrintf("ERROR: Network error.\n");
        break;
    case DWC_ERROR_DISCONNECTED:
        OS_TPrintf("ERROR: Connection to AP was lost.\n");
        break;
    case DWC_ERROR_FATAL:
        OS_TPrintf("FATAL ERROR: Please turn the power off.\n");
        break;
    case DWC_ERROR_FRIENDS_SHORTAGE:
        OS_TPrintf("ERROR: Your friends seem not to login yet.\n");
        break;
    case DWC_ERROR_NOT_FRIEND_SERVER:
        OS_TPrintf("ERROR: He is not buddy or not game server.\n");
        break;
    case DWC_ERROR_MO_SC_CONNECT_BLOCK:
        OS_TPrintf("ERROR: Game server is busy now.\n");
        break;
    case DWC_ERROR_SERVER_FULL:
        OS_TPrintf("ERROR: Game server is fully occupied.\n");
        break;
    default:
        OS_TPrintf("ERROR: Undefined yet.\n");
        break;
    }
}


/*---------------------------------------------------------------------------*
  Function for displaying the message list (with cursor)
 *---------------------------------------------------------------------------*/
static void DispMenuMsgWithCursor(const GameSequence* gameSeq, int index, void* param)
{
    int i;

    OS_TPrintf("===============================\n");
    OS_TPrintf("%s\n", gameSeq->modeName);

    for (i = 0; i < gameSeq->numMenu; i++){
        if (i == index) OS_TPrintf("> ");
        else OS_TPrintf("  ");

        OS_TPrintf("%s\n", gameSeq->menuList[i].msg);
    }

    // Call the callback function for making additional displays used by each mode
    if (gameSeq->dispCallback) gameSeq->dispCallback(index, param);

    OS_TPrintf("===============================\n\n");
    OS_TPrintf("+++VALID CONSOLE %d +++\n", DWC_CheckValidConsole( &stUserData ) );
    Heap_Print();
}


/*---------------------------------------------------------------------------*
  Function for displaying the friend roster (version that uses the GameSpy Status string)
 *---------------------------------------------------------------------------*/
static void DispFriendList(BOOL overLoad)
{
    u8   maxEntry = 0;
    u8   numEntry;
    char statusString[DWC_GS_STATUS_STRING_LEN];
    int  i;

    statusString[0] = '\0';

    for (i = 0; i < GAME_MAX_FRIEND_LIST; i++){
        if (!DWC_IsValidFriendData(&stGameCnt.friendList.keyList[i]))
            continue;
        
        if (overLoad){
            // Also update friend information if a specification is overwritten
            stGameCnt.friendStatus[i] =
                DWC_GetFriendStatusSC(&stGameCnt.friendList.keyList[i],
                                      &maxEntry, &numEntry, statusString);
        }
        
        if (maxEntry){
            // If the friend is a server-client matchmaking server DS,
            // display (Current Number of Connected Players/Maximum Number of Connected Players)
            OS_TPrintf("Friend[%d] '%s' : profileID %u, status %d, server (%d/%d), statusString : %s\n",
                       i, stGameCnt.friendList.playerName[i],
                       DWC_GetGsProfileId(&stUserData, &stGameCnt.friendList.keyList[i]),
                       stGameCnt.friendStatus[i],
                       numEntry, maxEntry, statusString);
        }
        else {
            OS_TPrintf("Friend[%d] '%s' : profileID %u, status %d, statusString : %s\n",
                       i, stGameCnt.friendList.playerName[i],
                       DWC_GetGsProfileId(&stUserData, &stGameCnt.friendList.keyList[i]),
                       stGameCnt.friendStatus[i], statusString);
        }
    }
}


/*---------------------------------------------------------------------------*
  Function for displaying the friend roster (version that uses the GameSpy Status data)
 *---------------------------------------------------------------------------*/
static void DispFriendListData(void)
{
    u8   maxEntry = 0;
    u8   numEntry;
    char statusData[DWC_GS_STATUS_STRING_LEN];
    int  size;
    int  i, j;

    for (i = 0; i < GAME_MAX_FRIEND_LIST; i++){
        if (!DWC_IsValidFriendData(&stGameCnt.friendList.keyList[i]))
            continue;
        
        // Updates friend information
        stGameCnt.friendStatus[i] =
            DWC_GetFriendStatusDataSC(&stGameCnt.friendList.keyList[i],
                                      &maxEntry, &numEntry, statusData, &size);
        
        if (maxEntry){
            // If the friend is a server-client matchmaking server DS,
            // display (Current Number of Connected Players/Maximum Number of Connected Players)
            OS_TPrintf("Friend[%d] '%s' : profileID %u, status %d, server (%d/%d)\n",
                       i, stGameCnt.friendList.playerName[i],
                       DWC_GetGsProfileId(&stUserData, &stGameCnt.friendList.keyList[i]),
                       stGameCnt.friendStatus[i],
                       numEntry, maxEntry);
        }
        else {
            OS_TPrintf("Friend[%d] '%s' : profileID %u, status %d\n",
                       i, stGameCnt.friendList.playerName[i],
                       DWC_GetGsProfileId(&stUserData, &stGameCnt.friendList.keyList[i]),
                       stGameCnt.friendStatus[i]);
        }

        // Display GameSpy Status character string
        OS_TPrintf("StatusData : ");
        for (j = 0; j < size; j++){
            OS_TPrintf("%02d, ", statusData[j]);
        }
        OS_TPrintf("\n");
    }
}


/*---------------------------------------------------------------------------*
  Callback function for making additional displays during friend add/delete mode
 *---------------------------------------------------------------------------*/
static void RegFriendModeDispCallback(int curIdx, void* param)
{
    RegFriendParam cntParam;
    int i;

    if (!param){
        // Called the first time with param = NULL
        cntParam.step = 0;
    }
    else {
        cntParam = *(RegFriendParam *)param;
    }

    OS_TPrintf("\n");

    if (cntParam.step == 0){
        // Display the friend roster during command selection
        DispFriendList(FALSE);
    }
    else {
        // Currently entering a numeric value
        if (curIdx == 0){
            // Register friend Currently entering a profile ID
            OS_TPrintf("Set profileID to friend list : %010u\n", cntParam.value);

            // Display an underline to indicate the numeric value currently being selected
            for (i = 0; i < 9-cntParam.figure; i++){
                OS_TPrintf(" ");
            }
            OS_TPrintf("                               -\n");
        }
        else {
            // Delete friend Currently entering a friend roster index
            DispFriendList(FALSE);  // Display the friend roster
            
            OS_TPrintf("Delete index : %02u\n", cntParam.value);

            for (i = 0; i < 1-cntParam.figure; i++){
                OS_TPrintf(" ");
            }
            OS_TPrintf("               -\n");
        }
    }
}


/*---------------------------------------------------------------------------*
  Callback function for additional display after login
 *---------------------------------------------------------------------------*/
static void LogonModeDispCallback(int curIdx, void* param)
{
#pragma unused(param)

    // Display the pointer to the number of player entries excluding oneself
    OS_TPrintf("\nnumber of entry = ");
    if (stNumEntry == 2){
        OS_TPrintf("   %d >>\n", stNumEntry);
    }
    else if (stNumEntry == DWC_MAX_CONNECTIONS){
        OS_TPrintf("<< %d\n", stNumEntry);
    }
    else {
        OS_TPrintf("<< %d >>\n", stNumEntry);
    }

    if (curIdx == 3){
        // when connecting to the server-client matchmaking server DS
        // the friend list index is also shown
        OS_TPrintf("GameServerIndex = ");
        if (DWC_GetNumFriend(stGameCnt.friendList.keyList, GAME_MAX_FRIEND_LIST) == 1){
            OS_TPrintf("   %d\n", stServerIdx);
        }
        else if (stServerIdx == 0){
            OS_TPrintf("   %d >>\n", stServerIdx);
        }
        else if (stServerIdx == GAME_MAX_FRIEND_LIST-1){
            OS_TPrintf("<< %d\n", stServerIdx);
        }
        else {
            OS_TPrintf("<< %d >>\n", stServerIdx);
        }
    }
}


/*---------------------------------------------------------------------------*
  Callback function for making an additional display during matchmaking mode
 *---------------------------------------------------------------------------*/
static void MatchModeDispCallback(int curIdx, void* param)
{
#pragma unused(curIdx)
#pragma unused(param)
    
}

/*---------------------------------------------------------------------------*
  Selected GUI
 *---------------------------------------------------------------------------*/
FS_EXTERN_OVERLAY(main_overlay_1);

static GameMode NetConfigMain(void)
{
    GameMode returnSeq = GAME_MODE_MAIN;

    (void)FS_LoadOverlay( MI_PROCESSOR_ARM9, FS_OVERLAY_ID(main_overlay_1) ) ;

    DWC_CleanupInet(); // Disconnect from the network.
    sPrintOverride = FALSE; // Temporarily restore OS_TPrintf() output to original status.
    dbs_DemoFinalize();

#if defined( USE_AUTHSERVER_RELEASE )
    DWC_SetAuthServer(DWC_CONNECTINET_AUTH_RELEASE);
#endif

#if defined( NETCONFIG_USE_HEAP )
    {
        void* work = OS_Alloc(DWC_UTILITY_WORK_SIZE);

        (void)DWC_StartUtility(work, DWC_LANGUAGE_JAPANESE, DWC_UTILITY_TOP_MENU_FOR_JPN);

        OS_Free(work);
    }
#else
    (void)DWC_StartUtilityEx(DWC_LANGUAGE_JAPANESE, 0);
#endif

    (void)FS_UnloadOverlay( MI_PROCESSOR_ARM9, FS_OVERLAY_ID(main_overlay_1) ) ;

    OS_SetIrqFunction(OS_IE_V_BLANK, VBlankIntr);
    (void)OS_EnableIrqMask(OS_IE_V_BLANK);
    (void)OS_EnableIrq();
    (void)GX_VBlankIntr( TRUE );

    dbs_DemoInit();
    sPrintOverride = TRUE; // OS_TPrintf() output is linked to the console
    OS_WaitVBlankIntr();              // Waiting the end of V-Blank interrupt
    GX_DispOn();
    GXS_DispOn();

    return returnSeq;
}

static GameMode StartIPMain(void)
{
	DWCApInfo apinfo;
	
    DWC_InitInet( &stConnCtrl );

#if defined( USE_AUTHSERVER_RELEASE )
    DWC_SetAuthServer(DWC_CONNECTINET_AUTH_RELEASE);
#endif

    DWC_ConnectInetAsync();
    // Wait for stable start.
    while ( !DWC_CheckInet() )
    {
        DWC_ProcessInet();

        // Wait for V-blank
        GameWaitVBlankIntr();
    }

    switch ( DWC_GetInetStatus() )
    {
    case DWC_CONNECTINET_STATE_ERROR:
        {
            // error display
            DWCError err;
            int errcode;
            err = DWC_GetLastError( &errcode );

            OS_TPrintf("   Error occurred %d %d.\n", err, errcode );
        }
        DWC_ClearError();

        DWC_CleanupInet();
        break;
    case DWC_CONNECTINET_STATE_FATAL_ERROR:
        {
            // error display
            DWCError err;
            int errcode;
            err = DWC_GetLastError( &errcode );

            OS_Panic("   Error occurred %d %d.\n", err, errcode );
        }
        break;

    case DWC_CONNECTINET_STATE_CONNECTED:
        OS_TPrintf("   CONNECTED!!!.\n");
		if(DWC_GetApInfo(&apinfo) == TRUE) {
			OS_TPrintf("DWCApInfo.aptype = %d\n", apinfo.aptype);
			OS_TPrintf("DWCApInfo.area   = %d\n", apinfo.area);
			OS_TPrintf("DWCApInfo.spotinfo = %s\n", apinfo.spotinfo);
			OS_TPrintf("DWCApInfo.essid = %s\n", apinfo.essid);
			OS_TPrintf("DWCApInfo.bssid = %02x:%02x:%02x:%02x:%02x:%02x\n", apinfo.bssid[0], apinfo.bssid[1], apinfo.bssid[2], apinfo.bssid[3], apinfo.bssid[4], apinfo.bssid[5]);
		}
		else {
			OS_TPrintf("Failed to get DWCApInfo\n");
		}
			
        break;
    }

    return GAME_MODE_MAIN;
}

static GameMode StopIPMain(void)
{
    DWC_CleanupInet();

    return GAME_MODE_MAIN;
}

static GameMode BackupUserdataMain( void )
{
    GameSequence *gameSeq = &gameSeqList[GAME_MODE_BACKUP];
    int curIdx = 0;
    GameMode returnSeq = GAME_MODE_NUM;

    while (1){
        dbs_Print( 0, 0, "s:%d", DWC_GetMatchingState() );
        dbs_Print( 7, 0, "n:%d", DWC_GetNumConnectionHost() );
        dbs_Print( 0, 1, "w:%d", DWC_GetLinkLevel() );
        dbs_Print( 10,1, "p:%d", stGameCnt.userData.profileID );
	    dbs_Print( 30, 0, "%c",   loopstr[((loopcnt)>>2)&3] );
		loopcnt++;

        ReadKeyData();  // Get key data

        HandleDWCError(&returnSeq);  // Error processing

        // Escapes the loop if the next mode to enter has been set
        if (returnSeq != GAME_MODE_NUM) break;

        ////////// Process the following key operations
        if (stKeyCnt.trg & PAD_BUTTON_A){
            // Select the menu with the A button
            BOOL end = FALSE;
            BOOL error = FALSE;
            int last_result;
            CARD_LockBackup((u16)card_lock_id);
            end = CARD_IdentifyBackup(stBackupTypeTable[stNumBackup].type);
            if(!end)
            {
                error = TRUE;
                last_result = CARD_GetResultCode();
            }
            CARD_UnlockBackup(card_lock_id);


            switch (curIdx){
            case 0:  // Load
                if ( error ) {
                    OS_TPrintf("Error occur with %d\n", last_result);
                    continue;
                }
                CARD_LockBackup((u16)card_lock_id);
                if (CARD_IsBackupEeprom())
                {
                    CARD_ReadEeprom(0, &stUserData, sizeof(stUserData) );
					CARD_ReadEeprom(sizeof(stUserData), &stGameCnt.friendList, sizeof(GameFriendList));
                }
                else if (CARD_IsBackupFlash())
                {
                    CARD_ReadFlash(0, &stUserData, sizeof(stUserData) );
					CARD_ReadFlash(sizeof(stUserData), &stGameCnt.friendList, sizeof(GameFriendList));
                }
                else if (CARD_IsBackupFram())
                {
                    CARD_ReadFram(0, &stUserData, sizeof(stUserData) );
					CARD_ReadFram(sizeof(stUserData), &stGameCnt.friendList, sizeof(GameFriendList));
                }
                CARD_UnlockBackup(card_lock_id);
                last_result = CARD_GetResultCode();

                if (last_result != CARD_RESULT_SUCCESS)
                {
                    OS_TPrintf("Error occur with %d\n", last_result);
                    continue;
                }
                else
                {
                    OS_TPrintf("Load OK!\n");
                    returnSeq = GAME_MODE_MAIN;
                }
                break;
            case 1:  // Save
                if ( error ) {
                    OS_TPrintf("Error occur with %d\n", last_result);
                    continue;
                }
                CARD_LockBackup((u16)card_lock_id);
                if (CARD_IsBackupEeprom())
                {
                    CARD_WriteAndVerifyEeprom(0, &stUserData, sizeof(stUserData) );
					CARD_WriteAndVerifyEeprom(sizeof(stUserData), &stGameCnt.friendList, sizeof(GameFriendList));
                }
                else if (CARD_IsBackupFlash())
                {
                    CARD_WriteAndVerifyFlash(0, &stUserData, sizeof(stUserData) );
					CARD_WriteAndVerifyFlash(sizeof(stUserData), &stGameCnt.friendList, sizeof(GameFriendList));
                }
                else if (CARD_IsBackupFram())
                {
                    CARD_WriteAndVerifyFram(0, &stUserData, sizeof(stUserData) );
					CARD_WriteAndVerifyFram(sizeof(stUserData), &stGameCnt.friendList, sizeof(GameFriendList));
                }
                CARD_UnlockBackup(card_lock_id);
                last_result = CARD_GetResultCode();

                if (last_result != CARD_RESULT_SUCCESS)
                {
                    OS_TPrintf("Error occur with %d\n", last_result);
                    continue;
                }
                else
                {
                    OS_TPrintf("Save OK!\n");
                    returnSeq = GAME_MODE_MAIN;
                }
                break;
            case 2: // report
                DWC_ReportUserData( &stUserData );
                break;
            case 3:
                returnSeq = GAME_MODE_MAIN;
                break;
            }
        }
        else if (stKeyCnt.trg & PAD_KEY_UP){
            // Move the cursor using up on the crosskey
            curIdx--;
            if (curIdx < 0) curIdx = gameSeq->numMenu-1;
            // Redisplay menu list
            DispMenuMsgWithCursor(gameSeq, curIdx, NULL);
        }
        else if (stKeyCnt.trg & PAD_KEY_DOWN){
            // Move the cursor using down on the crosskey
            curIdx++;
            if (curIdx >= gameSeq->numMenu) curIdx = 0;
            // Redisplay menu list
            DispMenuMsgWithCursor(gameSeq, curIdx, NULL);
        }
        else if (stKeyCnt.trg & PAD_KEY_LEFT){
            // Decrement the number of entries using the left button on the crosskey
            stNumBackup--;
            stNumBackup %= BACKUP_TABLE_MAX;

            DispMenuMsgWithCursor(gameSeq, curIdx, NULL);
        }
        else if (stKeyCnt.trg & PAD_KEY_RIGHT){
            // Increment the number of entries using the right button on the crosskey
            stNumBackup++;
            stNumBackup %= BACKUP_TABLE_MAX;

            DispMenuMsgWithCursor(gameSeq, curIdx, NULL);
        }
        ////////// Key operation processing ends here

        // Wait for V-blank
        GameWaitVBlankIntr();
    }

    return returnSeq;
}


/*---------------------------------------------------------------------------*
  Callback function for additional display after login
 *---------------------------------------------------------------------------*/
static void BackupModeDispCallback(int curIdx, void* param)
{
#pragma unused(param)
#pragma unused(curIdx)
    OS_TPrintf("\n BACKUP: %s\n", stBackupTypeTable[stNumBackup].comment);
}


/*---------------------------------------------------------------------------*
  Function for waiting for V-blank
 *---------------------------------------------------------------------------*/
static void GameWaitVBlankIntr(void)
{
    int i;

    dbs_DemoUpdate();

    // Change the number V-blank waits to match the assumed game frame
    for (i = 0; i < GAME_FRAME; i++){
        OS_WaitIrq(TRUE, OS_IE_V_BLANK);
    }

    Heap_Debug();
    dbs_DemoLoad();

    // Check for stack overflow
    OS_CheckStack(OS_GetCurrentThread());
}


/*---------------------------------------------------------------------------*
  V-blank interrupt function
 *---------------------------------------------------------------------------*/
static void VBlankIntr(void)
{

    OS_SetIrqCheckFlag(OS_IE_V_BLANK);
}
