#include <nitro.h>
#include <ninet/ip.h>
#include <dwc.h>

#include "main.h"
#include "init.h"
#include "utility.h"
#include "userdata.h"
#include "screen.h"

#include "dbs.h"

/** --------------------------------------------------------------------
  defines
  ----------------------------------------------------------------------*/
#define GAME_NAME        "dwctest" // Game name used in this sample
#define GAME_SECRET_KEY  "d4q9GZ"  // Secret key used in this sample
#define GAME_PRODUCTID   10824      // Product ID used in this sample

#define CB_CONNECT_SERVER 0        // the server-client matchmaking server DS
#define CB_CONNECT_CLIENT 1        // the server-client matchmaking client DS

/** --------------------------------------------------------------------
  globals
  ----------------------------------------------------------------------*/
KeyControl		g_KeyCtrl;

/** --------------------------------------------------------------------
  statics
  ----------------------------------------------------------------------*/
static u8 s_Work[ DWC_INIT_WORK_SIZE ] ATTRIBUTE_ALIGN( 32 );
static DWCFriendsMatchControl	s_FMCtrl;
static DWCInetControl			s_ConnCtrl;
static int  s_state;
static BOOL s_logined;
static BOOL s_keepmatch;
static BOOL s_matched;
static BOOL s_canceld;
static BOOL s_closed;
static int  s_matched_frame;
static int  s_profileID = 0;
static u8   s_entry = 0;

const int	SIZE_SEND_BUFFER = 512;
const int	SIZE_RECV_BUFFER = 4 * 1024;

static u8	s_send_buf[ SIZE_SEND_BUFFER ];
static u8	s_recv_buf[ DWC_MAX_CONNECTIONS ][ SIZE_RECV_BUFFER ];

/** --------------------------------------------------------------------
  static functions
  ----------------------------------------------------------------------*/
static int update_init( void );
static int update_connect( void );
static int update_login( void );
static int update_online( void );
static int update_server( void );
static int update_client( void );
static int update_matching( void );
static int update_matched( void );
static int update_close( void );
static int update_disconnect( void );
static int update_save( void );
static int update_error( void );

static BOOL InputNumber( u8* number );
static void DispMenuMsg( const u8 figure, const u8 number );

static void cb_login(DWCError error, int profileID, void* param);
static void cb_updateServers(DWCError error, BOOL isChanged, void* param);
static void cb_connect(DWCError error, BOOL cancel, BOOL self, BOOL isServer, int index, void* param);
static void cb_close(DWCError error, BOOL isLocal, BOOL isServer, u8  aid, int index, void* param);

static void DispMessage( const char* main_msg, const char* msg_a, const char* msg_b, const char* msg_x, const char* msg_y,  const char* msg_l );

/*---------------------------------------------------------------------------*
  V-blank interrupt function
 *---------------------------------------------------------------------------*/
static void VBlankIntr(void)
{
    //---- upload pseudo screen to VRAM
    DC_FlushRange(gScreen, sizeof(gScreen));
    GX_LoadBG0Scr(gScreen, 0, sizeof(gScreen));

    OS_SetIrqCheckFlag(OS_IE_V_BLANK);
}

/** --------------------------------------------------------------------
  Initialization processing
  ----------------------------------------------------------------------*/
static void
initFunctions( void )
{
	OS_Init();
    OS_InitTick();
    OS_InitAlarm();
    RTC_Init();
    CARD_Init();
    OS_SetThreadStackWarningOffset(OS_GetCurrentThread(), 0x100);
    PXI_Init();
    GX_Init();
    FS_Init( MI_DMA_MAX_NUM ); 

    GX_DispOff();
    GXS_DispOff();

    // Enable V-blank interrupt
    OS_SetIrqFunction(OS_IE_V_BLANK, VBlankIntr);
    (void)OS_EnableIrqMask(OS_IE_V_BLANK);
    (void)OS_EnableIrq();
    (void)OS_EnableInterrupts();
    (void)GX_VBlankIntr(TRUE);
	GX_SetDispSelect(GX_DISP_SELECT_SUB_MAIN);

    Heap_Init();
    Heap_SetDebug( TRUE );

    // Initialize output
    InitScreen();
    dbs_DemoInit();
    DTDB_SetPrintOverride( TRUE );

    // Initialize key input
    InitKeyData( &g_KeyCtrl );

    // Starts display
    OS_WaitVBlankIntr();              // Waiting the end of V-Blank interrupt
    GX_DispOn();
    GXS_DispOn();

}

/** --------------------------------------------------------------------
  function called once per frame
  ----------------------------------------------------------------------*/
static void
update( void )
{
    int errcode;
    DWCError err;
    DWCErrorType errtype;

    switch ( s_state )
    {
    case state_init:			s_state = update_init();			break;
    case state_connect:			s_state = update_connect();			break;
    case state_login:			s_state = update_login();			break;
    case state_online:			s_state = update_online();			break;
    case state_server:          s_state = update_server();          break;
    case state_client:          s_state = update_client();          break;
    case state_matching:		s_state = update_matching();		break;
    case state_matched:			s_state = update_matched();			break;
    case state_close:			s_state = update_close();			break;
    case state_disconnect:		s_state = update_disconnect();		break;
    case state_save:            s_state = update_save();            break;
    case state_error:			s_state = update_error();			break;
    }

#if 1
    err = DWC_GetLastErrorEx(&errcode, &errtype);

    if ( err != DWC_ERROR_NONE && s_state != state_error )
    {
        // Here, just change the status to "waiting to resolve error," if an error occurs.
        OS_TPrintf( "\n*******************************\n" );
        OS_TPrintf( "ERROR : %d : %d\n", err, errcode );
        OS_TPrintf( "Push START BUTTON to continue\n" );
        OS_TPrintf( "*******************************\n\n" );
        s_state = state_error;
    }

#else
    err = DWC_GetLastErrorEx(&errcode, &errtype);
    if ( err != DWC_ERROR_NONE )
    {
        OS_TPrintf("ERROR: %d:%d\n", err, errcode);
        DWC_ClearError();
        
        if ( errtype == DWC_ETYPE_DISCONNECT )
        {
            // When disconnecting, run the "FriendsMatch" end operation and Internet connection Cleanup.
            // Runs Cleanup
            DWC_ShutdownFriendsMatch();
            s_state = state_disconnect;
        }
        else if ( errtype == DWC_ETYPE_FATAL )
        {
            // This is a Fatal Error, so end
            OS_Panic("++FATAL ERROR\n");
        }
        else if ( errtype == DWC_ETYPE_LIGHT )
        {
            // For small errors, don't do anything and go back to online status.
            DTUD_DispFriendList();
            s_state = state_online;
        }
        else {
            // Once only the FriendsMatch operation is completed, run initialization of FriendsMatch again.
            pushResult( err );
            DWC_ShutdownFriendsMatch();
            s_state = state_connect;
        }
    }
#endif
}

/** --------------------------------------------------------------------
  Connect to the Internet.
  ----------------------------------------------------------------------*/
static int
update_init( void )
{
    // Does the user data on the DS device coincide with that on the DS card?
    if ( DWC_CheckValidConsole( DTUD_GetUserData() ) )
    {
        DispMessage( "STATE : CONNECTING", NULL, NULL, NULL, NULL, NULL);
        
        DWC_InitInet( &s_ConnCtrl );
        
        DWC_SetAuthServer( DWC_CONNECTINET_AUTH_RELEASE );
        
        DWC_ConnectInetAsync();
        
        return state_connect;
    }
    else
    {
        // IF it does not coincide, a connection to the Internet cannot be made.
        OS_TPrintf( "UserData is not Valid Console.\n" );
        return state_init;
    }

    return state_init;
}

/** --------------------------------------------------------------------
  connecting
  ----------------------------------------------------------------------*/
static int
update_connect( void )
{
    DWC_ProcessInet();

    if ( DWC_CheckInet() )
    {
        int status;

        status = DWC_GetLastError(NULL);

        if ( status == DWC_ERROR_NONE )
        {

            OS_TPrintf("Connected!..start FriendsMatch.\n");
            
            s_logined = FALSE;
            
            DWC_InitFriendsMatch(&s_FMCtrl, DTUD_GetUserData(),
                                 GAME_PRODUCTID, GAME_NAME, GAME_SECRET_KEY,
                                 0, 0,
                                 DTUD_GetFriendList(), FRIEND_LIST_LEN );
            
            // Login using the authentication function
            DWC_LoginAsync(L"Name", NULL, cb_login, NULL);
            
            return state_login;
            
        }
        else
        {
            OS_TPrintf("Error\n");
            return state_disconnect;
        }
    }

    return state_connect;
}

/** --------------------------------------------------------------------
  During login
  ----------------------------------------------------------------------*/
int
update_login( void )
{
    u8 i;
    
    DWC_ProcessFriendsMatch();  // Update DWC communication processing

    if ( s_logined )
    {
        // Gets the result of the ingamesn check.
	    if( DWC_GetIngamesnCheckResult() == DWC_INGAMESN_INVALID )
        {
            // If an inappropriate name is determined, a special operation is necessary.
		    OS_TPrintf("BAD ingamesn is detected by NAS.\n");
        }
        
        OS_TPrintf("================================\n");

        s_keepmatch = FALSE;
        s_matched   = FALSE;
        s_canceld   = FALSE;
        s_closed    = FALSE;

        DTUD_DispFriendList();

        // At this point, it sets the connection close callback
        DWC_SetConnectionClosedCallback(cb_close, NULL);

        // Also sets the receive buffer
        for ( i = 0; i < DWC_MAX_CONNECTIONS; ++i )
        {
            DWC_SetRecvBuffer( i, &s_recv_buf[i], SIZE_RECV_BUFFER );
        }

        // Login successful
        DispMessage( "STATE : ONLINE", "connect to server", "shutdown", "setup game server", "friend state", "save firned data");

        return state_online;
    }

    return state_login;
}

/** --------------------------------------------------------------------
  Online
  ----------------------------------------------------------------------*/
int
update_online( void )
{
    BOOL changed = FALSE;
    
    DWC_ProcessFriendsMatch();  // Update DWC communication processing

    if ( g_KeyCtrl.trg & PAD_BUTTON_B )
    {
        // Pressing the B button cuts the connection, and returns to the initial state.
        DWC_ShutdownFriendsMatch();
        return state_disconnect;
    }
    else if ( g_KeyCtrl.trg & PAD_BUTTON_A )
    {
        // Pressing the A button goes to client mode.
        s_matched = FALSE;
        DTUD_DispFriendList();
        return state_client;
    }
    else if ( g_KeyCtrl.trg & PAD_BUTTON_X )
    {
        // Pressing the X button goes to server mode.
        s_matched = FALSE;
        DTUD_DispFriendList();
        return state_server;
    }
    else if ( g_KeyCtrl.trg & PAD_BUTTON_Y )
    {
        // Pressing the Y button displays friend information.
        DTUD_DispFriendList();
    }
    else if ( g_KeyCtrl.trg & PAD_BUTTON_START )
    {
        // Pressing the start button cuts the connection, and returns to the initial state.
        DWC_ShutdownFriendsMatch();
        return state_disconnect;
    }
    else if ( g_KeyCtrl.trg & PAD_BUTTON_L )
    {
        DispMessage( "SAVE DATA?", "save", "cancel", NULL, NULL, NULL );
        return state_save;
    }

    return state_online;
}

/** --------------------------------------------------------------------
  Startup of the server DS
  ----------------------------------------------------------------------*/
static int
update_server( void )
{
    static u8 s_entry;
    
    DWC_ProcessFriendsMatch();  // Update DWC communication processing

    DispMessage( "SETUP GAME SERVER", "set", "cancel", NULL, NULL, NULL );
    PrintString( 1, 14, 0xf, "MAX MATCHING NUMBER:");
    PrintString( 1, 15, 0xf, "Please input 2 or more.");
    
    if ( g_KeyCtrl.trg & PAD_BUTTON_B )
    {
        DispMessage( "STATE : ONLINE", "connect to server", "shutdown", "setup game server", "friend state", "save firned data" );
        // Cancel connection using the B button
        return state_online;
    }
    else if( InputNumber( &s_entry ) )
    {
        // The maximum number of players that can be connected must be 2 or more.
        if( s_entry > 1 ){
            DispMessage( "STATE : MATCHING", NULL, "cancel", NULL, NULL, NULL );
            
            DWC_SetupGameServer( s_entry,
                                 cb_connect, (void *)CB_CONNECT_SERVER,
                                 NULL, NULL );
            
            s_entry = 0;
            
            return state_matching;
        }
    }
    return state_server;
}

/** --------------------------------------------------------------------
  Startup of the client DS
  ----------------------------------------------------------------------*/
static int
update_client( void )
{
    static u8 s_entry;
    u8   maxEntry, numEntry;
    DWCFriendData* s_FriendList;

    DWC_ProcessFriendsMatch();  // Update DWC communication processing

    DispMessage( "CONNECT TO SERVER", "set", "cancel", NULL, NULL, NULL );
    PrintString( 1, 14, 0xf, "CONNECT FRIEND INDEX:");

    // Release with the B button when waiting for automatic matchmaking to restart.
    if ( g_KeyCtrl.trg & PAD_BUTTON_B )
    {
        // Cancel connection using the B button
        DispMessage( "STATE : ONLINE", "connect to server", "shutdown", "setup game server", "friend state", "save firned data" );
        return state_online;
    }
    else if( InputNumber( &s_entry ) )
    {
        s_FriendList = DTUD_GetFriendList();

        DWC_GetFriendStatusSC( &s_FriendList[s_entry], &maxEntry, &numEntry, NULL );
        if ( numEntry >= maxEntry )
        {
            // The server DS is already over the set number of players.
            OS_TPrintf( "GameServer is full!\n" );
            OS_TPrintf( "Do over again.\n" );
        }
        else
        {
            DispMessage( "STATE : MATCHING", NULL, "cancel", NULL, NULL, NULL );
        
            DWC_ConnectToGameServerAsync( s_entry,
                                          cb_connect, (void *)CB_CONNECT_CLIENT,
                                          NULL, NULL );
            s_entry = 0;
            
            return state_matching;
        }
    }
    return state_client;
}
        
/** --------------------------------------------------------------------
  Running matchmaking
  ----------------------------------------------------------------------*/
static int
update_matching( void )
{
    DWC_ProcessFriendsMatch();  // Update DWC communication processing

    if ( g_KeyCtrl.trg & PAD_BUTTON_B )
    {
        // Cancel matchmaking using the B button
        s_canceld = TRUE;
    }

    if ( s_matched )
    {
        DispMessage( "STATE : MATCHED", "send all", "shutdown", "my state", "friend state", NULL );

        s_matched_frame = 0;

        return state_matched;
    }
    else if ( s_canceld )
    {
        // If canceled, it returns to online status
            DispMessage( "STATE : ONLINE", "connect to server", "shutdown", "setup game server", "friend state", "save firned data" );
        OS_TPrintf("Canceled connection ... retry\n");

        DWC_CancelMatching();

        s_canceld = FALSE;

        DTUD_DispFriendList();

        return state_online;
    }

    return state_matching;
}

/** --------------------------------------------------------------------
  Matchmaking complete
  ----------------------------------------------------------------------*/
static int
update_matched( void )
{
    s_matched_frame++;

    DWC_ProcessFriendsMatch();  // Update DWC communication processing

    // Leaves immediately if the server DS closes and the state changes.
    if ( s_state != state_matched )
    {
        DTUD_DispFriendList();
        
        return s_state;
    }
    
    if ( g_KeyCtrl.trg & PAD_BUTTON_A )
    {
        // Runs Reliable send to all connection hosts with the A button
        DWC_SendReliableBitmap( DWC_GetAIDBitmap(), s_send_buf, SIZE_SEND_BUFFER );
    }
    else if ( g_KeyCtrl.trg & PAD_BUTTON_B )
    {
        // Closes the connection with the B button
        int result = DWC_CloseAllConnectionsHard();
        
        OS_TPrintf( "CLOSE! result %d.\n", result );
        
        DTUD_DispFriendList();
        DispMessage( "STATE : ONLINE", "connect to server", "shutdown", "setup game server", "friend state", "save firned data" );
        return state_online;
    }
    else if ( g_KeyCtrl.trg & PAD_BUTTON_X )
    {
        // Pressing the X button displays your own information.
        u8* pAidList;
        int num;
        int i;

        num = DWC_GetAIDList(&pAidList);
        for ( i = 0; i < num; i++ )
        {
            if ( pAidList[i] == DWC_GetMyAID() ) break;
        }
        OS_TPrintf( "[%d] aid:%d, num:%d\n", i, DWC_GetMyAID(), num );
    }
    else if ( g_KeyCtrl.trg & PAD_BUTTON_Y )
    {
        // Pressing the Y button displays friend information.
        DTUD_DispFriendList();
    }

    return state_matched;
}

/** --------------------------------------------------------------------
  Closing
  ----------------------------------------------------------------------*/
static int
update_close( void )
{
    s_logined = FALSE;

    return state_login;
}

/** --------------------------------------------------------------------
  Disconnected.
  ----------------------------------------------------------------------*/
static int
update_disconnect( void )
{
    if ( DWC_CleanupInetAsync() )
    {
        return state_init;
    }

    return state_disconnect;
}

/** --------------------------------------------------------------------
  Saves data
  ----------------------------------------------------------------------*/
static int
update_save( void )
{
    DWC_ProcessFriendsMatch();  // Update DWC communication processing
    
    if ( g_KeyCtrl.trg & PAD_BUTTON_A )
    {
        if( DWC_CheckDirtyFlag( DTUD_GetUserData() ) ){
            DTUD_SaveUserData();
            OS_TPrintf( "Saved User Data.\n" );
        }
        
        DTUD_SaveFriendList();
        OS_TPrintf( "Saved Friend Data.\n" );

        DispMessage( "STATE : ONLINE", "connect to server", "shutdown", "setup game server", "friend state", "save firned data" );
        return state_online;
    }
    else if ( g_KeyCtrl.trg & PAD_BUTTON_B )
    {
        DispMessage( "STATE : ONLINE", "connect to server", "shutdown", "setup game server", "friend state", "save firned data" );
        return state_online;
    }
    return state_save;
}

/** --------------------------------------------------------------------
  Waiting for error release
  ----------------------------------------------------------------------*/
static int update_error( void )
{
    int errcode;
    DWCError err;
    DWCErrorType errtype;
    
    if ( g_KeyCtrl.trg & PAD_BUTTON_START )
    {
        // Releases errors with the start button
        err = DWC_GetLastErrorEx(&errcode, &errtype);

        DWC_ClearError();
        
        if ( errtype == DWC_ETYPE_DISCONNECT )
        {
            // When disconnecting, run the "FriendsMatch" end operation and Internet connection Cleanup.
            // Runs Clearup
            DWC_ShutdownFriendsMatch();
            return state_disconnect;
        }
        else if ( errtype == DWC_ETYPE_FATAL )
        {
            // This is a Fatal Error, so end
            OS_Panic("++FATAL ERROR\n");
            return state_init;  // not reached;
        }
        else if ( errtype == DWC_ETYPE_LIGHT )
        {
            // For small errors, don't do anything and go back to online status.
            DTUD_DispFriendList();
            return state_online;
        }
        else {
            // Once only the FriendsMatch operation is completed, run initialization of FriendsMatch again.
            DWC_ShutdownFriendsMatch();
            return state_connect;
        }
    }

    return state_error;
}

/** --------------------------------------------------------------------
  ----------------------------------------------------------------------*/
/** --------------------------------------------------------------------
  Callback function
  ----------------------------------------------------------------------*/
static void
cb_login(DWCError error, int profileID, void* param)
{
    (void)param;
    if (error == DWC_ERROR_NONE)
    {
        // Get authentication success status and profile ID
        OS_TPrintf("Login succeeded. profileID = %u\n\n", profileID);

        s_profileID = profileID;  // Records the profile ID.

        // Check whether or not stUserData has been updated.
        if ( DTUD_CheckUpdateUserData() )
        {
            DTUD_SaveUserData();
        }

        // Synchronous processing of the friend register and the buddy list on the GameSpy server DS
        DWC_UpdateServersAsync(NULL,
                               cb_updateServers, NULL,
                               NULL, NULL,
                               NULL, NULL);
    }
    else
    {
        OS_TPrintf("Error\n");
    }
}

static void
cb_updateServers(DWCError error, BOOL isChanged, void* param)
{
    (void)param;
    (void)isChanged;

    if ( error == DWC_ERROR_NONE )
    {
        // Successful synchronous processing of friend roster.
        OS_TPrintf( "Update Servers succeeded.\n" );
        
        s_logined = TRUE;
    }
    else
    {
        OS_TPrintf("Error\n");
    }
}

static void
cb_connect(DWCError error, BOOL cancel, BOOL self, BOOL isServer, int index, void* param)
{

    if ( (int)param == CB_CONNECT_SERVER )
    {
        // Server DS
        OS_TPrintf( "connected callback : SERVER\n" );
    }
    else if ( (int)param == CB_CONNECT_CLIENT )
    {
        // Client DS
        OS_TPrintf( "connected callback : CLIENT\n" );
    }
    OS_TPrintf( "err:%d, cancel:%d, self:%d, srv:%d, idx:%d\n",
                error, cancel, self, isServer, index );

    if ( error == DWC_ERROR_NONE )
    {
        if ( !cancel )
        {
            // Connection successful
            s_matched = TRUE;
        }
        else if ( self || isServer )
        {
            // You canceled matchmaking, or the server DS canceled matchmaking when you are the client.
            // the server DS canceled matchmaking.
            s_canceld = TRUE;
        }
        // Even if a newly connected client cancels matchmaking, nothing happens.
    }
    else
    {
        OS_TPrintf("Error\n");
    }

    s_keepmatch = FALSE;
}

static void
cb_close(DWCError error, BOOL isLocal, BOOL isServer, u8  aid, int index, void* param)
{
    (void)param;

    OS_TPrintf( "Closed Callback err:%d, local:%d, srv:%d, aid:%d, idx:%d\n",
               error, isLocal, isServer, aid, index );

    if ( error == DWC_ERROR_NONE )
    {
        if ( !isLocal && isServer )
        {
            // The server DS closed the connection with the server when you are a client.
            // Since the server DS has closed, the local host also closes.
            s_closed = TRUE;
            DWC_CloseAllConnectionsHard();
        }
        
        if ( s_closed && DWC_GetNumConnectionHost() == 1 ){
            // When all connections are closed due to the above close operation.
            s_closed = FALSE;
            s_state  = state_online;
            DispMessage( "STATE : ONLINE", "connect to server", "shutdown", "setup game server", "friend state", "save firned data" );
        }
    }
    else
    {
        OS_TPrintf("Error\n");
    }
}

/** -------------------------------------------------------------------------
  ---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*
  Name:         InputNumber

  Description:  performs friends registration key numerical value input

  Arguments:    maxFigure - Maximum number of input digits

  Returns:      If input is complete: returns TRUE,
                if input is incomplete: returns FALSE.
 *---------------------------------------------------------------------------*/
static BOOL InputNumber( u8* number )
{
    static u8 figure = 0;
    u8 maxFigure = 2;
    u8 num;

    num = *number;

    DispMenuMsg( figure, num );
    
    if (g_KeyCtrl.trg & PAD_BUTTON_A)
    {
        figure = 0;
        return TRUE;
    }
    else if ( g_KeyCtrl.trg & PAD_KEY_UP )
    {
        // Increment numeric value during selection using the up/down buttons on the crosskey
        if ( (num/UtilPow(10, figure) )%10 == 9)
        {
            num -= UtilPow(10, figure)*9;
        }
        else
        {
            num += UtilPow(10, figure);
        }
        // Displays numerical values during input
        DispMenuMsg( figure, num );
    }
    else if ( g_KeyCtrl.trg & PAD_KEY_DOWN )
    {
        // Increment numeric value during selection using the up/down buttons on the crosskey
        if ( (num/UtilPow(10, figure) )%10 == 0)
        {
            num += UtilPow(10, figure)*9;
        }
        else
        {
            num -= UtilPow(10, figure);
        }
        // Displays numerical values during input
        DispMenuMsg( figure, num );
    }
    else if ( g_KeyCtrl.trg & PAD_KEY_LEFT )
    {
        // Move the number of selected digits using the left/right buttons on the crosskey
        if (figure < maxFigure-1) figure++;
        // Displays numerical values during input
        DispMenuMsg( figure, num );
    }
    else if ( g_KeyCtrl.trg & PAD_KEY_RIGHT )
    {
        // Move the number of selected digits using the left/right buttons on the crosskey
        if (figure > 0) figure--;
        // Displays numerical values during input
        DispMenuMsg( figure, num );
    }

    *number = num;
    return FALSE;
}

/*---------------------------------------------------------------------------*
  Name:         DispMenuMsg

  Description:  Displays the numerical value during friend registration

  Arguments:    figure    - number of digits during input
                maxFigure - maximum number of digits

  Returns:      NONE.
 *---------------------------------------------------------------------------*/
static void DispMenuMsg( const u8 figure, const u8 number )
{
    PrintString( 1, 17, 0xf, "Input Number:" );
    PrintString( 15, 17, 0xf, "%2.2u", number );

// Display an underline to indicate the numeric value currently being selected
    PrintString( (s16)(16-figure), 18, 0xf, "-" );
}

/*---------------------------------------------------------------------------*
  Name:         DispMessage

  Description:  Displays general messages

  Arguments:    main_msg - current mode
                msg_a    - operation when the A button is pressed
                msg_b    - operation when the B button is pressed
                msg_x    - operation when the X button is pressed
                msg_y    - operation when the Y button is pressed
                msg_l    - operation when the L button is pressed

  Returns:      NONE.
 *---------------------------------------------------------------------------*/
static void
DispMessage( const char* main_msg, const char* msg_a, const char* msg_b, const char* msg_x, const char* msg_y,  const char* msg_l )
{
    s16 x = 1;
    s16 y = 5;

    ClearScreen();
    
    PrintString( 0, 0, 0xf, "================================" );
    PrintString( 2, 1, 0xf, "*** SC MATCHING SAMPLE ***" );
    PrintString( 0, 2, 0xf, "================================" );
    
    if ( main_msg == NULL ) return;

    PrintString( 0, y, 0xf, "--------------------------------" );
    y++;
    // Displays a message
    while ( *main_msg != '\0' )
    {
        if ( *main_msg == '\n' )
        {
            x = 1;  y++;
        }
        else
        {
            PrintString( x, y, 0xf, "%c", *main_msg );
            x++;
        }
        main_msg++;
    }
    y++;
    PrintString( 0, y, 0xf, "--------------------------------" );

    y += 2;

    // Displays an explanation of the A button and explanations of the B, X, and Y buttons
    if ( msg_a != NULL )
    {
        PrintString( 1, y, 0xf, "A BUTTON:%s", msg_a );
    }
    y++;
    if ( msg_b != NULL )
    {
        PrintString( 1, y, 0xf, "B BUTTON:%s", msg_b );
    }
    y += 2;
    if ( msg_x != NULL )
    {
        PrintString( 1, y, 0xf, "X BUTTON:%s", msg_x );
    }
    y++;
    if ( msg_y != NULL )
    {
        PrintString( 1, y, 0xf, "Y BUTTON:%s", msg_y );
    }
    y += 2;
    if ( msg_l != NULL )
    {
        PrintString( 1, y, 0xf, "L BUTTON:%s", msg_l );
    }
}

/** --------------------------------------------------------------------
  Status display function called every frame.
  ----------------------------------------------------------------------*/
void
drawConsole( void )
{
    static int i = 0;
    char loop[] = "/|\\-";
    char state[16];
    int  matchType = DWC_GetLastMatchingType();

    if ( ( DWC_GetState() != DWC_STATE_CONNECTED ) &&
         ( DWC_GetState() != DWC_STATE_MATCHING ) )
    {
        strcpy( state, "**" );
    }
    else if ( matchType == DWC_MATCH_TYPE_SC_SV )
    {
        strcpy( state, "SV" );
    }
    else if ( matchType == DWC_MATCH_TYPE_SC_CL )
    {
        strcpy( state, "CL" );
        if ( DWC_GetState() == DWC_STATE_CONNECTED )
        {
            strcat( state, "-CONN" );
        }
        else if ( DWC_GetState() == DWC_STATE_MATCHING )
        {
            strcat( state, "-MATCH" );
        }
    }
    else
    {
        strcpy( state, "**" );
    }
    
    dbs_Print( 0, 0, "s:%s", state );
    dbs_Print( 13, 0, "n:%d", DWC_GetNumConnectionHost() );
    dbs_Print( 0, 1, "w:%d", DWC_GetLinkLevel() );
    dbs_Print( 4, 1, "ss:%d", s_state );
    dbs_Print( 13, 1, "p:%d", s_profileID );

    dbs_Print( 30,0, "%c", loop[((i)>>2)&3] );
    i++;
}

/** --------------------------------------------------------------------
  Entrypoint
  ----------------------------------------------------------------------*/
void NitroMain ()
{
    initFunctions();

    // Specify debug display level
    DWC_SetReportLevel((u32)( DWC_REPORTFLAG_ALL &
                              ~DWC_REPORTFLAG_QR2_REQ ));
    // Initialize DWC library
    DWC_Init( s_Work );

    // Reads user data
    DTUD_Init();
    // Set function for allocating memory
    DWC_SetMemFunc( DTDB_GetAllocFunc(), DTDB_GetFreeFunc() );

    while (1)
    {
        drawConsole();

        ReadKeyData();

        update();

        dbs_DemoUpdate();
        OS_WaitIrq(TRUE, OS_IE_V_BLANK);
        Heap_Debug();
        dbs_DemoLoad();
    }

    OS_Terminate();
}
