#include <nitro.h>
#include <ninet/ip.h>
#include <dwc.h>

#include "main.h"
#include "init.h"
#include "userdata.h"
#include "utility.h"
#include "screen.h"
#include "font.h"
#include "dbs.h"
#include "wc.h"

#include "mp.h"

/** --------------------------------------------------------------------
  define
  ----------------------------------------------------------------------*/
#define     PICTURE_FRAME_PER_GAME_FRAME    1

#define     DEFAULT_GGID            0x003fff11
#define     DEFAULT_CHAN            1
#define     NUM_MAX_CHILD           15

/** --------------------------------------------------------------------
  statics
  ----------------------------------------------------------------------*/
static u16* s_SendBuf;
static DWCFriendData* s_FriendList;
//Data receive flag
static BOOL s_RecvFlag[NUM_MAX_CHILD];

/** --------------------------------------------------------------------
  static function
  ----------------------------------------------------------------------*/
//MP communication mode
static u8 ModeSelect(void);             // Parent/child select
static void ModeError(void);            // error display
static u8 ModeWorking(void);            // Busy
static void ModeParent(u16* gSendBuf, DWCFriendData* friendlist);  // Parent communication
static void ModeChild(u16* gSendBuf, DWCFriendData* friendlist);   // Child communication

static void PrintSharedData(u16 bitmap, DWCFriendData* friendlist);// Displays shared data

//Initial parameter for parent during MP communications
static WMParentParam defaultParameter ATTRIBUTE_ALIGN(32) = {
    NULL, 0, 0,
    DEFAULT_GGID,                             // Game group ID
    0x0000,                                   // Temporary ID (incremented by 1 each time)
    1,                                        // Entry permit flag
    NUM_MAX_CHILD,                            // Maximum number of connected children
    0,                                        // Multiboot flag
    0,                                        // Key sharing
    0,                                        // Continuous send flag
    100,                                      // Beacon interval
    {0, 0, 0, 0},                             // User name
    {0, 0, 0, 0, 0, 0, 0, 0},                 // Game name
    DEFAULT_CHAN,                             // Connection channel
    sizeof(DWCFriendData)*(1+NUM_MAX_CHILD)+4,// Size of parent send data
    sizeof(DWCFriendData)                     // Size of child send data
};

/*---------------------------------------------------------------------------*
  Name:         MP_SetData

  Description:  Runs send data initialization.

  Arguments:    NONE.

  Returns:      NONE.
 *---------------------------------------------------------------------------*/
void MP_SetData(u16* sendbuf, DWCFriendData* friendlist)
{
    s_SendBuf = sendbuf;
    s_FriendList = friendlist;

    // initialization of the receive flag
    MI_CpuClear8( s_RecvFlag, sizeof(s_RecvFlag) );

    // WM initialization
    WcInit(&defaultParameter, TRUE, FALSE, PICTURE_FRAME_PER_GAME_FRAME);

}

/*---------------------------------------------------------------------------*
  Name:         MP_Match

  Description:  Runs send / receive of data in MP

  Arguments:    NONE.

  Returns:      If communications end or are canceled: TRUE
                If communications are being made: returns FALSE.
 *---------------------------------------------------------------------------*/
BOOL MP_Match()
{
    static u8 child_flag = 0;//Communications control flag

    // Direct to process based on communication status
    switch (WcGetStatus())
    {
    case WC_STATUS_READY:// Parent/child selection
        if(child_flag){// Ends communications of a child device disconnected from a parent.
            child_flag = 0;
            WcEnd();
            return TRUE;
            break;
        }
        ModeSelect();
        break;
    case WC_STATUS_ERROR:
        ModeError();
        break;
    case WC_STATUS_BUSY:// Currently executing
        ModeWorking();
        break;
    case WC_STATUS_PARENT:// For parent
        ModeParent(s_SendBuf, s_FriendList);
        break;
    case WC_STATUS_CHILD:// For child
        ModeChild(s_SendBuf, s_FriendList);
        child_flag = 1;// Measures to return to Ready state when disconnected from parent
        break;
    }
    if (g_KeyCtrl.trg & PAD_BUTTON_B){//End communication
        child_flag = 0;
        WcEnd();
        return TRUE;
    }

    return FALSE;
}

/*---------------------------------------------------------------------------*
  Name:         MP_Finish

  Description:  Ends MP communication.

  Arguments:    NONE.

  Returns:      NONE.
 *---------------------------------------------------------------------------*/
void MP_Finish()
{
    // Quit WM
    WcFinish();
}

/*---------------------------------------------------------------------------*
  Name:         ModeSelect

  Description:  Process in parent/child selection screen

  Arguments:    None.

  Returns:      0: Select mode
                1: Cancel
 *---------------------------------------------------------------------------*/
static u8 ModeSelect(void)
{
    // Character display
    DispMessage( "MP MODE SELECT", NULL, "MP Cancel", "Parent Mode", "Child Mode", NULL);

    if (g_KeyCtrl.trg & PAD_BUTTON_X)
    {
        // Start parent communication
        WcStartParent();
        return 0;
    }
    else if (g_KeyCtrl.trg & PAD_BUTTON_Y)
    {
        // Start child communication
        WcStartChild();
        return 0;
    }

    return 0;
}

/*---------------------------------------------------------------------------*
  Name:         ModeError

  Description:  Processing in error display screen

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
static void ModeError(void)
{
    // Character display
    DispMessage( "ERROR!\n Fatal error occured.\n Please reboot program.", NULL, NULL, NULL, NULL, NULL);
}

/*---------------------------------------------------------------------------*
  Name:         ModeWorking

  Description:  Processing in busy screen

  Arguments:    None.

  Returns:      0: Currently executing
                1: End communication
 *---------------------------------------------------------------------------*/
static u8 ModeWorking(void)
{
    // Character display
    DispMessage( "NOW WORKING...", NULL, "MP Cancel", NULL, NULL, NULL);

    return 0;
}

/*---------------------------------------------------------------------------*
  Name:         ModeParent

  Description:  Processing in parent communications screen.

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
static void ModeParent(u16* gSendBuf, DWCFriendData* friendlist)
{
    BOOL result;
    u16 bitmap;

    // Synchronize data sharing
    result = WcStepDataSharing((void *)gSendBuf, &bitmap);

    // Character display
    DispMessage( "PARENT MODE", NULL, "MP Cancel", NULL, NULL, NULL);
    // Display shared data
    if (result)
    {
        PrintSharedData(bitmap, friendlist);
    }
    else
    {
        // Try StepDataSharing also in the next frame.
    }
}

/*---------------------------------------------------------------------------*
  Name:         ModeChild

  Description:  Processing in child communications screen.

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
static void ModeChild(u16* gSendBuf, DWCFriendData* friendlist)
{
    BOOL result;
    u16 bitmap;

    // Synchronize data sharing
    result = WcStepDataSharing((void *)gSendBuf, &bitmap);

    // Character display
    DispMessage( "CHILD MODE", NULL, "MP Cancel", NULL, NULL, NULL);
    // Display shared data
    if (result)
    {
        PrintSharedData(bitmap, friendlist);
    }
    else
    {
        // Try StepDataSharing also in the next frame.
    }
}

/*---------------------------------------------------------------------------*
  Name:         PrintSharedData

  Description:  Displays shared data from each terminal

  Arguments:    bitmap - Specifies ID bitmap indicating terminal receiving data

  Returns:      None.
 *---------------------------------------------------------------------------*/
static void PrintSharedData(u16 bitmap, DWCFriendData* friendlist)
{
    u16 i;
    u8 *p;
    int friendIdx;
    u8 friend_number = 0;


    // Loop for maximum number of parent device + child devices
    for (i = 0; i < (1 + WM_NUM_MAX_CHILD); i++)
    {
        if (bitmap & (0x0001 << i))
        {
            // Get data from terminal with ID of 'i'
            p = WcGetSharedDataAddress(i);

            if (p != NULL)
            {
                friend_number++;
                if(!s_RecvFlag[i] && WcGetAid() != i){// Get friend data excluding self
                    friendIdx = DTUD_GetAvailableFriendListIndex();
                    s_RecvFlag[i] = 1;
                    MI_CpuCopy8(p, &friendlist[friendIdx], sizeof(DWCFriendData));
                }
//                PrintString(2, (s16)(15 + i), 0x4, "Friend %d MP Data Received!!", i);
            }
        }
        else{
            s_RecvFlag[i] = 0;
        }
    }
    PrintString(2, 15, 0x4, "Changed Number of Friend %d", friend_number-1);
    // Change color only for data for this device
//    ColorString(2, (s16)(15 + WcGetAid()), 25, 0x1);
}
