#include <nitro.h>
#include <ninet/ip.h>
#include <dwc.h>

#include "main.h"
#include "init.h"
#include "utility.h"
#include "dbs.h"

#include "gamemain.h"

extern GameSequence gameSeqList[GAME_MODE_NUM];
extern GameControl stGameCnt;  // Game control data structure
extern KeyControl stKeyCnt;    // Key input control structure
extern DWCFriendsMatchControl stDwcCnt;    // DWC control structure

extern DWCUserData stUserData;
extern GameSaveData saveData;

static volatile int stNameCbLevel;  // Number of instances of waiting to get a player name

static void LoginCallback(DWCError error, int profileID, void* param);
static void UpdateServersCallback(DWCError error, BOOL isChanged, void* param);
static void FriendStatusCallback(int index, u8 status, const char* statusString, void* param);
static void DeleteFriendListCallback(int deletedIndex, int srcIndex,void* param);
static void SaveToServerCallback(BOOL success, BOOL isPublic, void* param);
static void LoadFromServerCallback(BOOL success, int profileID, char* data, int len, void* param);

/*---------------------------------------------------------------------------*
  Main function when offline
 *---------------------------------------------------------------------------*/
GameMode GameMain(void)
{
    GameSequence *gameSeq = &gameSeqList[GAME_MODE_MAIN];
    GameMode returnSeq = GAME_MODE_NUM;
    int curIdx = 0;

    while (1){
        dbs_Print( 0, 0, "s:%d", DWC_GetMatchingState() );
        dbs_Print( 7, 0, "n:%d", DWC_GetNumConnectionHost() );
        dbs_Print( 0, 1, "w:%d", DWC_GetLinkLevel() );
        dbs_Print( 10,1, "p:%d", stGameCnt.userData.profileID );
        
        ReadKeyData();  // Get key data

        DWC_ProcessFriendsMatch();  // Update DWC communication processing

        if (DWC_GetLastError(NULL)){
            // process when login/friend list synchronization fails
            ShutdownInet();  // Network disconnection
            DWC_ClearError();
            stGameCnt.blocking = FALSE;

            returnSeq = GAME_MODE_MAIN;
        }

        if (stGameCnt.blocking){
            // Prohibit key operations during DWC processing
            dbs_DemoUpdate();
            //SVC_WaitVBlankIntr();
            OS_WaitIrq(TRUE, OS_IE_V_BLANK);
            Heap_Debug();
            dbs_DemoLoad();

            // Check for stack overflow
            OS_CheckStack(OS_GetCurrentThread());
            continue;
        }

        // Authentication complete if the next mode to go to has been set
        if (returnSeq == GAME_MODE_LOGIN){
            // If login has been selected
            if (stGameCnt.userData.isValidProfile){
                // Escape from loop when the profile ID can be retrieved when authentication succeeded
                break;
            }
            else {
                // Cannot proceed to next mode if authentication fails
                returnSeq = GAME_MODE_NUM;
            }
        }
        else if (returnSeq != GAME_MODE_NUM){
            // Always escape from loop in other cases
            break;
        }
            
        ////////// Process the following key operations
        if (stKeyCnt.trg & PAD_BUTTON_A){
            // Select the menu with the A button
            returnSeq = gameSeq->menuList[curIdx].mode;
            stGameCnt.blocking = gameSeq->menuList[curIdx].blocking;

            switch (returnSeq){
            case GAME_MODE_LOGIN:  // Login

                // FriendsMatch Library initialization
                // The send/receive packet size of packets used by the library is set to default (8KB)
                //  
                DWC_InitFriendsMatch(&stDwcCnt, &stUserData, GAME_PRODUCTID, GAME_NAME,
                                     GAME_SECRET_KEY, 0, 0,
                                     stGameCnt.friendList.keyList, GAME_MAX_FRIEND_LIST);

                // Login using the authentication function
                //DWC_LoginAsync(stUserID, stPassword, LoginCallback, &returnSeq);
                DWC_LoginAsync(L"&#x201a;&#xc8;&#x201a;&#xdc;&#x201a;&#xa6;", NULL, LoginCallback, &returnSeq);
                break;
            default:
                break;
            }
        }
        else if (stKeyCnt.trg & PAD_KEY_UP){
            // Move the cursor using up on the crosskey
            curIdx--;
            if (curIdx < 0) curIdx = gameSeq->numMenu-1;
            // Redisplay menu list
            DispMenuMsgWithCursor(gameSeq, curIdx, NULL);
        }
        else if (stKeyCnt.trg & PAD_KEY_DOWN){
            // Move the cursor using down on the crosskey
            curIdx++;
            if (curIdx >= gameSeq->numMenu) curIdx = 0;
            // Redisplay menu list
            DispMenuMsgWithCursor(gameSeq, curIdx, NULL);
        }
        else if ( stKeyCnt.trg & PAD_BUTTON_Y){
            Heap_Dump();
        }
        ////////// Key operation processing ends here

        dbs_DemoUpdate();
        //SVC_WaitVBlankIntr();
        OS_WaitIrq(TRUE, OS_IE_V_BLANK);
        Heap_Debug();
        dbs_DemoLoad();
        // Check for stack overflow
        OS_CheckStack(OS_GetCurrentThread());
    }

    return returnSeq;
}

/*---------------------------------------------------------------------------*
  Login callback function
 *---------------------------------------------------------------------------*/
static void LoginCallback(DWCError error, int profileID, void *param)
{
    BOOL result;
    u64 friend_key;
    
    if (error == DWC_ERROR_NONE){
        // Get authentication success status and profile ID
        OS_TPrintf("Login succeeded. profileID = %u\n\n", profileID);
        stGameCnt.userData.profileID = profileID;
        stGameCnt.userData.isValidProfile = TRUE;

        // Gets the result of the ingamesn check.
        if( DWC_GetIngamesnCheckResult() == DWC_INGAMESN_INVALID )
        {
            // If an inappropriate name is determined, a special operation is necessary.
            OS_TPrintf("BAD ingamesn is detected by NAS.\n");
        }

        //Create friend registration key
        if( (friend_key = DWC_CreateFriendKey( &stUserData )) ){
            OS_TPrintf("friend_key = %lld\n", friend_key);
        }
        
        // begin friend list synchronization
        result = DWC_UpdateServersAsync(stGameCnt.userData.playerName,
                                        UpdateServersCallback, param,
                                        FriendStatusCallback, param,
                                        DeleteFriendListCallback, param);
        if (result == FALSE){
            // FALSE is returned only when this function is called in a status where it is okay to call this function (when login is not complete), otherwise TRUE is returned.
            //  
            OS_Panic("--- DWC_UpdateServersAsync error teminated.\n");
        }

        // Register a callback for making notification that storage save/load is complete
        DWC_SetStorageServerCallback(SaveToServerCallback,
                                     LoadFromServerCallback);
    }
    else {
        // Authentication failed
        DispErrorMessage();
        OS_TPrintf("Login failed. %d\n\n", error);

        // One more time from the beginning
        *(GameMode *)param = GAME_MODE_NUM;

        stGameCnt.blocking = FALSE;  // Cancel blocking
    }
}


/*---------------------------------------------------------------------------*
  friend list synchronization complete callback function
 *---------------------------------------------------------------------------*/
static void UpdateServersCallback(DWCError error, BOOL isChanged, void* param)
{
#pragma unused(param)
    int i;

    if (error == DWC_ERROR_NONE){
        // friend list synchronization completed
        OS_TPrintf("Updating servers succeeded.\n");

        if (isChanged){
            // Save the friend roster if it has been modified
            // (Should be saved)
            MI_CpuCopy32(&stGameCnt.friendList, &saveData.friendList, sizeof(GameFriendList));
        }

        // Get the name and communication status of a friend
        stNameCbLevel = 0;
        for (i = 0; i < GAME_MAX_FRIEND_LIST; i++){
            if (DWC_IsValidFriendData(&stGameCnt.friendList.keyList[i])){
                stGameCnt.friendStatus[i] =
                    DWC_GetFriendStatus(&stGameCnt.friendList.keyList[i], NULL);
    
                // Load player names
                if (DWC_LoadOthersDataAsync("\\dwc_name", i, (void *)i)){
                    stNameCbLevel++;
                }
            }
        }

        // Wait here until it is possible to get the friend's name
        while (stNameCbLevel){
            ReadKeyData();  // Get key data

            DWC_ProcessFriendsMatch();  // Update DWC communication processing

            dbs_DemoUpdate();
            OS_WaitIrq(TRUE, OS_IE_V_BLANK);
            Heap_Debug();
            dbs_DemoLoad();
        }

        // Display the friend roster
        DispFriendList(FALSE);
    }
    else {
        // Do nothing particular even if processing failed
        OS_TPrintf("Failed to update GameSpy servers. %d\n\n", error);
    }

    stGameCnt.blocking = FALSE;  // Cancel blocking
}

/*---------------------------------------------------------------------------*
  Callback function for notification of changes in friend status
 *---------------------------------------------------------------------------*/
static void FriendStatusCallback(int index, u8 status, const char* statusString, void* param)
{
#pragma unused(param)
#ifdef SDK_FINALROM
#pragma unused(index, status, statusString)
#endif

    OS_TPrintf("Friend[%d] changed status -> %d (statusString : %s).\n",
               index, status, statusString);
}


/*---------------------------------------------------------------------------*
  Friend roster delete callback
 *---------------------------------------------------------------------------*/
static void DeleteFriendListCallback(int deletedIndex, int srcIndex, void* param)
{
#pragma unused(param)
#ifdef SDK_FINALROM
#pragma unused(deletedIndex, srcIndex)
#endif
#if 0
    // As friend data is deleted and padded toward the front, corresponding data is also padded toward the front
    if (index < GAME_MAX_FRIEND_LIST-1){
        // Friend communication status is padded toward the front
        MI_CpuCopy8(&stGameCnt.friendStatus[index+1], &stGameCnt.friendStatus[index],
                    (u32)(GAME_MAX_FRIEND_LIST-index-1));

        // Friend player name is padded toward the front
        MI_CpuCopy8(&stGameCnt.friendList.playerName[index+1],
                    &stGameCnt.friendList.playerName[index],
                    (u32)(GAME_MAX_PLAYER_NAME*(GAME_MAX_FRIEND_LIST-index-1)));
    }

    stGameCnt.friendStatus[GAME_MAX_FRIEND_LIST-1] = DWC_STATUS_OFFLINE;
    MI_CpuClear32(&stGameCnt.friendList.playerName[GAME_MAX_FRIEND_LIST-1],
                  GAME_MAX_PLAYER_NAME);

#else
    OS_TPrintf("friend[%d] was deleted (equal friend[%d]).\n",
               deletedIndex, srcIndex);
#endif
}


/*---------------------------------------------------------------------------*
  Callback function for completing data save to storage server
 *---------------------------------------------------------------------------*/
static void SaveToServerCallback(BOOL success, BOOL isPublic, void* param)
{
#ifdef SDK_FINALROM
#pragma unused(success, isPublic, param)
#endif

    OS_TPrintf("Saved data to server. %08x\n", (u32)param);
    OS_TPrintf("result %d, isPublic %d.\n", success, isPublic);
}


/*---------------------------------------------------------------------------*
  Callback function for completing data load from storage server
 *---------------------------------------------------------------------------*/
static void LoadFromServerCallback(BOOL success, int profileID, char* data, int len, void* param)
{
#ifdef SDK_FINALROM
#pragma unused(profileID, len)
#endif

    OS_TPrintf("Loaded data from server. %08x\n", (u32)(param));
    OS_TPrintf("result %d, profileID %u, data '%s', len %d\n",
               success, profileID, data, len);

    if (success){
        // Get player name using the function for manipulating general-purpose key/value type text strings
        (void)DWC_GetCommonValueString("dwc_name",
                                       stGameCnt.friendList.playerName[(int)param],
                                       data, '\\');

        stNameCbLevel--;
    }
}


