#include <nitro.h>
#include <ninet/ip.h>
#include <dwc.h>

#include "main.h"
#include "init.h"
#include "utility.h"
#include "screen.h"
#include "font.h"
#include "dbs.h"
#include "wc.h"

#include "mp.h"

#define     PICTURE_FRAME_PER_GAME_FRAME    1

#define     DEFAULT_GGID            0x003fff11
#define     DEFAULT_CHAN            1
#define     NUM_MAX_CHILD           15

extern KeyControl stKeyCnt;    // Key input control structure

//MP communication mode
static u8 ModeSelect(void);             // Parent/child select
static void ModeError(void);            // error display
static u8 ModeWorking(void);            // Busy
static void ModeParent(u16* gSendBuf, DWCFriendData* friendlist);  // Parent communication
static void ModeChild(u16* gSendBuf, DWCFriendData* friendlist);   // Child communication

static void PrintSharedData(u16 bitmap, DWCFriendData* friendlist);// Displays shared data

//Initial parameter for parent during MP communications
static WMParentParam defaultParameter ATTRIBUTE_ALIGN(32) = {
    NULL, 0, 0,
    DEFAULT_GGID,                             // Game group ID
    0x0000,                                   // Temporary ID (incremented by 1 each time)
    1,                                        // Entry permit flag
    NUM_MAX_CHILD,                            // Maximum number of connected children
    0,                                        // Multiboot flag
    0,                                        // Key sharing
    0,                                        // Continuous send flag
    100,                                      // Beacon interval
    {0, 0, 0, 0},                             // User name
    {0, 0, 0, 0, 0, 0, 0, 0},                 // Game name
    DEFAULT_CHAN,                             // Connection channel
    sizeof(DWCFriendData)*(1+NUM_MAX_CHILD)+4,// Size of parent send data
    sizeof(DWCFriendData)                     // Size of child send data
};

//Data receive flag
static BOOL gRecvFlag[NUM_MAX_CHILD];

void mp_match(u16* gSendBuf, DWCFriendData* friendlist){
    u8 child_flag = 0, mp_flag = 0;//Communications control flag

    MI_CpuClear8( gRecvFlag, sizeof(gRecvFlag) );

    // WM initialization
    WcInit(&defaultParameter, TRUE, FALSE, PICTURE_FRAME_PER_GAME_FRAME);
    // Direct to process based on communication status
    while(!mp_flag){
        switch (WcGetStatus())
        {
        case WC_STATUS_READY:// Parent/child selection
            if(child_flag){
                mp_flag = 1;
                break;
            }
            mp_flag = ModeSelect();
            break;
        case WC_STATUS_ERROR:
            ModeError();
             break;
        case WC_STATUS_BUSY:// Currently executing
            mp_flag = ModeWorking();
            break;
        case WC_STATUS_PARENT:// For parent
            ModeParent(gSendBuf, friendlist);
            break;
        case WC_STATUS_CHILD:// For child
            ModeChild(gSendBuf, friendlist);
            child_flag = 1;//Measures to return to Ready state when disconnected from parent
            break;
        }
        ReadKeyData();  // Get key data
        if (stKeyCnt.trg & PAD_BUTTON_B){//End communication
        	WcEnd();
            mp_flag = 1;
        }

        dbs_DemoUpdate();
        OS_WaitIrq(TRUE, OS_IE_V_BLANK);
            Heap_Debug();
        dbs_DemoLoad();
        // Check for stack overflow
        OS_CheckStack(OS_GetCurrentThread());
    }
    // Quit WM
    WcFinish();
    ClearScreen();
}

/*---------------------------------------------------------------------------*
  Name:         ModeSelect

  Description:  Process in parent/child selection screen

  Arguments:    None.

  Returns:      0: Select mode
                1: Cancel
 *---------------------------------------------------------------------------*/
static u8 ModeSelect(void)
{
    // Clears the screen
    ClearScreen();
    // Character display
    PrintString(3, 8, 0xf,  "Push L to connect as PARENT");
    PrintString(3, 10, 0xf, "Push R to connect as CHILD");
    PrintString(3, 16, 0xf, "Push B to STOP");

    dbs_DemoUpdate();
    OS_WaitIrq(TRUE, OS_IE_V_BLANK);
        Heap_Debug();
    dbs_DemoLoad();
    // Check for stack overflow
    OS_CheckStack(OS_GetCurrentThread());
    
    while(1){
        ReadKeyData();  // Get key data
        if (stKeyCnt.trg & PAD_BUTTON_L)
        {
            // Start parent communication
            WcStartParent();
            return 0;
        }
        else if (stKeyCnt.trg & PAD_BUTTON_R)
        {
            // Start child communication
            WcStartChild();
            return 0;
        }
        else if (stKeyCnt.trg & PAD_BUTTON_B)
        {
            return 1;
        }
    }
    return 0;
}

/*---------------------------------------------------------------------------*
  Name:         ModeError

  Description:  Processing in error display screen

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
static void ModeError(void)
{
    // Clears the screen
    ClearScreen();
    // Character display
    PrintString(5, 10, 0x1, "======= ERROR! =======");
    PrintString(5, 13, 0xf, " Fatal error occured.");
    PrintString(5, 14, 0xf, "Please reboot program.");
}

/*---------------------------------------------------------------------------*
  Name:         ModeWorking

  Description:  Processing in busy screen

  Arguments:    None.

  Returns:      0: Currently executing
                1: End communication
 *---------------------------------------------------------------------------*/
static u8 ModeWorking(void)
{
    // Clears the screen
    ClearScreen();
    // Character display
    PrintString(9, 9, 0xf, "Now working...");
    PrintString(3, 16, 0xf, "Push B to STOP");

    ReadKeyData();  // Get key data
    if (stKeyCnt.trg & PAD_BUTTON_B)
    {
        // End communication
        WcEnd();
        return 1;
    }
    return 0;
}

/*---------------------------------------------------------------------------*
  Name:         ModeParent

  Description:  Processing in parent communications screen.

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
static void ModeParent(u16* gSendBuf, DWCFriendData* friendlist)
{
    BOOL result;
    u16 bitmap;

    // Synchronize data sharing
    result = WcStepDataSharing((void *)gSendBuf, &bitmap);

    // Clears the screen
    ClearScreen();
    // Character display
    PrintString(8, 1, 0x2, "Parent mode");
    // Display shared data
    if (result)
    {
        PrintSharedData(bitmap, friendlist);
    }
    else
    {
        // Try StepDataSharing also in the next frame.
    }
}

/*---------------------------------------------------------------------------*
  Name:         ModeChild

  Description:  Processing in child communications screen.

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
static void ModeChild(u16* gSendBuf, DWCFriendData* friendlist)
{
    BOOL result;
    u16 bitmap;

    // Synchronize data sharing
    result = WcStepDataSharing((void *)gSendBuf, &bitmap);

    // Clears the screen
    ClearScreen();
    // Character display
    PrintString(8, 1, 0x2, "Child mode");
    // Display shared data
    if (result)
    {
        PrintSharedData(bitmap, friendlist);
    }
    else
    {
        // Try StepDataSharing also in the next frame.
    }
}

/*---------------------------------------------------------------------------*
  Name:         PrintSharedData

  Description:  Displays shared data from each terminal

  Arguments:    bitmap - Specifies ID bitmap indicating terminal receiving data

  Returns:      None.
 *---------------------------------------------------------------------------*/
static void PrintSharedData(u16 bitmap, DWCFriendData* friendlist)
{
    u16 i;
    u8 *p;
    int friendIdx;


    // Loop for maximum number of parent device + child devices
    for (i = 0; i < (1 + WM_NUM_MAX_CHILD); i++)
    {
        if (bitmap & (0x0001 << i))
        {
            // Get data from terminal with ID of 'i'
            p = WcGetSharedDataAddress(i);

            if (p != NULL)
            {
                if(!gRecvFlag[i] && WcGetAid() != i){// Get friend data excluding self
                    friendIdx = GetAvailableFriendListIndex();
                    gRecvFlag[i] = 1;
                    MI_CpuCopy8(p, &friendlist[friendIdx], sizeof(DWCFriendData));
                }
                PrintString(2, (s16)(3 + i), 0x4, "Number%d Receive Success!!", i);
            }
            else
            {
                PrintString(2, (s16)(3 + i), 0x1, "xxxxxxxxxxxxxxxx");
            }
        }
        else{
            gRecvFlag[i] = 0;
        }
    }
    PrintString(3, 20, 0xf, "Push B to STOP");
    // Change color only for data for this device
    ColorString(2, (s16)(3 + WcGetAid()), 25, 0x1);
}
