#include <nitro.h>
#include <ninet/ip.h>
#include <dwc.h>

#include "main.h"
#include "init.h"
#include "utility.h"

#include "dbs.h"

#include "transport.h"

extern GameSequence gameSeqList[GAME_MODE_NUM];
extern GameControl stGameCnt;  // Game control data structure
extern KeyControl stKeyCnt;    // Key input control structure

//callback prototype
static void ConnectionClosedCallback(DWCError error,
                                     BOOL isLocal,
                                     BOOL isServer,
                                     u8 aid,
                                     int index,
                                     void* param);

/** -------------------------------------------------------------------------
  DWC Transport
  ---------------------------------------------------------------------------*/
#define			MAX_PLAYERS				3
const int		SIZE_SEND_BUFFER = 512;
const int		SIZE_RECV_BUFFER = 4 * 1024;

static u8		sSendBuffer[ SIZE_SEND_BUFFER ];
static u8		sRecvBuffer[ MAX_PLAYERS ][ SIZE_RECV_BUFFER ];

/** -------------------------------------------------------------------------
  A transmission completion callback.
  ---------------------------------------------------------------------------*/
static void
SendDoneCallback( int size, u8 aid )
{
#ifdef SDK_FINALROM
#pragma unused(size, aid)
#endif
	OS_TPrintf( "to aid = %d  size = %d\n", aid, size );
}

/** -------------------------------------------------------------------------
  Receive complete callback
  ---------------------------------------------------------------------------*/
static void
UserRecvCallback( u8 aid, u8* buffer, int size )
{
#ifdef SDK_FINALROM
#pragma unused(aid, buffer, size)
#endif
	OS_TPrintf( "from aid = %d  size = %d  buffer[0] = %X\n", aid, size, buffer[0] );
}

/** -------------------------------------------------------------------------
  Set receive buffer for DWC transport
  ---------------------------------------------------------------------------*/
void
SetRecvBuffer( void )
{
	int i;
    int j = 0;
	int connectingNum = DWC_GetNumConnectionHost()	;

    OS_TPrintf( "connection host = %d\n", connectingNum );
    
    for ( i = 0; i < connectingNum; ++i )
    {
		if ( i == DWC_GetMyAID() )
        {
            j++;
			continue;
        }

        DWC_SetRecvBuffer( (u8)i, &sRecvBuffer[i-j], SIZE_RECV_BUFFER );
    }
}

/** -------------------------------------------------------------------------
  Reliable send
  ---------------------------------------------------------------------------*/
static void
BroadCastReliable( void )
{
    u8* pAidList;
    int numHost;
    int i;

    numHost = DWC_GetAIDList(&pAidList);

    for (i = 0; i < numHost; i++){
        if (pAidList[i] == DWC_GetMyAID())
        {
            continue;
        }

        DWC_SendReliable( pAidList[i], sSendBuffer, SIZE_SEND_BUFFER );
    }
}

/** -------------------------------------------------------------------------
  Unreliable transfer
  ---------------------------------------------------------------------------*/
static void
BroadCastUnreliable( void )
{
    u8* pAidList;
    int numHost;
    int i;

    numHost = DWC_GetAIDList(&pAidList);

    for (i = 0; i < numHost; i++){
        if (pAidList[i] == DWC_GetMyAID()) continue;

        DWC_SendUnreliable( pAidList[i], sSendBuffer, SIZE_SEND_BUFFER );
    }
}

/*---------------------------------------------------------------------------*
  Main function after mesh connection has been completed
 *---------------------------------------------------------------------------*/
GameMode GameConnectedMain(void)
{
    GameSequence *gameSeq = &gameSeqList[GAME_MODE_CONNECTED];
    GameMode returnSeq = GAME_MODE_NUM;
    int curIdx = 0;
    
    // Sets the send callback	
    DWC_SetUserSendCallback( SendDoneCallback ); 

    // Set the receive callback	
    DWC_SetUserRecvCallback( UserRecvCallback ); 
    
    // Set the receive buffer	
    SetRecvBuffer();

    // Set the callback for closing connections
    DWC_SetConnectionClosedCallback(ConnectionClosedCallback, &returnSeq);

#if 0
	// Set the packet loss rate
	if ( DWC_GetMyAID() != 0 )
    {
		DWC_SetSendDrop( 30, 0 );
		DWC_SetRecvDrop( 30, 0 );        
    }
#endif
#if 0
    // Set the delay (unit: msec)  
	if ( DWC_GetMyAID() != 0 )
    {
		DWC_SetSendDelay( 500, 0 );
		DWC_SetRecvDelay( 500, 0 );        
    }
#endif    

    OS_TPrintf( "my aid = %d\n", DWC_GetMyAID() );
    
    while (1){
        dbs_Print( 0, 0, "s:%d", DWC_GetMatchingState() );
        dbs_Print( 7, 0, "n:%d", DWC_GetNumConnectionHost() );
        dbs_Print( 0, 1, "w:%d", DWC_GetLinkLevel() );
        dbs_Print( 10,1, "p:%d", stGameCnt.userData.profileID );

        ReadKeyData();  // Get key data

        DWC_ProcessFriendsMatch();  // Update DWC communication processing

        if (DWC_GetLastError(NULL)){
            // Processing when matchmaking failed
            ShutdownInet();  // Network disconnection
            DWC_ClearError();
            returnSeq = GAME_MODE_MAIN;
            break;
        }

        if (stGameCnt.blocking){
            // Prohibit key operations during DWC processing
            dbs_DemoUpdate();
            OS_WaitIrq(TRUE, OS_IE_V_BLANK);
            Heap_Debug();
            dbs_DemoLoad();
            // Check for stack overflow
            OS_CheckStack(OS_GetCurrentThread());
            continue;
	    }

        // Escapes the loop if the next mode to enter has been set
        if (returnSeq != GAME_MODE_NUM) break;

        ////////// Process the following key operations
        if (stKeyCnt.trg & PAD_BUTTON_A){
            // Select the menu with the A button
            returnSeq = gameSeq->menuList[curIdx].mode;
            stGameCnt.blocking = gameSeq->menuList[curIdx].blocking;

            switch (curIdx){
            case 0:             // Reliable transfer
                BroadCastReliable();
				break;
            case 1:             // Unreliable transfer
				BroadCastUnreliable();
                break;
            case 2:             // Close connection
                DWC_CloseConnectionsAsync();

                if (DWC_GetNumConnectionHost() <= 1){
                    // Terminate immediately and do not call the callback if the number of hosts already connected is 0.
                    stGameCnt.blocking = 0;
                }
                break;
            default:
                break;
            }
        }
        else if (stKeyCnt.trg & PAD_KEY_UP){
            // Move the cursor using up on the crosskey
            curIdx--;
            if (curIdx < 0) curIdx = gameSeq->numMenu-1;
            // Redisplay menu list
            DispMenuMsgWithCursor(gameSeq, curIdx, NULL);
        }
        else if (stKeyCnt.trg & PAD_KEY_DOWN){
            // Move the cursor using down on the crosskey
            curIdx++;
            if (curIdx >= gameSeq->numMenu) curIdx = 0;
            // Redisplay menu list
            DispMenuMsgWithCursor(gameSeq, curIdx, NULL);
        }
        ////////// Key operation processing ends here

        dbs_DemoUpdate();
        OS_WaitIrq(TRUE, OS_IE_V_BLANK);
            Heap_Debug();
        dbs_DemoLoad();
        // Check for stack overflow
        OS_CheckStack(OS_GetCurrentThread());
    }

    return returnSeq;
}

/*---------------------------------------------------------------------------*
  Callback function for closing connections
 *---------------------------------------------------------------------------*/
static void ConnectionClosedCallback(DWCError error,
                                     BOOL isLocal,
                                     BOOL isServer,
                                     u8  aid,
                                     int index,
                                     void* param)
{
#ifdef SDK_FINALROM
#pragma unused(aid, index)
#endif

    if (error == DWC_ERROR_NONE){
        if (!isLocal && isServer){
            // When using server-client matchmaking and the local host is a client,
            // the server DS closed the connection.
            // Since the server DS has closed, the local host also closes.
            DWC_CloseConnectionsAsync();
        }
        
        if (DWC_GetNumConnectionHost() == 1){
            // When all connections are closed,
            if (isLocal){
                // Proceed if the connections with all members were closed when closed by the local host
                //  
                stGameCnt.blocking = FALSE;  // Cancel blocking
            }
            if ((DWC_GetLastMatchingType() != DWC_MATCH_TYPE_SC_SV) || isServer){
                // if closed by the local host when isServer==TRUE,
                // or if closed by another host, set the next mode here
                *(GameMode *)param = GAME_MODE_LOGIN;
            }
            // When the self  is on the server-client matchmaking server DS,
            // do nothing if a client closes the connections of all members
        }

        if (isLocal){
            OS_TPrintf("Closed connection to aid %d (friendListIndex = %d) Rest %d.\n",
                       aid, index, DWC_GetNumConnectionHost());
        }
        else {
            OS_TPrintf("Connection to aid %d (friendListIndex = %d) was closed. Rest %d.\n",
                       aid, index, DWC_GetNumConnectionHost());
        }
    }
    else {
        DispErrorMessage();
        OS_TPrintf("Failed to close connections. %d\n\n", error);

        *(GameMode *)param = GAME_MODE_MAIN;

        stGameCnt.blocking = FALSE;  // Cancel blocking
    }
}


