/*---------------------------------------------------------------------------*
  Project:  NitroDWC - ranking sample program - 
  File:     main.c

  Copyright 2005 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: main.c,v $
  Revision 1.14  06/16/2006 04:53:39  hayashi
  added crc feature to the user defined data

  Revision 1.13  06/01/2006 13:25:25  hayashi
  added sample to cancel

  Revision 1.12  05/24/2006 09:17:02  hayashi
  changed to descending order at getting order part

  Revision 1.11  05/24/2006 06:20:40  hayashi
  change the initialize data for Ranking which is adapted to "dwctest"

  Revision 1.10  2006/05/22 07:29:57  hayashi
  changed to be set the server by DWC_SetAuthServer()

  Revision 1.9  05/17/2006 02:47:09  hayashi
  add "last update" output

  Revision 1.8  05/12/2006 08:28:54  hayashi
  NOTINITIALIZED -> NOTREADY
  RIVAL -> NEAR
  DWCAccUserData -> DWCUserData
  BOOL	DWC_RnkShutdown() -> DWCRnkError

  Revision 1.7  04/26/2006 09:25:40  hayashi
  specification is changed related to PID handling

  Revision 1.6  04/17/2006 06:06:38  takayama
  SVC_WaitVBlankIntr() -> OS_WaitVBlankIntr()

  Revision 1.5  04/13/2006 11:35:19  hayashi
  Change prefix Ranking->Rnk

  Revision 1.4  04/13/2006 09:01:37  hayashi
  character code converting

  Revision 1.3  04/11/2006 09:01:33  hayashi
  Made the connection target server selectable with DWC_RnkInitialize

  Revision 1.2  04/11/2006 07:04:19  hayashi
  The base64 encoding size calculation was erroneous, so revised

  Revision 1.1  04/10/2006 10:16:58  hayashi
  added the ranking module


  $NoKeywords: $
 *---------------------------------------------------------------------------*/

#include <nitro.h>
#include <ninet/ip.h>
#include <dwc.h>

// Header file include
//-------------------------------------
#include "main.h"
#include "init.h"
#include "utility.h"


// Constant definitions
//---------------------------------------------------------

#define RANKING_INITDATA "hxBfMOkOvlOHpUGfQlOb0001ed350000685d0040000037012e6bdwctest"
#define CATEGORY		10

// Structure type declaration
//---------------------------------------------------------

// User-defined data
typedef struct _tagMYUSERDEFINEDDATA{
	u32		crc;		// CRC
	u32		size;		// size
	u8		data[32];	// Data (32 bytes)
}MYUSERDEFINEDDATA;


// Global variable declarations
//---------------------------------------------------------
DWCInetControl stConnCtrl;


// Function Prototype
//---------------------------------------------------------
static BOOL StartIPMain(void);
static void VBlankIntr(void);

/*---------------------------------------------------------------------------*
  Function for allocating memory
 *---------------------------------------------------------------------------*/
static void*
AllocFunc( DWCAllocType name, u32   size, int align )
{
	void * ptr;
	OSIntrMode old;
	(void)name;
	(void)align;

	old = OS_DisableInterrupts();

	ptr = OS_AllocFromMain( size );

	OS_RestoreInterrupts( old );

	return ptr;
}

/*---------------------------------------------------------------------------*
  Function for freeing memory
 *---------------------------------------------------------------------------*/
static void
FreeFunc( DWCAllocType name, void* ptr,  u32 size  )
{
	OSIntrMode old;
	(void)name;
	(void)size;

	if ( !ptr ) return;

	old = OS_DisableInterrupts();

	OS_FreeToMain( ptr );

	OS_RestoreInterrupts( old );
}


//=============================================================================
/*!
 *	@brief	Get IP
 *
 *	@retval	TRUE		Succeeded
 *	@retval	FALSE   	Failed
 */
//=============================================================================
static BOOL StartIPMain(void)
{
	DWC_InitInet( &stConnCtrl );

	DWC_ConnectInetAsync();

	// Wait for stable start.
	while ( !DWC_CheckInet() )
	{
		DWC_ProcessInet();

		OS_WaitVBlankIntr();

	}

	switch ( DWC_GetInetStatus() )
	{
	case DWC_CONNECTINET_STATE_ERROR:
		{
			// error display
			DWCError err;
			int errcode;
			err = DWC_GetLastError( &errcode );

			OS_TPrintf("   Error occurred %d %d.\n", err, errcode );
		}
		DWC_ClearError();

		DWC_CleanupInet();
		break;
	case DWC_CONNECTINET_STATE_FATAL_ERROR:
		{
			// error display
			DWCError err;
			int errcode;
			err = DWC_GetLastError( &errcode );

			OS_Panic("   Error occurred %d %d.\n", err, errcode );
		}
		break;

	case DWC_CONNECTINET_STATE_CONNECTED:
		OS_TPrintf("   CONNECTED!!!.\n");
		return TRUE;
	}
	return FALSE;
}

/*---------------------------------------------------------------------------*
  V-blank interrupt function
 *---------------------------------------------------------------------------*/
static void VBlankIntr(void)
{
	OS_SetIrqCheckFlag(OS_IE_V_BLANK);
}

//=============================================================================
/*!
 *	@brief	Main
 */
//=============================================================================
static u8 s_Work[ DWC_INIT_WORK_SIZE ] ATTRIBUTE_ALIGN( 32 );

void NitroMain()
{

	DWCAccUserData	userdata;	// DWC user data
	DWCRnkError		res;		// For ranking library error acquisition
	DWCRnkGetParam	param;		// Parameters when acquiring ranking

	MYUSERDEFINEDDATA	userdefineddata;	// User-defined data

	u32 order;
	u32 count;
	s32 timeout;
	int i;

	// initialize
	//-------------------------------------------
	OS_Printf("initialize\n");

	OS_Init();
	OS_InitTick();
	OS_InitAlarm();
	RTC_Init();

	// Initial settings for checking for stack overflow
	OS_SetThreadStackWarningOffset(OS_GetCurrentThread(), 0x100);
	
	PXI_Init();
	
	Heap_Init();
	Alarm_Init();

	GX_Init();
	FS_Init( MI_DMA_MAX_NUM ); 

	GX_DispOff();
	GXS_DispOff();

	// Enable V-blank interrupt
	OS_SetIrqFunction(OS_IE_V_BLANK, VBlankIntr);
	(void)OS_EnableIrqMask(OS_IE_V_BLANK);
	(void)OS_EnableIrq();
	(void)OS_EnableInterrupts();
	(void)GX_VBlankIntr(TRUE);
	GX_SetDispSelect(GX_DISP_SELECT_SUB_MAIN);

	// Starts display
	OS_WaitVBlankIntr();			  // Waiting the end of V-Blank interrupt
	GX_DispOn();
	GXS_DispOn();

	// Initialize DWC library
	DWC_Init( s_Work );
	
	// Specify debug display level
	DWC_SetReportLevel(DWC_REPORTFLAG_ALL);
	
	DWC_SetMemFunc( AllocFunc, FreeFunc );

	//Get IP
	while(!StartIPMain())
	{
		OS_Printf("error at StartIPMain()\n");	
	}

	//---------------------------------
	// Login process
	//---------------------------------
	//
	// Perform the user data acquisition process here. The ranking library
	// is dependent upon GameSpy, so unless this is user data that has logged in at least once
	// and is acquiring a valid GameSpyProfileID, the initialization will fail.
	//  A temporary ID is not valid.
	//
	// If there was already a login in the past and the acquired user data is being kept as a backup,
	// you will be able to use that data. However, as outlined in the guidelines,
	// be aware that you will need to log in each time you connect to the Internet.
	//  
	//
	// If using an external authentication server, etc., you will be unable to acquire a valid GameSpyProfileID,
	// so you cannot use the ranking library.  
	// In such a case, the GameSpy server resources cannot be used.
	//
	// <Precaution>
	//
	// For the sake of simplicity in this sample, we are forcing the user data to be generated,
	// but you should actually try to use the proper user data acquired by logging in.
	//   If any other data is passed, it is possible for the ranking module initialization
	// to succeed, though the operations are undefined.
	//
	{
		// User Data Configuration
		//
		//
		u32 dummyuserdata[DWC_ACC_USERDATA_BUFSIZE] = {
								0x00000040,0xa1b2c3d4,0x5e6f7a8b,0xc9d0e1f2,
								0x00000996,0x3a4b5c6d,0xe7f8a9b0,0x0001e240,
								0x1c2d3e4f,0xa5b6c7d8,0x9e0f1a2b,0xc3d4e5f6,
								0x7a8b9c0d,0xe1f2a3b4,0x5c6d7e8f,0xcfa4db0e };

		memcpy( &userdata, dummyuserdata, DWC_ACC_USERDATA_BUFSIZE );

		// Connection Target Server Configuration
		//
		// DWC_CONNECTINET_AUTH_TEST	Development-oriented server
		// DWC_CONNECTINET_AUTH_RELEASE	Product-oriented server
		//
		// There are also two servers prepared in the ranking library. Which one to use is determined
		// by the authentication server that is used. Here, for the sake of convenience,
		// the call is explicit, but be sure to implement the proper login step.
		// other than obj.
		//
		DWC_SetAuthServer( DWC_CONNECTINET_AUTH_TEST );
	}

	//---------------------------------
	// Ranking Module Process
	//---------------------------------

	// Ranking Module Initialization
	res = DWC_RnkInitialize(	RANKING_INITDATA,
								&userdata );

	switch( res ){
	case DWC_RNK_SUCCESS:
		OS_Printf("DWC_RnkInitialize success\n");
		break;
	case DWC_RNK_ERROR_INIT_ALREADYINITIALIZED:
		OS_Printf("[err]DWC_RnkInitialize already initialized\n");
		goto exit;
	case DWC_RNK_ERROR_INIT_INVALID_INITDATASIZE:
		OS_Printf("[err]DWC_RnkInitialize invalid init data size\n");
		goto exit;
	case DWC_RNK_ERROR_INIT_INVALID_INITDATA:
		OS_Printf("[err]DWC_RnkInitialize invalid init data\n");
		goto exit;
	case DWC_RNK_ERROR_INIT_INVALID_USERDATA:
		OS_Printf("[err]DWC_RnkInitialize invaild user data\n");
		goto exit;
	}

	//
	// Upload
	//

	// User-defined data generation
	//
	// 768 bytes of optional binary data can be designated in the user-defined data.
	//  This data can be acquired at the same time the ranking list is acquired.
	//  
	//
	// It is possible that a malicious user can set up a fake server to illegally re-write this data.
	//  Such illegally re-written data can cause the program to
	// run out of control, so in order to prevent the execution of any malicious code,
	// you should code with the assumption that any data input by a client cannot be trusted.
	//  
	//
	// In this sample, the size and CRC are recorded in the user-defined data, and the validity of that data is checked.
	//  Actually, you should be aware that overflows, etc. can result from searches for data presumed to have NULL terminators,
	// and you should code so that no memory region greater than the size of the user-defined data is accessed.
	//  
	//
	userdefineddata.size = 32;
	strcpy((char *)userdefineddata.data, "test data");
	userdefineddata.crc = (u32)MATH_CalcChecksum8(
										(const void*)userdefineddata.data,
										userdefineddata.size );

	// Start score upload request
	res = DWC_RnkPutScoreAsync(	CATEGORY,				// category
								DWC_RNK_REGION_JP,		// region
								1234,					// score
								(void*)&userdefineddata,// user defined data
								userdefineddata.size );

	switch( res ){
	case DWC_RNK_SUCCESS:
		OS_Printf("DWC_RnkPutScoreAsync success\n");
		break;
	case DWC_RNK_IN_ERROR:
		OS_Printf("[err]DWC_RnkPutScoreAsync in error\n");
		goto exit;
	case DWC_RNK_ERROR_PUT_NOTREADY:
		OS_Printf("[err]DWC_RnkPutScoreAsync not ready\n");
		goto exit;
	case DWC_RNK_ERROR_INVALID_PARAMETER:
		OS_Printf("[err]DWC_RnkPutScoreAsync invalid parameter\n");
		goto exit;
	case DWC_RNK_ERROR_PUT_INVALID_KEY:
		OS_Printf("[err]DWC_RnkPutScoreAsync invalid key\n");
		goto exit;
	case DWC_RNK_ERROR_PUT_NOMEMORY:
		OS_Printf("[err]DWC_RnkPutScoreAsync nomemory\n");
		goto exit;
	}

	// Asynchronous process

	timeout = 5000;

	while( (res = DWC_RnkProcess()) == DWC_RNK_SUCCESS ){
	
		OS_Sleep(10);

		// Time out after 5 seconds have passed
		//  -> Cancel and end
		if( (timeout -= 10) < 0 ){

			res = DWC_RnkCancelProcess();

			switch(res){
			case DWC_RNK_SUCCESS:
				OS_Printf("[timeout] DWC_RnkCancelProcess\n");
				goto exit;
			case DWC_RNK_IN_ERROR:
				break;
			case DWC_RNK_ERROR_CANCEL_NOTASK:
				break;
			}

		}
	
	}

	switch( res ){
	case DWC_RNK_PROCESS_NOTASK:
		OS_Printf("DWC_RnkProcess done\n");
		break;
	case DWC_RNK_IN_ERROR:
		OS_Printf("[err]DWC_RnkProcess in error [code:%d]\n", DWC_GetLastError( NULL ) );
		goto exit;
	}

	// Get network communications results
	switch( DWC_RnkGetState() ){
	
	case DWC_RNK_STATE_COMPLETED:
		OS_Printf("DWC_RnkProcess Succeeded \n");
		break;

	case DWC_RNK_STATE_ERROR:
		OS_Printf("[err]DWC_RnkGetState [code:%d]\n", DWC_RnkGetState() );
		goto exit;
	
	}



	//
	// Get local order
	//

	// Start order acquisition request

	param.size = sizeof( param.order );
	param.order.since = 0;
	param.order.sort = DWC_RNK_ORDER_DES;

	res = DWC_RnkGetScoreAsync(	DWC_RNK_GET_MODE_ORDER,		// mode
								CATEGORY,					// category
								DWC_RNK_REGION_JP,			// region
								&param );					// parameter

	switch( res ){
	case DWC_RNK_SUCCESS:
		OS_Printf("DWC_RnkGetScoreAsync success\n");
		break;
	case DWC_RNK_IN_ERROR:
		OS_Printf("[err]DWC_RnkGetScoreAsync in error\n");
		goto exit;
	case DWC_RNK_ERROR_GET_NOTREADY:
		OS_Printf("[err]DWC_RnkGetScoreAsync not ready\n");
		goto exit;
	case DWC_RNK_ERROR_INVALID_PARAMETER:
		OS_Printf("[err]DWC_RnkGetScoreAsync invalid parameter\n");
		goto exit;
	case DWC_RNK_ERROR_GET_INVALID_KEY:
		OS_Printf("[err]DWC_RnkGetScoreAsync invalid key\n");
		goto exit;
	case DWC_RNK_ERROR_GET_NOMEMORY:
		OS_Printf("[err]DWC_RnkGetScoreAsync nomemory\n");
		goto exit;
	}

	// Asynchronous process

	timeout = 5000;

	while( (res = DWC_RnkProcess()) == DWC_RNK_SUCCESS ){
	
		OS_Sleep(10);

		// Time out after 5 seconds have passed
		//  -> Cancel and end
		if( (timeout -= 10) < 0 ){

			res = DWC_RnkCancelProcess();

			switch(res){
			case DWC_RNK_SUCCESS:
				OS_Printf("[timeout] DWC_RnkCancelProcess\n");
				goto exit;
			case DWC_RNK_IN_ERROR:
				break;
			case DWC_RNK_ERROR_CANCEL_NOTASK:
				break;
			}

		}
	
	}

	switch( res ){
	case DWC_RNK_IN_ERROR:
		OS_Printf("[err]DWC_RnkProcess in error [code:%d]\n", DWC_GetLastError( NULL ) );
		goto exit;
	case DWC_RNK_PROCESS_NOTASK:
		OS_Printf("DWC_RnkProcess done\n");
		break;
	}

	// Get network communications results
	switch( DWC_RnkGetState() ){
	
	case DWC_RNK_STATE_COMPLETED:
		OS_Printf("DWC_RnkProcess success \n");
		break;

	case DWC_RNK_STATE_ERROR:
		OS_Printf("[err]DWC_RnkGetState [code:%d]\n", DWC_GetLastError( NULL ) );
		goto exit;
	
	}

	// Display
	res = DWC_RnkResGetOrder( &order );
	switch( res ){
	case DWC_RNK_SUCCESS:
		break;
	case DWC_RNK_IN_ERROR:
		OS_Printf("[err]DWC_RnkResGetOrder in error [code:%d]\n", DWC_GetLastError( NULL ) );
		goto exit;
	case DWC_RNK_ERROR_INVALID_MODE:
		OS_Printf("[err]DWC_RnkResGetOrder invalid mode\n");
		goto exit;
	case DWC_RNK_ERROR_INVALID_PARAMETER:
		OS_Printf("[err]DWC_RnkResGetOrder invalid parameter\n");
		goto exit;
	case DWC_RNK_ERROR_NOTCOMPLETED:
		OS_Printf("[err]DWC_RnkResGetOrder not completed\n");
		goto exit;
	case DWC_RNK_ERROR_EMPTY_RESPONSE:
		OS_Printf("[err]DWC_RnkResGetOrder empty response\n");
		goto exit;
	}

	OS_Printf("order=%d\n", order);


	//
	// Get ranking
	//

	// Start top ranking acquisition request
	param.size = sizeof( param.toplist );
	param.toplist.sort = DWC_RNK_ORDER_DES;
	param.toplist.limit = 10;
	param.toplist.since = 0;

	res = DWC_RnkGetScoreAsync(	DWC_RNK_GET_MODE_TOPLIST,	// mode
									CATEGORY,				// category
									DWC_RNK_REGION_JP,		// region
									&param );				// parameter

	switch( res ){
	case DWC_RNK_SUCCESS:
		OS_Printf("DWC_RnkGetScoreAsync success\n");
		break;
	case DWC_RNK_IN_ERROR:
		OS_Printf("[err]DWC_RnkGetScoreAsync in error\n");
		goto exit;
	case DWC_RNK_ERROR_GET_NOTREADY:
		OS_Printf("[err]DWC_RnkGetScoreAsync not ready\n");
		goto exit;
	case DWC_RNK_ERROR_INVALID_PARAMETER:
		OS_Printf("[err]DWC_RnkGetScoreAsync invalid parameter\n");
		goto exit;
	case DWC_RNK_ERROR_GET_INVALID_KEY:
		OS_Printf("[err]DWC_RnkGetScoreAsync invalid key\n");
		goto exit;
	case DWC_RNK_ERROR_GET_NOMEMORY:
		OS_Printf("[err]DWC_RnkGetScoreAsync nomemory\n");
		goto exit;
	}

	// Asynchronous process

	timeout = 5000;

	while( (res = DWC_RnkProcess()) == DWC_RNK_SUCCESS ){
	
		OS_Sleep(10);

		// Time out after 5 seconds have passed
		//  -> Cancel and end
		if( (timeout -= 10) < 0 ){

			res = DWC_RnkCancelProcess();

			switch(res){
			case DWC_RNK_SUCCESS:
				OS_Printf("[timeout] DWC_RnkCancelProcess\n");
				goto exit;
			case DWC_RNK_IN_ERROR:
				break;
			case DWC_RNK_ERROR_CANCEL_NOTASK:
				break;
			}

		}
	
	}

	switch( res ){
	case DWC_RNK_IN_ERROR:
		OS_Printf("[err]DWC_RnkProcess in error [code:%d]\n", DWC_GetLastError( NULL ) );
		goto exit;
	case DWC_RNK_PROCESS_NOTASK:
		OS_Printf("DWC_RnkProcess done\n");
		break;
	}

	// Get network communications results
	switch( DWC_RnkGetState() ){

	case DWC_RNK_STATE_COMPLETED:
		OS_Printf("DWC_RnkProcess success \n");
		break;

	case DWC_RNK_STATE_ERROR:
		OS_Printf("[err]DWC_RnkGetState [code:%d]\n", DWC_GetLastError( NULL ) );
		goto exit;

	}


	// Get list

	res = DWC_RnkResGetRowCount( &count );

	switch( res ){
	case DWC_RNK_SUCCESS:
		break;
	case DWC_RNK_IN_ERROR:
		OS_Printf("[err]DWC_RnkResGetRowCount in error [code:%d]\n", DWC_GetLastError( NULL ) );
		goto exit;
	case DWC_RNK_ERROR_INVALID_MODE:
		OS_Printf("[err]DWC_RnkResGetRowCount invalid mode\n");
		goto exit;
	case DWC_RNK_ERROR_INVALID_PARAMETER:
		OS_Printf("[err]DWC_RnkResGetRowCount invalid parameter\n");
		goto exit;
	case DWC_RNK_ERROR_NOTCOMPLETED:
		OS_Printf("[err]DWC_RnkResGetRowCount not completed\n");
		goto exit;
	case DWC_RNK_ERROR_EMPTY_RESPONSE:
		OS_Printf("[err]DWC_RnkResGetRowCount empty response\n");
		goto exit;
	}
	
	for( i=0; i<count; i++ ){

		DWCRnkData data;

		if( DWC_RnkResGetRow( &data, (u32)i ) != DWC_RNK_SUCCESS ){
			OS_Printf("[err]DWC_RnkResGetRow()\n");
			break;
		}

		// Output
		OS_TPrintf("%dth score=%d pid=%d region=%d update=%dmin before ",
						data.order,
						data.score,
						data.pid,
						data.region,
						data.lastupdate );

		// Check validity of user-defined data
		{
			MYUSERDEFINEDDATA*	mydata;
			
			mydata = (MYUSERDEFINEDDATA*)data.userdata;

			if(mydata->size != 32){

				// Data size is invalid
				OS_TPrintf("userdata=[invalid data size]\n");

			}else if(mydata->crc != (u32)MATH_CalcChecksum8(
										(const void*)mydata->data,
										mydata->size)){

				// CRC is invalid
				OS_TPrintf("userdata=[invalid crc]\n");
			
			}else{

				// Valid data
				OS_TPrintf("userdata=%s\n", mydata->data);

			}

		}

	}

exit:
	// End ranking module
	DWC_RnkShutdown();
	
	// Quit
	OS_Printf("Terminated.\n");	

	OS_Terminate();

}
